/*
 ***** BEGIN LICENSE BLOCK *****
 * Version: EPL 2.0/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Eclipse Public
 * License Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.eclipse.org/legal/epl-v20.html
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * Copyright (C) 2013-2017 The JRuby Team (jruby@jruby.org)
 * 
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the EPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the EPL, the GPL or the LGPL.
 ***** END LICENSE BLOCK *****/
package org.jruby.ext.ripper;

public class YyTables {
   private static short[] combine(short[] t1, short[] t2, 
                                  short[] t3, short[] t4) {
      short[] t = new short[t1.length + t2.length + t3.length + t4.length];
      int index = 0;
      System.arraycopy(t1, 0, t, index, t1.length);
      index += t1.length;
      System.arraycopy(t2, 0, t, index, t2.length);
      index += t2.length;
      System.arraycopy(t3, 0, t, index, t3.length);
      index += t3.length;
      System.arraycopy(t4, 0, t, index, t4.length);
      return t;
   }

   public static final short[] yyTable() {
      return combine(yyTable1(), yyTable2(), yyTable3(), yyTable4());
   }

   public static final short[] yyCheck() {
      return combine(yyCheck1(), yyCheck2(), yyCheck3(), yyCheck4());
   }
   private static final short[] yyTable1() {
      return new short[] {

          309,  378,  312,   88,   88,  135,  135,  429,  219,  693, 
          254,  254,  254,  108,  232,  280,  254,  254,  358,  489, 
          254,  311,  419,  218,  218,  422,  356,  721,  296,  296, 
          264,  467,  296,  647,  232,  675,  648,  258,  261,  292, 
          292,  712,  234,  292,  218,  764,   88,  699,  711,    8, 
          354,  711,  254,  393,  364,  305,  308,  647,  642,    8, 
          335,  339,  234,  528,  544,  661,  547,  530,  279,  218, 
           83,   83,  295,  295,  328,  627,  295,  583,  659,  647, 
          712,  647,  647,  810,  858,  582,  134,  134,  606,  634, 
          433,  635,  623,  437,  631,  636,  134,  399,  637,  830, 
          260,  354,  393,  673,  334,  442,  647,  260,    8,  647, 
          575,  583,  839,  281,  958,  260,  647,  306,  636,  689, 
          647,  432,  674,  647,  264,  894,  583,  878,  647,  567, 
          254,  134,   88,  702,  835, 1046,  623,  606,  631,  376, 
         1030,  434,  647,  851,  575,  260,  399,  305,  350,  259, 
          500,  647,  674,  823,  502,  750,  259,  516,  647,  575, 
          301,  647,  833,  304,  259,  647,  318,  582,  414,  647, 
          581,  434,  520,  521,   74,  593,  308,  567,  567,  636, 
          544,  575,  134,  280,  871,  414,  674,  314,  463,  552, 
          319,  647,  733,  555,  259,  317,  636,  350,  556,  413, 
          606,  582,  376,  435,  581,   93,   93,  136,  136,  593, 
          647,  674,  567,  567,  477,  575,  582,  281,  884,  581, 
          489,  377,  503,  504,  593,   72,  567, 1046,  386,  777, 
          575,  561,  562,  252,  252,  252,   84,  379,  552,   88, 
          975, 1030,  384,  369,  847,  387,  740,  382,   93,  606, 
          511,  438,  355,  385,  872,  809,   75,  638,   53,  638, 
          532,  780,  653,  304,  638,  814,  254,  254,  380,  712, 
          383,  582,  525,  958,  813,  563,  815,  468,  711,  711, 
          264,  740,  642,  567,  377,  381,   76,   84,  536,  742, 
           87,  746,  472,  473,  743,  853,  854,  647,  855,  370, 
          642,  525,  254,  355,  254,  254,  524,   73,  254,  519, 
          254,  907,  415,  386,  296,   88,  296,  914,   82,  416, 
          421,   88,   80,   88,  769,  292,  373,  292,  292,  770, 
          818,  993,   88,  713,   93,  527,  878,   76,  384,  385, 
          518,  717,  718,  264,  308,  401,  647,  972,  440,  403, 
          404,  328,  441,  439,  354,  453,  979,  458,  295,  726, 
          295,  983,  374,   59,   60,  425,  647,  902,  648,   82, 
          647,  648,  458,   80,  647,  647,  323,  647,  696,  647, 
          254,  254,  254,  254,   88,  254,  254,  307,  514,  292, 
          582,  639,  582,  639,  928,  636,  647,  657,  639,  522, 
          279,  636,  393,  279,  453,  655,  588,  589,  590,  591, 
          647,  647,    8,  647,  950,  903,  905, 1016,  647,  254, 
          672,  910,  912,  538,  539,  641,  583,  641,   88,  354, 
          541,  254,  641,  843,   88,  134,  632,  606,  647,  859, 
          632,   93,  292,  663,  532,  281,  399,  254,  281,   88, 
          254,  592,  634,  633,  635,  921,  328,  849,  647,  575, 
           88,  567,  323,  254,  632,  581,  859,  836,   88,  582, 
          292,  582,  623,  822,  631,  469,  575,  254,  746,  863, 
          632,  633,  736,  652,  552,  898,  292,  652, 1018,  252, 
          252,  698,  698, 1097,  632,  606,  803,  633,  710,  719, 
          292,   83,  581,  123,  712,   84,   74,  254, 1074,   88, 
          134,  633,  232,  575,  711,  280,  582,   93,  280,  812, 
           88,  552,  218,   93,  593,   93,  694,  317,  632,   96, 
           96,  499,  606,  501,   93,  581,  254,  729,  354,  424, 
          234,  282,  992,  426,  567,  633,  575,  418,  601,   97, 
          647,  647,  123,  632,  552,   76,  355,  538,  539,  292, 
          386,  532,   77,  845,  541,  606,  876,  843,   84,  379, 
          633,  848,   96,  591,  384,  799,   83,  387,  430,  382, 
          757,  759,  601, 1050,  440,  385,   93,   82,   75,  647, 
          778,   80,  865,   85,  333,  337,  746,  601,  746,  804, 
          380,  786,  383,  252,  252,  252,  252,  591,  593,  594, 
          431,  573,  453,   77,  638,  638,   78,  381,   76,  134, 
          638,  638,  591,  307,  384,  385,  771,  644,  443,  644, 
           93,  355,   83,  638,  644,   81,   93,  850,  647,  538, 
          539,  543,  391,  392,   85,  509,  541,  786,  825,  822, 
           82,   93,  454,  323,   80,  354,  931,  783,   96,  457, 
          772,  464,   93,  786,  706,  789,  308,   78,  254,   88, 
           93,  789,  465,  676,   79,  817,  786,  407,  408,   88, 
          582,  466,  816,   83,  453,  807,   81,  469,  232,  898, 
          746, 1026,  471,  898,  808,  103,  103,  218,  218,  474, 
         1061,  561,  562,  478,  954,  783,  254,  786,  789, 1013, 
         1083,   93,  532,  293,  299, 1015,  234,  300,  484,  784, 
          485,  916,   93,  488,  867,   79,   88,  790,  844,  674, 
          720,  354,  840,  793,  451,  874,  134,  877,  103,  652, 
          355,  638,  452,  647,  647,  508,  867,  386,  639,  639, 
          926,  886,   87,  512,  639,  639,  513,  746,  323,  746, 
          123,  320,  452,  452,  862,   96,  523,  639,  452,   73, 
          517,  876,  529,  440,  531,   88,  955,  537,   88,  656, 
          664,   88,  641,  641,  415,  898,  876,  506,  641,  641, 
          685,  416,  417,   83,  746,  320,   88,  573,  692,  254, 
          677,  641, 1025,  682,  684,  714,  354,  695,   88,  686, 
          320,  715,  134,  538,  539,  716,  632,  722,  730,  725, 
          541,  440,  292,   88,  103,  546,  254,  929,  440,  440, 
          751,   77,  752,  633,  933,  753,  446,   88,   88,   88, 
          756,   96,  320,  601,  758,  108,  647,   96,  760,   96, 
          762,  495,  942,  462,  763,  440,  766,  355,   96,  698, 
          776,  867,   85,  917,  108,  779,  781,  782,  591,  787, 
         1024,   93,   88,   88,  788,  639,  800,  962,  876,  108, 
          601,   93,  795,  966,  968,   78,  970,  134,  971,  802, 
          805,  252,  643,  811,   77,  786,  108,  601,  987,  662, 
          822,   83,   98,  828,   81,  591,  829,  831,  832,  641, 
           96,  320,  834,  601,  846,  860,  573,  864,  573,  885, 
          901,  909,  591,   92,  908,   85,  320,  927,   93,  252, 
          930,  103,  643,  355,  662,  308, 1010,  932,  591, 1059, 
           88,  857,  935,   79, 1054,  476, 1034,  939,   78,  521, 
          943,  645,  944,  645,   96, 1040,  945,  951,  645, 1044, 
           96,  538,  539,  540,   83,  495,  956,   81,  541,  647, 
          515,  538,  539,  543, 1090,   96,  952,   93,  541,  134, 
           93,  724,  959,   93,  644,  644,   96,   73,  960, 1049, 
          644,  644,  961,  495,   96,  965,  967,   88,   93,  490, 
          969,  493, 1062,  644,  973,  976,   79,  103,  355,  495, 
           93,  104,  127,  103,   88,  103,   88,  134,  647,  978, 
          980,  982,   88,  495,  103,   93,  550,  988,  551,  552, 
          553,  554,  989,  995, 1019,   96, 1020, 1023, 1032,   93, 
           93,   93, 1063, 1065, 1066, 1067,   96, 1031, 1039,  252, 
         1047, 1094, 1051, 1060,  254,   97, 1064,  786,  440,   88, 
         1068,  127,   92, 1070,  550, 1072,  551,  552,  553,  554, 
          134,  632, 1075,  415,   93,   93,  103,  292, 1080,  320, 
          416,  445,  495, 1088, 1101, 1098,  320,  320,  633,  634, 
          415,  647,  622,  320,  647,  636,   92,  416,  461,  127, 
          136, 1099,  117,  635,  640,  899,   95,  640,  900,  642, 
          320,  644,  636,   92,  108,  429,  116,  103,  636,  103, 
          103,  647,  636,  375,  103,  429,  103,  482,   98,  367, 
          138,  768,  775, 1087,  775,  727,  221,  117, 1011,  794, 
           95,  103,   93,  515,   61,   62,   63,   64,  320,  996, 
          875,  117,  103,  938,  889,  320,  320,   95,  111,  618, 
          103,  622,   98,  320,  880,  116,  775,  947,  948,  949, 
          320,  320,   97, 1042,  429, 1041,  622,  108,  792,   98, 
          915, 1086,  415,  521,  792,  462,  521,  518,  283,  416, 
          475,  136,  861,  521,  521,   96,  806,  520, 1091,   93, 
          420,  103,  963,  964,    0,   96,    0,  515,    0,  521, 
            0,    0,  103,    0,  515,  515,   93,   99,   93,  520, 
          618,  792,    0,  647,   93,    0,    0,  647,  647,  821, 
          824,  550,  824,  551,  552,  553,  554,  555,   17,  824, 
            0,  618,  556,    0,    0,    0,  123,  440,    0,  904, 
          906,    0,   96,    0,    0,  911,  913,    0,  535,    0, 
            0,   93,    0,    0,  522,  557,    0,   16,  798,  127, 
         1014,    0,   97,  559,  560,  561,  562,    0,    0,  118, 
            0,    0,  300,    0,    0,  291,  291,   17,    0,  291, 
            0,  515,  300,    0,    0,  440,  904,  906,    0,  911, 
          913,   96,  440,  440,   96,    0,    0,   96,  645,  645, 
            0,  315,  317,    0,  645,  645,   16,  291,  291,  563, 
          357,  359,   96,   97,    0,  515,  293,  645,  118,  440, 
          320,   92,  618,    0,   96,    0,    0,  320,  320,    0, 
          515,  300,    0,  622, 1055,  495, 1056,    0,    0,   96, 
          446,    0, 1057,    0,  647,  100,    0,   97,    0,  117, 
          522,  103,    0,   96,   96,   96,  106,  937,    0,    0, 
          941,  103,  515,  116,   97,   95,  619,  429,  429,  429, 
          515, 1053,  429,  429,  429,    0,  429,  515,  515, 1089, 
            0,    0,    0,  618,   92,    0,  429,   98,   96,   96, 
          106,    0,  523,  647,  520,  429,  429,  116,  429,  429, 
          429,  429,  429,    0,  618,    0,    0,  106,  103,    0, 
          521,    0,  415,    0,  518,    0,    0,  521,  521,  416, 
          510,  518,  518,    0,  520,  645,    0,  619,   95,  137, 
          137,  520,  520,  429,    0,    0,  521,  625,    0,  137, 
          429,  429,  118,  521,  521,    0,  520,  429,  619,    0, 
           98,    0, 1012,  520,  520,    0,   96,  103,    0,  206, 
          103,    0,    0,  103,  103,  103,  137,  137,  429,    0, 
          103,  103,  137,  137,  137,  137,    0,    0,  103,  336, 
           17,   17,   17,  103,  824,  415,   17,   17,    0,   17, 
          103,  522,  416,  534,    0,  415,    0,    0,  522,  522, 
          429,    0,  416,  797,    0,  103,    0,    0,  206,   16, 
           16,   16,   53,   96,    0,   16,   16,    0,   16,  103, 
          103,  103,   53,    0,    0,  137,  118,    0,    0,    0, 
           96,    0,   96,    0,  300,  300,  300,    0,   96,  300, 
          300,  300,    0,  300,    0,  291,  291,  291,  359,    0, 
          291,  618,    0,    0,  103,  103,    0,    0,    0,    0, 
           17,  291,    0,  291,  291,  300,  300,  300,  300,  300, 
          105,   53,  507,    0,    0,   96,    0,  415,    0,    0, 
          119,   17,   97,    0,  416, 1048,    0,  522,    0,   16, 
          495,  103,    0,  824,  522,  522,  647,  647,  647,  515, 
            0,    0,    0,  647,  105,    0,  515,  515,  415,    0, 
           16,  300,  618,   17,  300,  416, 1052,  538,  539,  545, 
            0,  105,  103,    0,  541,  106,  625,    0,    0,  523, 
          515,  520,    0,  618,    0,  300,  523,  523,  520,  520, 
            0,    0,   16,    0,    0,   97,  595,  596,  597,  598, 
          599,  621,    0,  600,  601,  602,  603,  604,  605,  606, 
          607,  608,  609,  610,  611,  612,  613,  300,    0,  614, 
          615,  616,  617,  102,  625,    0,    0,    0,  291,  103, 
          634,  625,  625,    0,  640,    0,  647,  621,  106,    0, 
          291,  640,    0,    0,    0,    0,  103,  550,  103,  551, 
          552,  553,  554,    0,  103,    0,  291,    0,  625,    0, 
          137,  137,  137,  137,    0,  137,  479,  480,  481,    0, 
          525,  483,  291,  292,  640,  640,  640,  868,    0,  291, 
            0,  744,  879,  292,    0,    0,  291,  137,    0,  101, 
            0,  103,    0,    0,  291,  291,  888,    0,  291,  538, 
          539,  739,    0,    0,    0,  206,  541,  206,  206,  206, 
          206,  206,    0,  538,  539,  767,  206,  292,    0,    0, 
          541,    0,  723,  640,   53,   53,   53,  521,  137,   53, 
           53,   53,  292,   53,  291,  521,  524,  291,    0,  206, 
          206,    0,    0,   53,    0,  291,    0,  206,  206,  206, 
          206,    0,   53,   53,  107,   53,   53,   53,   53,   53, 
            0,  137,  137,  137,  137,  137,    0,    0,  137,  137, 
          137,  137,  137,  137,  137,  137,  137,  137,  137,  137, 
          137,  137,  103,  104,  137,  137,  137,  137,  107,  105, 
          621,  291,    0,  206,  518,  137,   99,    0,  100,  486, 
          486,  486,    0,  137,   53,  107,  486,  651,   94,    0, 
            0,  660,    0,  119,    0,    0,    0,  104,    0,  550, 
            0,  551,  552,  553,  554,   53,    0,    0,  518,  526, 
           99,    0,  100,    0,  104,  622,    0,    0,  621,  137, 
          137,  137,    0,  518,  137,  621,  621,   99,    0,  100, 
          651,  621,  105,  744,    0,  517,  291,   53,  121,  137, 
          137,  745,  119,  137,  440,  707,  709,    0,    0,  336, 
            0,    0,  621,    0,    0,  518,  102,  522, 1027,  523, 
          551,  552,  553,  554,    0,    0,    0,  137,  137, 1033, 
            0, 1035,  291, 1036,  523,    0,    0, 1037,    0,  137, 
            0,    0,  137,    0, 1045,  709,    0,  525,  336,    0, 
          102,    0,  137,    0,  525,  525,  291,    0,  291,  379, 
          380,  381,  382,  383,  120,  291,  291,  102,   55,    0, 
            0,    0,    0,    0,    0,  292,  292,  292,   55,  882, 
          292,  292,  292,  550,  292,  551,  552,  553,  554,    0, 
            0,  896,    0,   95,  292,  292,  137,  525,    0,    0, 
            0,    0,  765,  292,  292,    0,  292,  292,  292,  292, 
          292,    0,   54,  524,    0,    0,  291,  744,    0,    0, 
          524,  524,   54,    0,    0,  957,    0,   55, 1092,    0, 
            0,  640,  640,   93, 1093,  121, 1095,  640,  640, 1096, 
          518,    0,    0,   96,    0,    0,  918,    0,  291,    0, 
          640,    0,  291,    0,    0,  292,    0,  122,  292,    0, 
          292,  137,  120,  107,  622, 1105,    0,  801,    0,  137, 
            0,   54,    0,    0,    0,    0,  292,    0,  640,  640, 
          516,  640,  640,  113,  121,  291,    0,    0,    0,    0, 
          519,    0,  104,  440,    0,    0,    0,  137,    0,    0, 
            0,    0,    0,  819,  619,   99,  526,  100,  292,    0, 
          119,  120,  622,  526,  526,    0,  525,    0,    0,  622, 
          622,  137,    0,  137,    0,  622,  107,  841,    0,  842, 
          137,  137,  517,    0,    0,    0,  709,  336,    0,  517, 
          517,  440,    0,  524,    0,  137,  622,  515,  440,  440, 
            0,  990,  518,    0,  522,  104,  523,  515,  991,  518, 
          518,  522,  522,  523,  523,  619,    0,    0,   99,    0, 
          100,  523,    0,    0,  291,  440,    0,  386,  523,  523, 
            0,  137,    0,  518,  977,  102,  619,  883,  981,    0, 
            0,  515,  896,    0,  991,  896,    0, 1043,    0,  896, 
            0,  291,    0,    0,    0,  515,  515,    0,  111,  401, 
          515,  137,    0,  403,  404,  405,  406,  137,    0,    0, 
          137,    0,    0,  925,    0,    0,    0,    0,  114,  619, 
           55,   55,   55,    0,  525,   55,   55,   55,  515,   55, 
            0,  525,  525,  137,  137,    0,  137,  137,  102,   55, 
          137,    0,    0,    0,    0,    0,  946,    0,   55,   55, 
            0,   55,   55,   55,   55,   55,    0,    0,  112,    0, 
            0,    0,    0,    0,   54,   54,   54,  518,  115,   54, 
           54,   54,    0,   54,  518,  518,    0,  896,  518,  896, 
          619,  896,  121,   54,  896,    0,    0,    0,  518,    0, 
            0,    0,    0,  291,    0,   54,   54,   54,   54,   54, 
           55,  619,  137,    0,    0,    0,  137,  516,    0,  120, 
          896,    0,    0,    0,  516,  516,    0,  519,    0,    0, 
            0,   55,  518, 1069,  519,  519, 1071, 1073,    0,  137, 
         1076, 1077,    0,    0,    0,  994,  518,  518,    0,  114, 
          137,  518,    0,  525,   54,    0,    0,    0,    0,  137, 
          525,  525,    0,   55,    0,    0,  137,    0,    0,    0, 
            0,    0,  336,    0,    0,   54,    0,  199,    0,  518, 
          524,    0,    0,    0,    0,    0,    0,  524,  524, 1100, 
         1102, 1103, 1104,    0,    0,    0,    0,    0,    0,    0, 
            0, 1106,    0,  137,  618,  618,  618,   54,  618,  515, 
          515,  515,  618,  618,  515,  515,  515,  618,  515,  618, 
          618,  618,  618,  618,  618,  618,  199,  618,  515,  515, 
          618,  618,  618,  618,  618,  618,  618,  515,  515,  618, 
          515,  515,  515,  515,  515,    0,  618,    0,    0,  618, 
          618,  618,  515,  618,  618,  618,  618,  618,  618,  618, 
          618,  618,  618,  618,  515,  515,  515,  515,  515,  515, 
          515,  515,  515,  515,  515,  515,  515,  515,  308,    0, 
          515,  515,  515,  515,  618,  527,  515,  618,  618,  515, 
          618,  618,  515,  618,  515,  618,  515,  618,  515,  618, 
          515,  515,  515,  515,  515,  515,  515,  618,  515,  618, 
          515,    0,  618,  618,  618,  618,  618,  618,    0,    0, 
          110,  618,  515,  618,  618,    0,  618,  618,  618,  618, 
            0,  411,    0,    0,    0,  897,    0,    0,    0,  618, 
          618,  618,  515,    0,    0,  619,  619,  619,    0,  619, 
          518,  518,  518,  619,  619,  518,  518,  518,  619,  518, 
          619,  619,  619,  619,  619,  619,  619,  310,  619,  518, 
          518,  619,  619,  619,  619,  619,  619,  619,  518,  518, 
          619,  518,  518,  518,  518,  518,    0,  619,    0,    0, 
          619,  619,  619,  518,  619,  619,  619,  619,  619,  619, 
          619,  619,  619,  619,  619,  518,  518,  518,  518,  518, 
          518,  518,  518,  518,  518,  518,  518,  518,  518,  109, 
            0,  518,  518,  518,  518,  619,  530,  518,  619,  619, 
          518,  619,  619,  518,  619,  518,  619,  518,  619,  518, 
          619,  518,  518,  518,  518,  518,  518,  518,  619,  518, 
          619,  518,    0,  619,  619,  619,  619,  619,  619,    0, 
            0,  625,  619,  518,  619,  619,  321,  619,  619,  619, 
          619,  625,    0,  199,  386,  199,  199,  199,  199,  199, 
          619,  619,  619,  518,  199,    0,    0,    0,    0,    0, 
          399,  400,    0,    0,    0,  520,    0,    0,    0,    0, 
            0,    0,    0,  520,    0,  625,  401,  199,  199,    0, 
          403,  404,  405,  406,    0,  199,  199,  199,  199,  625, 
          625,    0,  125,    0,  625,    0,  550,    0,  551,  552, 
          553,  554,  555,    0,    0,    0,  897,  556,    0,  897, 
            0,  897,    0,  897,    0,    0,    0,    0,    0,    0, 
            0,    0,  625,    0,    0,  129,  634,  634,  634,    0, 
          557,  199,  634,  634,    0,  634,  635,  635,  635,    0, 
          561,  562,  635,  635,    0,  635,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  386,  387,  388,  389,  390, 
          391,  392,  393,  394,  395,  396,    0,  397,  398,    0, 
          621,  399,  400,    0,  310,    0,    0,    0,    0,    0, 
          621,  310,  310,    0,  563,    0,    0,  401,    0,  402, 
            0,  403,  404,  405,  406,  407,  408,  409,    0,  410, 
            0,  897,    0,  897,    0,  897,  634,    0,  897,    0, 
            0,    0,    0,    0,  621,    0,  635,    0,    0,  386, 
            0,    0,    0,    0,  128,    0,    0,  634,  621,  621, 
            0,  124,    0,  621,  897,  399,  400,  635,    0,    0, 
            0,  550,    0,  551,  552,  553,  554,  555,    0,    0, 
            0,  401,  556,  402,    0,  403,  404,  405,  406,  634, 
            0,  621,    0,    0,    0,    0,    0,    0,    0,  635, 
            0,    0,    0,  321,    0,  557,  558,    0,    0,    0, 
          321,  321,  209,  559,  560,  561,  562,    0,  621,  621, 
          621,    0,  621,  625,  625,  625,  621,  621,  625,  625, 
          625,  621,  625,  621,  621,  621,  621,  621,  621,  621, 
            0,  625,  625,  625,  621,  621,  621,  621,  621,  621, 
          621,  625,  625,  621,  625,  625,  625,  625,  625,  563, 
          621,  209,    0,  621,  621,  621,  625,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  106,    0,  625,  625,  625,  625,  621,  194, 
          625,  621,  621,  625,  621,  621,  625,  621,  625,  621, 
          625,  621,  625,  621,  625,  625,  625,  625,  625,  625, 
          625,  621,  625,  625,  625,    0,  621,  621,  621,  621, 
          621,  621,    0,    0,    0,  621,  625,  621,  621,  625, 
          621,  621,  621,  621,    0,    0,    0,    0,  194,    0, 
            0,  122,    0,  621,  621,  621,  625,  621,  621,  621, 
            0,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,    0, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,    0,  621, 
          122,    0,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,    0,  526,  621,  621,  621,  621,  621,    0,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,    0,  621,  621,  621,  621,  621, 
          621,    0,    0,  622,  621,  621,  621,  621,    0,  621, 
          621,  621,  621,  622,    0,    0,    0,    0,    0,    0, 
            0,    0,  621,  621,  621,  621,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  209,    0, 
          209,  209,  209,  209,  209,    0,    0,  622,    0,  209, 
            0,    0,    0,  386,    0,    0,    0,  125,  625,    0, 
          522,  622,  622,    0,  126,    0,  622,    0,  522,  399, 
          400,    0,  209,  209,    0,    0,    0,    0,    0,    0, 
          209,  209,  209,  209,    0,  401,    0,  402,    0,  403, 
          404,  405,  406,    0,  622,  409,    0,  410,    0,    0, 
            0,    0,    0,    0,    0,    0,  625,    0,    0,    0, 
            0,    0,    0,  625,  625,    0,    0,    0,    0,  621, 
            0,    0,    0,    0,    0,  194,  209,  194,  194,  194, 
          194,  194,    0,    0,    0,    0,  194,    0,  122,    0, 
          625,    0,  311,    0,    0,    0,    0,  523,    0,    0, 
            0,    0,  311,    0,    0,  523,    0,    0,    0,  194, 
          194,    0,    0,    0,    0,    0,    0,  194,  194,  194, 
          194,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  311,  386,  387,  388, 
          389,  390,  391,  392,  393,  394,  395,  396,    0,  526, 
          311,  311,    0,  399,  400,  311,  526,  526,    0,    0, 
            0,    0,    0,  194,    0,    0,    0,    0,    0,  401, 
            0,  402,    0,  403,  404,  405,  406,  407,  408,  409, 
            0,  410,    0,  311,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  176,    0,    0,    0,    0,    0, 
          622,  622,  622,    0,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,    0,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,    0,  622,  176,    0,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  105,    0,  622,  622,  622,  622, 
          622,  177,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,    0,  622,  622, 
          622,  622,  622,  622,    0,    0,    0,  622,  622,  622, 
          622,  621,  622,  622,  622,  622,    0,    0,    0,    0, 
          177,    0,    0,  113,    0,  622,  622,  622,  622,  620, 
          620,  620,    0,  620,  311,  311,  311,  620,  620,  311, 
          311,  311,  620,  311,  620,  620,  620,  620,  620,  620, 
          620,    0,  620,  311,  311,  620,  620,  620,  620,  620, 
          620,  620,  311,  311,  620,  311,  311,  311,  311,  311, 
            0,  620,  113,    0,  620,  620,  620,  311,  620,  620, 
          620,  620,  620,  620,  620,  620,  620,  620,  620,  311, 
          311,  311,  311,  311,  311,  311,  311,  311,  311,  311, 
          311,  311,  311,    0,  517,  311,  311,  311,  311,  620, 
            0,  311,  620,  620,  311,  620,  620,  311,  620,  311, 
          620,  311,  620,  311,  620,  311,  311,  311,  311,  311, 
          311,  311,  620,  311,  620,  311,    0,  620,  620,  620, 
          620,  620,  620,    0,    0,  626,  620,  311,  620,  620, 
            0,  620,  620,  620,  620,  626,    0,    0,    0,    0, 
            0,    0,    0,    0,  620,  620,  620,  311,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          176,    0,  176,  176,  176,  176,  176,    0,    0,  626, 
            0,  176,    0,    0,    0,    0,    0,    0,    0,  124, 
          621,  386,  525,  626,  626,    0,  391,  392,  626,    0, 
          525,    0,    0,    0,  176,  176,    0,  399,  400,    0, 
            0,    0,  176,  176,  176,  176,    0,    0,    0,    0, 
            0,    0,    0,  401,    0,  402,  626,  403,  404,  405, 
          406,  407,  408,  409,    0,  410,    0,    0,  621,    0, 
            0,    0,    0,    0,    0,  621,  621,    0,    0,    0, 
            0,  621,    0,    0,    0,    0,    0,  177,  176,  177, 
          177,  177,  177,  177,    0,    0,    0,    0,  177,    0, 
          113,    0,  621,    0,  627,    0,    0,    0,    0,  524, 
            0,    0,    0,    0,  627,    0,    0,  524,    0,    0, 
            0,  177,  177,    0,    0,    0,    0,    0,    0,  177, 
          177,  177,  177,    0,    0,    0,    0,    0,    0,    0, 
          386,  387,  388,  389,  390,  391,  392,  393,  627,  395, 
          396,    0,    0,    0,    0,    0,  399,  400,    0,    0, 
            0,  517,  627,  627,    0,    0,    0,  627,  517,  517, 
            0,    0,  401,    0,  402,  177,  403,  404,  405,  406, 
          407,  408,  409,    0,  410,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  627,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  178,    0,    0,    0, 
            0,    0,  623,  623,  623,    0,  623,  626,  626,  626, 
          623,  623,  626,  626,  626,  623,  626,  623,  623,  623, 
          623,  623,  623,  623,    0,  626,  626,  626,  623,  623, 
          623,  623,  623,  623,  623,  626,  626,  623,  626,  626, 
          626,  626,  626,    0,  623,  178,    0,  623,  623,  623, 
          626,  623,  623,  623,  623,  623,  623,  623,  623,  623, 
          623,  623,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  107,    0,  626,  626, 
          626,  626,  623,  135,  626,  623,  623,  626,  623,  623, 
          626,  623,  626,  623,  626,  623,  626,  623,  626,  626, 
          626,  626,  626,  626,  626,  623,  626,  626,  626,    0, 
          623,  623,  623,  623,  623,  623,    0,    0,    0,  623, 
          626,  623,  623,  622,  623,  623,  623,  623,    0,    0, 
            0,    0,  135,    0,    0,  112,    0,  623,  623,  623, 
          626,  624,  624,  624,    0,  624,  627,  627,  627,  624, 
          624,  627,  627,  627,  624,  627,  624,  624,  624,  624, 
          624,  624,  624,    0,  627,  627,  627,  624,  624,  624, 
          624,  624,  624,  624,  627,  627,  624,  627,  627,  627, 
          627,  627,    0,  624,  112,    0,  624,  624,  624,  627, 
          624,  624,  624,  624,  624,  624,  624,  624,  624,  624, 
          624,  627,  627,  627,  627,  627,  627,  627,  627,  627, 
          627,  627,  627,  627,  627,    0,  516,  627,  627,  627, 
          627,  624,    0,  627,  624,  624,  627,  624,  624,  627, 
          624,  627,  624,  627,  624,  627,  624,  627,  627,  627, 
          627,  627,  627,  627,  624,  627,  627,  627,    0,  624, 
          624,  624,  624,  624,  624,    0,    0,  320,  624,  627, 
          624,  624,    0,  624,  624,  624,  624,  320,    0,    0, 
            0,    0,    0,    0,    0,    0,  624,  624,  624,  627, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  178,    0,  178,  178,  178,  178,  178,    0, 
            0,  320,    0,  178,    0,    0,    0,    0,    0,    0, 
            0,  126,  622,    0,  526,  320,  320,    0,  127,    0, 
          320,    0,  526,    0,    0,    0,  178,  178,    0,    0, 
            0,    0,    0,    0,  178,  178,  178,  178,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  320,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          622,    0,    0,    0,    0,    0,    0,  622,  622,    0, 
            0,    0,    0,  622,    0,    0,    0,    0,    0,  135, 
          178,  135,  135,  135,  135,  135,    0,    0,    0,    0, 
          135,    0,  112,    0,  622,    0,  439,    0,    0,    0, 
            0,  515,    0,    0,    0,    0,  439,    0,    0,  515, 
            0,    0,    0,  135,  135,    0,    0,    0,    0,    0, 
            0,  135,  135,  135,  135,    0,    0,    0,    0,    0, 
            0,    0,  386,  387,  388,  389,  390,  391,  392,    0, 
          439,  395,  396,    0,    0,    0,    0,    0,  399,  400, 
            0,    0,    0,  516,    0,  439,    0,    0,    0,  439, 
          516,  516,    0,    0,  401,    0,  402,  135,  403,  404, 
          405,  406,  407,  408,  409,    0,  410,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  439,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  136,    0, 
            0,    0,    0,    0,  622,  622,  622,    0,  622,  320, 
          320,  320,  622,  622,  320,  320,  320,  622,  320,  622, 
          622,  622,  622,  622,  622,  622,    0,    0,  320,  320, 
          622,  622,  622,  622,  622,  622,  622,  320,  320,  622, 
          320,  320,  320,  320,  320,    0,  622,  136,    0,  622, 
          622,  622,  320,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  320,  320,  320,  320,  320,  320, 
          320,  320,  320,  320,  320,  320,  320,  320,    0,    0, 
          320,  320,  320,  320,  622,  545,  320,  622,  622,  320, 
          622,  622,  320,  622,  320,  622,  320,  622,  320,  622, 
          320,  320,  320,  320,  320,  320,  320,  622,  320,  411, 
          320,    0,  622,  622,  622,  622,  622,  622,    0,    0, 
            0,  622,  320,  622,  622,    0,  622,  622,  622,  622, 
            0,    0,    0,    0,  545,    0,    0,  115,    0,  622, 
          622,  622,  320,  287,  287,  287,    0,  287,  439,  439, 
          439,  287,  287,  439,  439,  439,  287,  439,  287,  287, 
          287,  287,  287,  287,  287,    0,  439,  439,  439,  287, 
          287,  287,  287,  287,  287,  287,  439,  439,  287,  439, 
          439,  439,  439,  439,    0,  287,  115,    0,  287,  287, 
          287,    0,  287,  287,  287,  287,  287,  287,  287,  287, 
          287,  287,  287,  439,  439,  439,  439,  439,  439,  439, 
          439,  439,  439,  439,  439,  439,  439,  411,  519,  439, 
          439,  439,  439,  287,    0,  439,  287,    0,  439,  287, 
          287,  439,  287,  439,  287,  439,  287,  439,  287,  439, 
          439,  439,  439,  439,  439,  439,  287,  439,  439,  439, 
            0,  287,  287,  287,  287,  287,  287,    0,    0,  647, 
          287,    0,  287,  287,    0,  287,  287,  287,  287,  647, 
            0,    0,    0,    0,    0,    0,    0,    0,  287,  287, 
          287,  439,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  136,    0,  136,  136,  136,  136, 
          136,    0,    0,  647,    0,  136,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  518,    0,  647,    0, 
            0,    0,  647,    0,  518,    0,    0,    0,  136,  136, 
            0,    0,    0,    0,    0,    0,  136,  136,  136,  136, 
            0,    0,    0,  796,    0,    0,    0,    0,    0,    0, 
          647,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  386,  387,  388,  389,  390,  391,  392, 
          393,  394,  395,  396,    0,  397,  398,    0,    0,  399, 
          400,  545,  136,  545,  545,  545,  545,  545,    0,    0, 
            0,    0,  545,    0,  115,  401,    0,  402,  327,  403, 
          404,  405,  406,  407,  408,  409,    0,  410,  327,    0, 
            0,    0,    0,    0,    0,  545,    0,    0,    0,    0, 
            0,    0,    0,  545,  545,  545,  545,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  327,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  519,    0,  327,    0,    0, 
            0,  327,  519,  519,    0,    0,    0,    0,    0,  545, 
            0,  386,  387,  388,  389,  390,  391,  392,  393,  394, 
          395,  396,    0,  397,  398,    0,    0,  399,  400,  327, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  401,    0,  402,    0,  403,  404,  405, 
          406,  407,  408,  409,    0,  410,  287,  287,  287,    0, 
          287,  647,  647,  647,  287,  287,  647,  647,  647,  287, 
          647,  287,  287,  287,  287,  287,  287,  287,    0,  647, 
          647,  647,  287,  287,  287,  287,  287,  287,  287,  647, 
          647,  287,  647,  647,  647,  647,  647,    0,  287,    0, 
            0,  287,  287,  287,    0,  287,  287,  287,  287,  287, 
          287,  287,  287,  287,  287,  287,  647,  647,  647,  647, 
          647,  647,  647,  647,  647,  647,  647,  647,  647,  647, 
          258,    0,  647,  647,  647,  647,  287,  625,  647,  287, 
          258,  647,  287,  287,  647,  287,  647,  287,  647,  287, 
          647,  287,  647,  647,  647,  647,  647,  647,  647,  287, 
          647,  647,  647,    0,  287,  287,  287,  287,  287,  287, 
            0,  625,    0,  287,  258,  287,  287,    0,  287,  287, 
          287,  287,    0,    0,    0,    0,  625,    0,  258,  258, 
            0,  287,  287,  287,  647,  287,  287,  287,    0,  287, 
          327,  327,  327,  287,  287,  327,  327,  327,  287,  327, 
          287,  287,  287,  287,  287,  287,  287,    0,  625,  327, 
          327,  287,  287,  287,  287,  287,  287,  287,  327,  327, 
          287,  327,  327,  327,  327,  327,    0,  287,    0,    0, 
          287,  287,  287,    0,  287,  287,  287,  287,  287,  287, 
          287,  287,  287,  287,  287,  327,  327,  327,  327,  327, 
          327,  327,  327,  327,  327,  327,  327,  327,  327,    0, 
            0,  327,  327,  327,  327,  287,    0,  327,  287,    0, 
          327,  287,  287,  327,  287,  327,  287,  327,  287,  327, 
          287,  327,  327,  327,  327,  327,  327,  327,  287,  327, 
          515,  327,    0,  287,  287,  287,  287,  287,  287,    0, 
          515,    0,  287,    0,  287,  287,    0,  287,  287,  287, 
          287,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          287,  287,  287,  327,  550,    0,  551,  552,  553,  554, 
          555,    0,    0,    0,   92,  556,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  515, 
            0,  111,    0,  515,    0,    0,    0,    0,  557,    0, 
            0,    0,    0,    0,    0,    0,  559,  560,  561,  562, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  515,  258,  258,  258,    0,    0,  258,  258,  258, 
            0,  258,    0,    0,    0,    0,    0,  625,    0,    0, 
            0,  258,  258,    0,    0,    0,    0,    0,    0,    0, 
          258,  258,  563,  258,  258,  258,  258,  258,    0,    0, 
            0,    0,    0,    0,    0,  258,    0,    0,    0,  518, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  518, 
            0,    0,    0,    0,    0,  625,    0,    0,    0,    0, 
            0,    0,  625,  625,    0,    0,    0,    0,  621,  258, 
            0,    0,  258,    0,    0,  258,    0,  258,    0,    0, 
            0,    0,    0,   95,    0,    0,  625,    0,    0,  625, 
            0,    0,    0,  258,    0,  635,  629,    0,  518,  636, 
          114,    0,  518,  188,  189,  258,  190,  191,  192,  193, 
          194,  195,  196,    0,    0,  197,  198,    0,    0,    0, 
          199,  200,  201,  202,    0,  258,    0,    0,    0,    0, 
          518,    0,    0,    0,    0,    0,    0,  204,  205,    0, 
          206,  207,  208,  209,  210,  211,  212,  213,  214,  215, 
          216,    0,    0,  217,    0,    0,    0,  618,  618,  618, 
            0,  618,  515,  515,  515,  618,  618,    0,  515,  515, 
          618,  515,  618,  618,  618,  618,  618,  618,  618,    0, 
          618,    0,    0,  618,  618,  618,  618,  618,  618,  618, 
          515,  515,  618,  515,  515,  515,  515,  515,    0,  618, 
            0,    0,  618,  618,  618,    0,  618,  618,  618,  618, 
          618,  618,  618,  618,  618,  618,  618,  515,  515,  515, 
          515,  515,  515,  515,  515,  515,  515,  515,  515,  515, 
          515,    0,    0,  515,  515,  515,  515,  618,  527,  111, 
          618,  618,  515,  618,  618,    0,  618,    0,  618,  515, 
          618,  515,  618,  515,  515,  515,  515,  515,  515,  515, 
          618,  515,  618,  515,    0,  618,  618,  618,  618,  618, 
          618,    0,  881,    0,  618,  881,  618,  618,    0,  618, 
          618,  618,  618,    0,  893,    0,    0,    0,  111,    0, 
          129,    0,  618,  618,  618,  515,  619,  619,  619,    0, 
          619,  518,  518,  518,  619,  619,    0,  518,  518,  619, 
          518,  619,  619,  619,  619,  619,  619,  619,    0,  619, 
          515,    0,  619,  619,  619,  619,  619,  619,  619,  518, 
          518,  619,  518,  518,  518,  518,  518,    0,  619,  129, 
            0,  619,  619,  619,    0,  619,  619,  619,  619,  619, 
          619,  619,  619,  619,  619,  619,  518,  518,  518,  518, 
          518,  518,  518,  518,  518,  518,  518,  518,  518,  518, 
            0,  310,  518,  518,  518,  518,  619,  530,    0,  619, 
          619,  518,  619,  619,    0,  619,    0,  619,  518,  619, 
          518,  619,  518,  518,  518,  518,  518,  518,  518,  619, 
          518,  619,  518,    0,  619,  619,  619,  619,  619,  619, 
            0,    0,  625,  619,    0,  619,  619,    0,  619,  619, 
          619,  619,  625,    0,    0,    0,    0,    0,    0,    0, 
            0,  619,  619,  619,  518,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  106,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  625,    0,  125,    0,  625,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  893,    0,    0,  893,    0, 
          881,    0,  893,    0,    0,    0,  111,    0,    0,    0, 
            0,    0,    0,  625,    0,    0,    0,    0,    0,  618, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  129,    0,    0, 
            0,  621,    0,    0,    0,    0,    0,  515,    0,    0, 
            0,  621,    0,    0,  515,  515,    0,    0,    0,    0, 
          618,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          893,    0,  893,    0,  893,    0,    0,  893,    0,    0, 
            0,  618,    0,    0,    0,  105,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  310,    0, 
          621,    0,  124,  893,  621,  310,  310,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  621,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  621, 
          621,  621,    0,  621,  625,  625,  625,  621,  621,    0, 
          625,  625,  621,  625,  621,  621,  621,  621,  621,  621, 
          621,    0,  625,    0,    0,  621,  621,  621,  621,  621, 
          621,  621,  625,  625,  621,  625,  625,  625,  625,  625, 
            0,  621,    0,    0,  621,  621,  621,    0,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,    0,    0,  625,  625,  625,  625,  621, 
            0,  114,  621,  621,  625,  621,  621,    0,  621,    0, 
          621,  625,  621,  625,  621,  625,  625,  625,  625,  625, 
          625,  625,  621,  625,  625,  625,    0,  621,  621,  621, 
          621,  621,  621, 
      };
   }

   private static final short[] yyTable2() {
      return new short[] {

            0,    0,    0,  621,    0,  621,  621,    0,  621,  621, 
          621,  621,    0,    0,    0,    0,    0,  114,    0,  128, 
            0,  621,  621,  621,  625,  621,  621,  621,    0,  621, 
          621,  621,  621,  621,  621,    0,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,    0,  621,  518, 
            0,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,    0,  621,  128,    0, 
          621,  621,  621,    0,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,    0, 
          321,  621,  621,  621,  621,  621,    0,    0,  621,  621, 
          621,  621,  621,    0,  621,    0,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,    0,  621,  621,  621,  621,  621,  621,    0, 
            0,  622,  621,    0,  621,  621,    0,  621,  621,  621, 
          621,  622,    0,    0,    0,    0,    0,    0,    0,    0, 
          621,  621,  621,  621,    0,    0,  620,  621,    0,    0, 
          622,    0,    0,    0,  188,  189,    0,  190,  191,  192, 
          193,  194,  195,  196,    0,  107,  197,  198,    0,    0, 
            0,  199,  200,  201,  202,    0,    0,    0,    0,    0, 
          622,  304,  126,    0,  622,    0,    0,    0,  204,  205, 
            0,  206,  207,  208,  209,  210,  211,  212,  213,  214, 
          215,  216,    0,    0,  217,  114,    0,    0,    0,    0, 
            0,    0,  622,    0,    0,    0,    0,    0,  619,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  128,    0,    0,    0, 
          625,    0,    0,    0,    0,    0,  518,    0,    0,    0, 
          625,    0,    0,  518,  518,    0,    0,    0,    0,  619, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          619,    0,    0,    0,  625,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  321,    0,  625, 
            0,  125,    0,  625,  321,  321,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  625,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  622,  622, 
          622,    0,  622,  622,  622,  622,  622,  622,    0,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
            0,  622,    0,    0,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,    0, 
          622,    0,    0,  622,  622,  622,    0,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,    0,    0,  622,  622,  622,  622,  622,    0, 
          125,  622,  622,  622,  622,  622,    0,  622,    0,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,    0,  622,  622,  622,  622, 
          622,  622,    0,    0,    0,  622,    0,  622,  622,    0, 
          622,  622,  622,  622,    0,    0,    0,    0,    0,  125, 
            0,    0,    0,  622,  622,  622,  622,  621,  621,  621, 
            0,  621,  625,  625,  625,  621,  621,    0,  625,  625, 
          621,  625,  621,  621,  621,  621,  621,  621,  621,    0, 
          625,  625,    0,  621,  621,  621,  621,  621,  621,  621, 
          625,  625,  621,  625,  625,  625,  625,  625,    0,  621, 
            0,    0,  621,  621,  621,    0,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,    0,    0,  625,  625,  625,  625,  621,    0,    0, 
          621,  621,  625,  621,  621,    0,  621,    0,  621,  625, 
          621,  625,  621,  625,  625,  625,  625,  625,  625,  625, 
          621,  625,  625,  625,    0,  621,  621,  621,  621,  621, 
          621,    0,    0,  621,  621,    0,  621,  621,    0,  621, 
          621,  621,  621,  621,    0,    0,    0,    0,    0,    0, 
            0,    0,  621,  621,  621,  625,    0,    0,  628,  629, 
            0,    0,  630,    0,    0,    0,  188,  189,    0,  190, 
          191,  192,  193,  194,  195,  196,    0,  621,  197,  198, 
            0,    0,    0,  199,  200,  201,  202,    0,    0,    0, 
            0,    0,  621,  304,  124,    0,  621,    0,    0,    0, 
          204,  205,    0,  206,  207,  208,  209,  210,  211,  212, 
          213,  214,  215,  216,    0,    0,  217,  125,    0,    0, 
            0,    0,    0,    0,  621,    0,    0,    0,    0,    0, 
          625,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  622,    0,    0,    0,    0,    0,  625,    0, 
            0,    0,  622,    0,    0,  625,  625,    0,    0,    0, 
            0,  621,    0,    0,  669,  621,    0,    0,  670,    0, 
            0,    0,  188,  189,    0,  190,  191,  192,  193,  194, 
          195,  196,  625,    0,  197,  198,  622,    0,    0,  199, 
          200,  201,  202,    0,    0,    0,    0,    0,    0,  304, 
            0,  622,    0,  126,    0,  622,  204,  205,    0,  206, 
          207,  208,  209,  210,  211,  212,  213,  214,  215,  216, 
            0,    0,  217,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  622,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          621,  621,  621,    0,  621,  621,  621,  621,  621,  621, 
            0,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,    0,  621,    0,    0,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,    0,  621,    0,    0,  621,  621,  621,    0,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,    0,    0,  621,  621,  621,  621, 
          621,    0,  124,  621,  621,  621,  621,  621,    0,  621, 
            0,  621,  621,  621,  621,  621,  621,  621,  621,  621, 
          621,  621,  621,  621,  621,  621,  621,    0,  621,  621, 
          621,  621,  621,  621,    0,    0,    0,  621,    0,  621, 
          621,    0,  621,  621,  621,  621,    0,    0,    0,    0, 
            0,  124,    0,    0,    0,  621,  621,  621,  621,  622, 
          622,  622,    0,  622,  622,  622,  622,  622,  622,    0, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,    0,  622,  621,    0,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
            0,  622,    0,    0,  622,  622,  622,    0,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,    0,    0,  622,  622,  622,  622,  622, 
            0,    0,  622,  622,  622,  622,  622,    0,  622,    0, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  622,  622,  622,  622,    0,  622,  622,  622, 
          622,  622,  622,    0,    0,  320,  622,    0,  622,  622, 
            0,  622,  622,  622,  622,  320,    0,    0,  101,    0, 
            0,    0,    0,    0,  622,  622,  622,  622,    0,    0, 
          635,  629,    0,    0,  636,    0,    0,    0,  188,  189, 
            0,  190,  191,  192,  193,  194,  195,  196,    0,  108, 
          197,  198,  101,    0,    0,  199,  200,  201,  202,    0, 
            0,    0,    0,    0,  320,  304,  127,    0,  320,  101, 
            0,    0,  204,  205,    0,  206,  207,  208,  209,  210, 
          211,  212,  213,  214,  215,  216,    0,    0,  217,  124, 
            0,    0,    0,    0,    0,    0,  320,    0,    0,  524, 
            0,    0,  621,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  320,    0,    0,    0,    0,    0, 
          621,    0,    0,    0,  320,    0,    0,  621,  621,    0, 
            0,    0,    0,  621,    0,    0,  687,  621,    0,    0, 
          688,    0,    0,    0,  188,  189,    0,  190,  191,  192, 
          193,  194,  195,  196,  621,    0,  197,  198,  320,    0, 
            0,  199,  200,  201,  202,    0,    0,    0,    0,    0, 
            0,  304,    0,  320,    0,  127,    0,  320,  204,  205, 
            0,  206,  207,  208,  209,  210,  211,  212,  213,  214, 
          215,  216,    0,    0,  217,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  320,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  622,  622,  622,    0,  622,  320,  320,  320, 
          622,  622,    0,  320,  320,  622,  320,  622,  622,  622, 
          622,  622,  622,  622,    0,    0,    0,  101,  622,  622, 
          622,  622,  622,  622,  622,  320,  320,  622,  320,  320, 
          320,  320,  320,    0,  622,    0,    0,  622,  622,  622, 
            0,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  622,  320,  320,  320,  320,  320,  320,  320,  320, 
          320,  320,  320,  320,  320,  320,  524,    0,  320,  320, 
          320,  320,  622,  524,  524,  622,  622,  320,  622,  622, 
          101,  622,    0,  622,  320,  622,  320,  622,  320,  320, 
          320,  320,  320,  320,  320,  622,  320,  647,  320,    0, 
          622,  622,  622,  622,  622,  622,    0,  647,    0,  622, 
            0,  622,  622,    0,  622,  622,  622,  622,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  622,  622,  622, 
          320,  622,  622,  622,    0,  622,  320,  320,  320,  622, 
          622,    0,  320,  320,  622,  320,  622,  622,  622,  622, 
          622,  622,  622,    0,    0,    0,  647,  622,  622,  622, 
          622,  622,  622,  622,  320,  320,  622,  320,  320,  320, 
          320,  320,    0,  622,    0,    0,  622,  622,  622,    0, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          622,  320,  320,  320,  320,  320,  320,  320,  320,  320, 
          320,  320,  320,  320,  320,    0,    0,  320,  320,  320, 
          320,  622,    0,    0,  622,  622,  320,  622,  622,    0, 
          622,    0,  622,  320,  622,  320,  622,  320,  320,  320, 
          320,  320,  320,  320,  622,  320,  440,  320,    0,  622, 
          622,  622,  622,  622,  622,    0,  440,    0,  622,    0, 
          622,  622,    0,  622,  622,  622,  622,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  622,  622,  622,  320, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          440,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  440,  440,    0,  123,    0,  440, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  440,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  126,    0,    0,    0, 
            0,    0,    0,    0,  287,  287,  287,    0,  287,  647, 
          647,  647,  287,  287,  647,  647,  647,  287,  647,  287, 
          287,  287,  287,  287,  287,  287,    0,  647,  647,    0, 
          287,  287,  287,  287,  287,  287,  287,  647,  647,  287, 
          647,  647,  647,  647,  647,  126,  287,    0,    0,  287, 
          287,  287,    0,  287,  287,  287,  287,  287,  287,  287, 
          287,  287,  287,  287,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  647,    0,  622,    0,    0, 
            0,    0,  647,  647,  287,    0,    0,  287,    0,  647, 
          287,  287,    0,  287,    0,  287,    0,  287,    0,  287, 
            0,  439,    0,    0,    0,    0,    0,  287,    0,  647, 
          647,  439,  287,  287,  287,  287,  287,  287,    0,    0, 
            0,  287,    0,  287,  287,    0,  287,  287,  287,  287, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  287, 
          287,  287,  647,    0,    0,  439,    0,    0,  440,  440, 
          440,    0,    0,  440,  440,  440,    0,  440,    0,  439, 
          439,    0,    0,    0,  439,    0,  440,  440,  440,    0, 
            0,    0,    0,    0,    0,    0,  440,  440,    0,  440, 
          440,  440,  440,  440,    0,    0,    0,    0,    0,    0, 
            0,  440,  439,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  440,  440,  440,  440,  440,  440,  440, 
          440,  440,  440,  440,  440,  440,  440,    0,    0,  440, 
          440,  440,  440,    0,  521,  440,    0,    0,  440,    0, 
            0,  440,    0,  440,  521,  440,    0,  440,    0,  440, 
          440,  440,  440,  440,  440,  440,    0,  440,  440,  440, 
            0,    0,    0,  126,    0,    0,    0,    0,    0,    0, 
            0,  440,    0,    0,    0,    0,  622,    0,  521,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  440,  521,  521,    0,  117,    0,  521,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  622,  521,    0,    0,    0,    0, 
            0,  622,  622,    0,    0,    0,    0,  622,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  520,    0,    0,    0,    0,    0,    0,  622,    0, 
            0,  520,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  439,  439,  439,    0,    0,  439,  439, 
          439,    0,  439,    0,    0,    0,    0,    0,    0,    0, 
            0,  439,  439,  439,    0,  520,    0,    0,    0,    0, 
            0,  439,  439,    0,  439,  439,  439,  439,  439,  520, 
          520,    0,  116,    0,  520,    0,  439,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  439,  439, 
          439,  439,  439,  439,  439,  439,  439,  439,  439,  439, 
          439,  439,  520,    0,  439,  439,  439,  439,    0,    0, 
          439,    0,    0,  439,    0,    0,  439,    0,  439,    0, 
          439,  103,  439,    0,  439,  439,  439,  439,  439,  439, 
          439,    0,  439,  439,  439,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  439,    0,    0,    0, 
            0,    0,    0,    0,    0,  103,  521,  521,  521,    0, 
            0,  521,  521,  521,    0,  521,  439,    0,    0,    0, 
            0,    0,  103,    0,    0,  521,  521,    0,    0,    0, 
            0,    0,    0,    0,  521,  521,    0,  521,  521,  521, 
          521,  521,    0,    0,    0,    0,    0,    0,    0,  521, 
          522,    0,  526,    0,    0,    0,    0,    0,    0,    0, 
          522,  521,  521,  521,  521,  521,  521,  521,  521,  521, 
          521,  521,  521,  521,  521,    0,    0,  521,  521,  521, 
          521,    0,  533,  521,    0,    0,  521,    0,    0,  521, 
            0,  521,    0,  521,  522,  521,    0,  521,  521,  521, 
          521,  521,  521,  521,    0,  521,    0,  521,  522,  522, 
            0,  118,    0,  522,    0,    0,    0,    0,    0,  521, 
            0,    0,    0,  520,  520,  520,    0,    0,  520,  520, 
          520,    0,  520,    0,    0,    0,    0,    0,    0,  521, 
            0,  522,  520,  520,    0,    0,    0,    0,    0,    0, 
            0,  520,  520,    0,  520,  520,  520,  520,  520,    0, 
            0,    0,    0,    0,    0,    0,  520,  523,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  523,  520,  520, 
          520,  520,  520,  520,  520,  520,  520,  520,  520,  520, 
          520,  520,    0,    0,  520,  520,  520,  520,    0,  532, 
          520,    0,    0,  520,    0,    0,  520,    0,  520,    0, 
          520,  523,  520,    0,  520,  520,  520,  520,  520,  520, 
          520,    0,  520,    0,  520,  523,  523,    0,  119,    0, 
          523,    0,    0,    0,    0,    0,  520,    0,    0,    0, 
          103,    0,   94,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  520,    0,  523,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,   94,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  526, 
          525,    0,    0,   94,    0,    0,  526,  526,    0,    0, 
          525,    0,    0,  103,    0,    0,    0,    0,    0,    0, 
            0,    0,  522,  522,  522,    0,    0,  522,  522,  522, 
            0,  522,    0,  517,    0,    0,    0,    0,    0,    0, 
            0,  522,  522,    0,  525,    0,    0,    0,    0,    0, 
          522,  522,    0,  522,  522,  522,  522,  522,  525,  525, 
            0,  121,    0,  525,    0,  522,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  522,  522,  522, 
          522,  522,  522,  522,  522,  522,  522,  522,  522,  522, 
          522,  525,    0,  522,  522,  522,  522,    0,  534,  522, 
            0,    0,  522,   93,    0,  522,    0,  522,    0,  522, 
            0,  522,    0,  522,  522,  522,  522,  522,  522,  522, 
           57,  522,    0,  522,    0,    0,    0,    0,    0,    0, 
           57,    0,    0,    0,    0,  522,    0,   93,    0,  523, 
          523,  523,    0,    0,  523,  523,  523,    0,  523,    0, 
            0,    0,    0,    0,   93,  522,    0,    0,  523,  523, 
            0,    0,    0,    0,    0,    0,    0,  523,  523,    0, 
          523,  523,  523,  523,  523,    0,    0,    0,    0,   57, 
            0,    0,  523,  524,  516,    0,    0,    0,    0,    0, 
            0,    0,    0,  524,  523,  523,  523,  523,  523,  523, 
          523,  523,  523,  523,  523,  523,  523,  523,    0,    0, 
          523,  523,  523,  523,    0,  535,  523,    0,    0,  523, 
            0,   94,  523,    0,  523,    0,  523,  524,  523,    0, 
          523,  523,  523,  523,  523,  523,  523,    0,  523,    0, 
          523,  524,  524,    0,  120,    0,  524,    0,    0,    0, 
            0,    0,  523,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  525,  525,  525,    0,    0,  525,  525,  525, 
          517,  525,  523,    0,  524,    0,    0,  517,  517,    0, 
            0,  525,  525,    0,   94,    0,    0,    0,    0,    0, 
          525,  525,    0,  525,  525,  525,  525,  525,    0,    0, 
            0,    0,    0,    0,    0,  525,  526,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  526,  525,  525,  525, 
          525,  525,  525,  525,  525,  525,  525,  525,  525,  525, 
          525,    0,    0,  525,  525,  525,  525,    0,  537,  525, 
            0,    0,  525,    0,    0,  525,    0,  525,    0,  525, 
          526,  525,    0,  525,  525,  525,  525,  525,  525,  525, 
            0,  525,    0,  525,  526,  526,    0,  122,    0,  526, 
            0,    0,   93,    0,    0,  525,    0,    0,    0,    0, 
            0,    0,   57,   57,   57,    0,    0,   57,   57,   57, 
            0,   57,    0,    0,    0,  525,    0,  526,    0,    0, 
            0,   57,    0,    0,    0,    0,    0,    0,    0,    0, 
           57,   57,    0,   57,   57,   57,   57,   57,    0,    0, 
            0,  516,    0,  517,    0,    0,    0,    0,  516,  516, 
            0,    0,    0,  517,    0,   93,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  524,  524,  524,    0,    0, 
          524,  524,  524,    0,  524,    0,    0,    0,    0,    0, 
            0,    0,   57,    0,  524,  524,    0,  517,    0,    0, 
            0,    0,    0,  524,  524,    0,  524,  524,  524,  524, 
          524,  517,  517,   57,  113,    0,  517,    0,  524,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          524,  524,  524,  524,  524,  524,  524,  524,  524,  524, 
          524,  524,  524,  524,  517,   57,  524,  524,  524,  524, 
            0,  536,  524,    0,    0,  524,    0,    0,  524,    0, 
          524,    0,  524,   96,  524,    0,  524,  524,  524,  524, 
          524,  524,  524,    0,  524,    0,  524,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  524,    0, 
            0,    0,    0,    0,    0,    0,    0,   96,  526,  526, 
          526,    0,    0,  526,  526,  526,    0,  526,  524,    0, 
            0,    0,    0,    0,   96,    0,    0,  526,  526,    0, 
            0,    0,    0,    0,    0,    0,  526,  526,    0,  526, 
          526,  526,  526,  526,    0,    0,    0,    0,    0,    0, 
            0,  526,  516,    0,  519,    0,    0,    0,    0,    0, 
            0,    0,  516,  526,  526,  526,  526,  526,  526,  526, 
          526,  526,  526,  526,  526,  526,  526,    0,    0,  526, 
          526,  526,  526,    0,  538,  526,    0,    0,  526,    0, 
            0,  526,    0,  526,    0,  526,  516,  526,    0,  526, 
          526,  526,  526,  526,  526,  526,    0,  526,    0,  526, 
          516,  516,    0,  112,    0,  516,    0,    0,    0,    0, 
            0,  526,    0,    0,    0,  517,  517,  517,    0,    0, 
          517,  517,  517,    0,  517,    0,    0,    0,    0,    0, 
            0,  526,    0,  516,  517,  517,    0,    0,    0,    0, 
            0,    0,    0,  517,  517,    0,  517,  517,  517,  517, 
          517,    0,    0,    0,    0,    0,    0,    0,  517,  519, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  519, 
          517,  517,  517,  517,  517,  517,  517,  517,  517,  517, 
          517,  517,  517,  517,    0,    0,  517,  517,  517,  517, 
            0,  529,  517,    0,    0,  517,    0,    0,  517,    0, 
          517,    0,  517,  519,  517,    0,  517,  517,  517,  517, 
          517,  517,  517,    0,  517,    0,  517,  519,  519,    0, 
          115,    0,  519,    0,    0,    0,    0,    0,  517,    0, 
            0,    0,   96,    0,  110,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  517,    0, 
          519,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  110,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  519,  647,    0,    0,  110,    0,    0,  519,  519, 
            0,    0,  647,    0,    0,   96,    0,    0,    0,    0, 
            0,    0,    0,    0,  516,  516,  516,    0,    0,  516, 
          516,  516,    0,  516,    0,  310,    0,    0,    0,    0, 
            0,    0,    0,  516,  516,    0,  647,    0,    0,    0, 
            0,    0,  516,  516,    0,  516,  516,  516,  516,  516, 
          647,  647,    0,    0,    0,  647,    0,  516,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  516, 
          516,  516,  516,  516,  516,  516,  516,  516,  516,  516, 
          516,  516,  516,  647,    0,  516,  516,  516,  516,    0, 
          528,  516,    0,    0,  516,  109,    0,  516,    0,  516, 
            0,  516,    0,  516,    0,  516,  516,  516,  516,  516, 
          516,  516,   58,  516,    0,  516,    0,    0,    0,    0, 
            0,    0,   58,    0,    0,    0,    0,  516,    0,  109, 
            0,  519,  519,  519,    0,    0,  519,  519,  519,    0, 
          519,    0,    0,    0,    0,    0,  109,  516,    0,    0, 
          519,  519,    0,    0,    0,    0,    0,    0,    0,  519, 
          519,    0,  519,  519,  519,  519,  519,    0,    0,    0, 
            0,   58,    0,    0,  519,  310,  321,    0,    0,    0, 
            0,    0,    0,    0,    0,  310,  519,  519,  519,  519, 
          519,  519,  519,  519,  519,  519,  519,  519,  519,  519, 
            0,    0,  519,  519,  519,  519,    0,  531,  519,    0, 
            0,  519,    0,  110,  519,    0,  519,    0,  519,  310, 
          519,    0,  519,  519,  519,  519,  519,  519,  519,    0, 
          519,    0,  519,  310,  310,    0,  129,    0,  310,    0, 
            0,    0,    0,    0,  519,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  647,  647,  647,    0,    0,  647, 
          647,  647,  310,  647,  519,    0,  310,    0,    0,  310, 
          310,    0,  647,  647,  647,    0,  110,    0,    0,    0, 
            0,    0,  647,  647,    0,  647,  647,  647,  647,  647, 
            0,    0,    0,    0,    0,    0,    0,  647,  321,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  321,  647, 
          647,  647,  647,  647,  647,  647,  647,  647,  647,  647, 
          647,  647,  647,    0,    0,  647,  647,  647,  647,    0, 
            0,  647,    0,    0,  647,    0,    0,  647,    0,  647, 
            0,  647,  321,  647,    0,  647,  647,  647,  647,  647, 
          647,  647,    0,  647,  647,  647,  321,  321,    0,  128, 
            0,  321,    0,    0,  109,    0,    0,  647,    0,    0, 
            0,    0,    0,    0,   58,   58,   58,    0,    0,   58, 
           58,   58,    0,   58,    0,    0,    0,  647,    0,  321, 
            0,    0,    0,   58,    0,    0,    0,    0,    0,    0, 
            0,    0,   58,   58,    0,   58,   58,   58,   58,   58, 
            0,    0,    0,  321,    0,  463,    0,    0,    0,    0, 
          321,  321,    0,    0,    0,  463,    0,  109,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  310,  310,  310, 
            0,    0,  310,  310,  310,    0,  310,    0,    0,    0, 
            0,    0,    0,    0,   58,    0,  310,  310,    0,  463, 
            0,    0,    0,    0,    0,  310,  310,    0,  310,  310, 
          310,  310,  310,  463,  463,   58,    0,    0,  463,    0, 
          310,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  310,  310,  310,  310,  310,  310,  310,  310, 
          310,  310,  310,  310,  310,  310,  463,   58,  310,  310, 
          310,  310,    0,    0,  310,    0,    0,  310,    0,    0, 
          310,    0,  310,    0,  310,    0,  310,    0,  310,  310, 
          310,  310,  310,  310,  310,    0,  310,    0,  310,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          310,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          321,  321,  321,    0,    0,  321,  321,  321,    0,  321, 
          310,    0,    0,    0,    0,    0,    0,    0,    0,  321, 
          321,    0,    0,    0,    0,    0,    0,    0,  321,  321, 
            0,  321,  321,  321,  321,  321,    0,    0,    0,    0, 
            0,    0,    0,  321,  332,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  332,  321,  321,  321,  321,  321, 
          321,  321,  321,  321,  321,  321,  321,  321,  321,    0, 
            0,  321,  321,  321,  321,    0,    0,  321,    0,    0, 
          321,    0,    0,  321,    0,  321,    0,  321,  332,  321, 
            0,  321,  321,  321,  321,  321,  321,  321,    0,  321, 
            0,  321,  332,  332,    0,    0,    0,  332,    0,    0, 
            0,    0,    0,  321,    0,    0,    0,  463,  463,  463, 
            0,    0,  463,  463,  463,    0,  463,    0,    0,    0, 
            0,    0,    0,  321,    0,  332,  463,  463,    0,    0, 
            0,    0,    0,    0,    0,  463,  463,    0,  463,  463, 
          463,  463,  463,    0,    0,    0,    0,    0,    0,    0, 
          463,  260,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  260,  463,  463,  463,  463,  463,  463,  463,  463, 
          463,  463,  463,  463,  463,  463,    0,    0,  463,  463, 
          463,  463,    0,    0,  463,    0,    0,  463,    0,    0, 
          463,    0,  463,    0,  463,  260,  463,    0,  463,  463, 
          463,  463,  463,  463,  463,    0,  463,    0,  463,  260, 
          260,    0,    0,    0,  260,    0,    0,    0,  690,  629, 
          463,    0,  691,    0,    0,    0,  188,  189,    0,  190, 
          191,  192,  193,  194,  195,  196,    0,    0,  197,  198, 
          463,    0,  363,  199,  200,  201,  202,    0,    0,    0, 
            0,    0,    0,  304,    0,    0,    0,    0,    0,    0, 
          204,  205,    0,  206,  207,  208,  209,  210,  211,  212, 
          213,  214,  215,  216,  327,    0,  217,    0,    0,    0, 
            0,    0,    0,    0,  327,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  332,  332,  332,    0, 
            0,  332,  332,  332,    0,  332,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  332,  332,    0,  327,    0, 
            0,    0,    0,    0,  332,  332,    0,  332,  332,  332, 
          332,  332,  327,  327,    0,    0,    0,  327,    0,  332, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  332,  332,  332,  332,  332,  332,  332,  332,  332, 
          332,  332,  332,  332,  332,  327,    0,  332,  332,  332, 
          332,    0,    0,  332,    0,    0,  332,    0,    0,  332, 
            0,  332,    0,  332,    0,  332,    0,  332,  332,  332, 
          332,  332,  332,  332,  298,  332,    0,  332,    0,    0, 
            0,    0,    0,    0,  298,    0,    0,    0,    0,  332, 
            0,    0,    0,  260,  260,  260,    0,    0,  260,  260, 
          260,    0,  260,    0,    0,    0,    0,    0,    0,  332, 
            0,    0,  260,  260,    0,    0,    0,    0,  294,    0, 
            0,  260,  260,    0,  260,  260,  260,  260,  260,    0, 
            0,    0,    0,  298,    0,    0,  260,  510,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  510,  260,  260, 
          260,  260,  260,  260,  260,  260,  260,  260,  260,  363, 
          260,  260,    0,    0,  260,  260,  363,  363,    0,    0, 
          260,    0,    0,  260,    0,    0,  260,    0,  260,    0, 
          260,  510,  260,    0,  260,  260,  260,  260,  260,  260, 
          260,    0,  260,    0,  260,  510,  510,    0,    0,    0, 
          510,    0,    0,    0,    0,    0,  260,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  327,  327,  327,    0, 
            0,  327,  327,  327,    0,  327,  260,    0,  510,    0, 
            0,    0,    0,    0,    0,  327,  327,    0,    0,    0, 
            0,    0,    0,    0,  327,  327,    0,  327,  327,  327, 
          327,  327,    0,    0,    0,    0,    0,    0,    0,  327, 
          359,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          359,  327,  327,  327,  327,  327,  327,  327,  327,  327, 
          327,  327,  327,  327,  327,    0,    0,  327,  327,  327, 
          327,    0,    0,  327,    0,    0,  327,    0,    0,  327, 
            0,  327,    0,  327,  359,  327,    0,  327,  327,  327, 
          327,  327,  327,  327,    0,  327,    0,  327,    0,  359, 
            0,    0,    0,  359,    0,    0,    0,    0,    0,  327, 
            0,    0,    0,    0,    0,    0,  298,  298,  298,    0, 
            0,  298,  298,  298,    0,  298,    0,    0,    0,  327, 
            0,  359,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  360,    0,    0,    0,  298,  298,  298, 
          298,  298,    0,  360,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  510, 
          510,  510,    0,    0,  510,  510,  510,  360,  510,    0, 
            0,    0,    0,  298,    0,    0,  298,    0,  510,  510, 
            0,    0,  360,    0,    0,    0,  360,  510,  510,    0, 
          510,  510,  510,  510,  510,    0,    0,  298,    0,    0, 
            0,    0,  510,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  360,  510,  510,  510,  510,  510, 
          510,  510,  510,  510,  510,  510,  510,  510,    0,  298, 
          510,  510,  510,  510,    0,  324,  510,    0,    0,  510, 
            0,    0,  510,    0,  510,  324,  510,    0,  510,    0, 
          510,  510,  510,  510,  510,  510,  510,    0,  510,    0, 
          510,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  510,    0,    0,    0,    0,    0,    0,  324, 
            0,    0,  359,  359,  359,    0,    0,  359,  359,  359, 
            0,  359,  510,    0,  324,    0,    0,    0,  324,    0, 
            0,  359,  359,    0,    0,    0,    0,    0,    0,    0, 
          359,  359,    0,  359,  359,  359,  359,  359,    0,    0, 
            0,    0,    0,    0,    0,    0,  324,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  359,  359,  359, 
          359,  359,  359,  359,  359,  359,  359,  359,  359,  359, 
          359,    0,    0,  359,  359,  359,  359,    0,    0,  359, 
            0,    0,  359,    0,    0,  359,    0,  359,    0,  359, 
            0,  359,   45,  359,  359,  359,  359,  359,  359,  359, 
            0,  359,   45,  359,    0,  360,  360,  360,    0,    0, 
          360,  360,  360,    0,  360,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  360,  360,    0,    0,    0,    0, 
            0,    0,    0,  360,  360,  359,  360,  360,  360,  360, 
          360,    0,    0,    0,    0,    0,    0,    0,    0,  241, 
            0,   45,    0,    0,    0,    0,    0,    0,    0,  241, 
          360,  360,  360,  360,  360,  360,  360,  360,  360,  360, 
          360,  360,  360,  360,    0,    0,  360,  360,  360,  360, 
            0,    0,  360,    0,    0,  360,    0,    0,  360,    0, 
          360,    0,  360,  241,  360,    0,  360,  360,  360,  360, 
          360,  360,  360,    0,  360,    0,  360,  241,  241,    0, 
            0,    0,  241,    0,    0,    0,    0,  324,  324,  324, 
            0,    0,  324,  324,  324,    0,  324,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  324,  324,  360,    0, 
            0,    0,    0,    0,    0,  324,  324,    0,  324,  324, 
          324,  324,  324,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  324,  324,  324,  324,  324,  324,  324,  324, 
          324,  324,  324,  324,  324,  324,  238,    0,  324,  324, 
          324,  324,    0,    0,  324,    0,  238,  324,    0,    0, 
          324,    0,  324,    0,  324,    0,  324,    0,  324,  324, 
          324,  324,  324,  324,  324,    0,  324,    0,  324,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          238,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  238,  238,    0,    0,    0,  238, 
          324,    0,    0,    0,   45,   45,   45,    0,    0,    0, 
           45,   45,    0,   45,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   45,   45,   45,   45,  235, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  235, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  241,  241,  241,    0,    0,  241,  241,  241,    0, 
          241,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          241,  241,    0,  235,   45,    0,    0,    0,    0,  241, 
          241,    0,  241,  241,  241,  241,  241,  235,  235,    0, 
            0,    0,  235,    0,  241,   45,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  241,  241,  241, 
          241,  241,  241,  241,  241,  241,  241,    0,  241,  241, 
            0,    0,  241,  241,    0,    0,    0,   45,  241,    0, 
            0,  241,    0,    0,  241,    0,  241,    0,  241,    0, 
          241,    0,  241,  241,  241,  241,  241,  241,  241,    0, 
          241,    0,  241,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  241,    0,  237,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  237,    0,    0,    0, 
            0,    0,    0,    0,  241,    0,    0,    0,  238,  238, 
          238,    0,    0,  238,  238,  238,    0,  238,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  238,  238,    0, 
          237,    0,    0,    0,    0,    0,  238,  238,    0,  238, 
          238,  238,  238,  238,  237,  237,    0,    0,    0,  237, 
            0,  238,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  238,  238,  238,  238,  238,  238, 
          238,  238,  238,  238,    0,  238,  238,    0,    0,  238, 
          238,    0,    0,    0,    0,  238,    0,    0,  238,    0, 
            0,  238,    0,  238,    0,  238,    0,  238,    0,  238, 
          238,  238,  238,  238,  238,  238,    0,  238,    0,  238, 
            0,  235,  235,  235,    0,    0,  235,  235,  235,    0, 
          235,  238,  236,    0,    0,    0,    0,    0,    0,    0, 
          235,  235,  236,    0,    0,    0,    0,    0,    0,  235, 
          235,  238,  235,  235,  235,  235,  235,    0,    0,    0, 
            0,    0,    0,    0,  235,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  236,  235,  235,  235, 
          235,  235,  235,  235,  235,  235,  235,    0,  235,  235, 
          236,  236,  235,  235,    0,  236,    0,    0,  235,    0, 
            0,  235,    0,    0,  235,    0,  235,    0,  235,    0, 
          235,    0,  235,  235,  235,  235,  235,  235,  235,    0, 
          235,    0,  235,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  235,    0,    0,    0,    0,    0, 
            0,  239,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  239,    0,    0,  235,    0,    0,    0,  237,  237, 
          237,    0,    0,  237,  237,  237,    0,  237,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  237,  237,    0, 
            0,    0,    0,    0,    0,  239,  237,  237,    0,  237, 
          237,  237,  237,  237,    0,    0,    0,    0,    0,  239, 
          239,  237,    0,    0,  239,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  237,  237,  237,  237,  237,  237, 
          237,  237,  237,  237,    0,  237,  237,    0,    0,  237, 
          237,    0,    0,    0,    0,  237,    0,    0,  237,    0, 
            0,  237,    0,  237,    0,  237,    0,  237,    0,  237, 
          237,  237,  237,  237,  237,  237,    0,  237,  440,  237, 
            0,    0,    0,    0,    0,    0,    0,    0,  440,    0, 
            0,  237,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  237,    0,    0,  236,  236,  236,    0,    0,  236, 
          236,  236,  104,  236,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  236,  236,    0,    0,  440,    0,  123, 
            0,  440,  236,  236,    0,  236,  236,  236,  236,  236, 
            0,    0,    0,   36,    0,    0,    0,  236,    0,    0, 
            0,    0,    0,   36,    0,    0,    0,    0,    0,  440, 
          236,  236,  236,  236,  236,  236,  236,  236,  236,  236, 
            0,  236,  236,    0,    0,  236,  236,    0,    0,    0, 
            0,  236,    0,    0,  236,    0,    0,  236,    0,  236, 
            0,  236,    0,  236,    0,  236,  236,  236,  236,  236, 
          236,  236,   36,  236,    0,  236,    0,    0,    0,    0, 
            0,    0,    0,  239,  239,  239,    0,  236,  239,  239, 
          239,    0,  239,    0,  272,    0,    0,    0,    0,    0, 
            0,    0,  239,  239,  272,    0,    0,  236,    0,    0, 
            0,  239,  239,    0,  239,  239,  239,  239,  239,    0, 
            0,    0,    0,    0,    0,    0,  239,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  267,  239, 
          239,  239,  239,  239,  239,  239,  239,  239,  239,  440, 
          239,  239,    0,  272,  239,  239,    0,    0,    0,  440, 
          239,    0,    0,  239,    0,    0,  239,    0,  239,    0, 
          239,    0,  239,    0,  239,  239,  239,  239,  239,  239, 
          239,    0,  239,    0,  239,    0,    0,    0,    0,    0, 
            0,    0,    0,  440,    0,    0,  239,    0,    0,    0, 
          440,  440,  440,    0,    0,    0,  440,  440,  440,  440, 
          123,    0,  440,    0,    0,    0,  239,    0,  440,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  440,  440, 
            0,  440,  440,  440,  440,  440,    0,    0,    0,    0, 
          440,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  440,  440,  440,  440,  440, 
          440,  440,  440,  440,  440,  440,  440,  440,  440,  521, 
            0,  440,  440,  440,  440,   36,   36,   36,    0,  521, 
          440,   36,   36,    0,   36,    0,    0,  440,    0,  440, 
            0,  440,  440,  440,  440,  440,  440,  440,    0,  440, 
          440,  440,    0,    0,    0,    0,   36,   36,   36,   36, 
           36,    0,    0,   98,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  521,    0, 
          117,    0,  521,  440,    0,    0,    0,    0,    0,    0, 
            0,    0,    0, 
      };
   }

   private static final short[] yyTable3() {
      return new short[] {

            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,   36,  272,  272,  272,    0,  521,    0,  272, 
          272,    0,  272,    0,    0,    0,    0,    0,    0,  520, 
            0,    0,    0,   36,    0,    0,    0,    0,    0,  520, 
            0,  272,  272,    0,  272,  272,  272,  272,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   36,    0,    0,  440,  440, 
          440,    0,    0,   97,  440,  440,    0,  440,    0,    0, 
            0,    0,    0,    0,    0,    0,  440,    0,  520,    0, 
          116,    0,  520,  272,    0,    0,  440,  440,    0,  440, 
          440,  440,  440,  440,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  272,    0,    0,    0,    0,    0, 
          520,    0,    0,  440,  440,  440,  440,  440,  440,  440, 
          440,  440,  440,  440,  440,  440,  440,  522,    0,  440, 
          440,  440,  440,    0,    0,    0,  272,  522,  440,    0, 
            0,    0,    0,    0,    0,  440,    0,  440,    0,  440, 
          440,  440,  440,  440,  440,  440,    0,  440,  440,  440, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,   99,    0,    0,    0,    0,    0,    0,  521,  521, 
          521,    0,    0,    0,  521,  521,  522,  521,  118,    0, 
          522,  440,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  521,  521,    0,  521, 
          521,  521,  521,  521,    0,    0,    0,    0,  522,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  521,  521,  521,  521,  521,  521,  521, 
          521,  521,  521,  521,  521,  521,  521,    0,    0,  521, 
          521,  521,  521,    0,  533,    0,    0,    0,  521,    0, 
            0,    0,    0,    0,    0,  521,    0,  521,    0,  521, 
          521,  521,  521,  521,  521,  521,    0,  521,    0,  521, 
            0,  520,  520,  520,    0,    0,    0,  520,  520,    0, 
          520,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  520, 
          520,  521,  520,  520,  520,  520,  520,    0,    0,    0, 
          523,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          523,    0,    0,    0,    0,    0,  520,  520,  520,  520, 
          520,  520,  520,  520,  520,  520,  520,  520,  520,  520, 
            0,    0,  520,  520,  520,  520,    0,  532,    0,    0, 
            0,  520,    0,    0,  100,    0,    0,    0,  520,    0, 
          520,    0,  520,  520,  520,  520,  520,  520,  520,  523, 
          520,  119,  520,  523,    0,    0,    0,    0,    0,  522, 
          522,  522,    0,    0,    0,  522,  522,    0,  522,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  523,    0,    0,  520,    0,    0,  522,  522,    0, 
          522,  522,  522,  522,  522,    0,    0,    0,  525,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  525,    0, 
            0,    0,    0,    0,  522,  522,  522,  522,  522,  522, 
          522,  522,  522,  522,  522,  522,  522,  522,    0,    0, 
          522,  522,  522,  522,    0,  534,    0,    0,    0,  522, 
            0,    0,  102,    0,    0,    0,  522,    0,  522,    0, 
          522,  522,  522,  522,  522,  522,  522,  525,  522,  121, 
          522,  525,    0,    0,    0,    0,    0,    0,    0,    0, 
          700,  621,    0,    0,  701,    0,    0,    0,  188,  189, 
            0,  190,  191,  192,  193,  194,  195,  196,    0,  525, 
          197,  198,  522,    0,    0,  199,  200,  201,  202,    0, 
            0,  524,    0,    0,    0,  304,    0,    0,    0,    0, 
            0,  524,  204,  205,    0,  206,  207,  208,  209,  210, 
          211,  212,  213,  214,  215,  216,    0,    0,  217,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  101,    0,    0,    0,    0, 
            0,    0,  523,  523,  523,    0,    0,    0,  523,  523, 
          524,  523,  120,    0,  524,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          523,  523,    0,  523,  523,  523,  523,  523,    0,    0, 
            0,    0,  524,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  523,  523,  523, 
          523,  523,  523,  523,  523,  523,  523,  523,  523,  523, 
          523,  526,    0,  523,  523,  523,  523,    0,  535,    0, 
            0,  526,  523,    0,    0,    0,    0,    0,    0,  523, 
            0,  523,    0,  523,  523,  523,  523,  523,  523,  523, 
            0,  523,    0,  523,    0,    0,    0,    0,    0,    0, 
          525,  525,  525,    0,    0,  103,  525,  525,    0,  525, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          526,    0,  122,    0,  526,  523,    0,    0,  525,  525, 
            0,  525,  525,  525,  525,  525,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  526,    0,    0,  525,  525,  525,  525,  525, 
          525,  525,  525,  525,  525,  525,  525,  525,  525,    0, 
            0,  525,  525,  525,  525,    0,  537,    0,    0,    0, 
          525,    0,    0,    0,    0,    0,    0,  525,    0,  525, 
            0,  525,  525,  525,  525,  525,  525,  525,    0,  525, 
            0,  525,    0,  524,  524,  524,    0,    0,    0,  524, 
          524,    0,  524,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  524,  524,  525,  524,  524,  524,  524,  524,    0, 
            0,    0,  517,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  517,    0,    0,    0,    0,    0,  524,  524, 
          524,  524,  524,  524,  524,  524,  524,  524,  524,  524, 
          524,  524,    0,    0,  524,  524,  524,  524,    0,  536, 
            0,    0,    0,  524,    0,    0,   94,    0,    0,    0, 
          524,    0,  524,    0,  524,  524,  524,  524,  524,  524, 
          524,  517,  524,  113,  524,  517,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  526,  526,  526,    0,    0,    0,  526, 
          526,    0,  526,  517,    0,    0,  524,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  526,  526,    0,  526,  526,  526,  526,  526,    0, 
            0,    0,  516,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  516,    0,    0,    0,    0,    0,  526,  526, 
          526,  526,  526,  526,  526,  526,  526,  526,  526,  526, 
          526,  526,    0,    0,  526,  526,  526,  526,    0,  538, 
            0,    0,    0,  526,    0,    0,   93,    0,    0,    0, 
          526,    0,  526,    0,  526,  526,  526,  526,  526,  526, 
          526,  516,  526,  112,  526,  516,    0,    0,    0,    0, 
            0,    0,    0,    0,  703,  629,    0,    0,  704,    0, 
            0,    0,  188,  189,    0,  190,  191,  192,  193,  194, 
          195,  196,    0,  516,  197,  198,  526,    0,    0,  199, 
          200,  201,  202,    0,    0,  519,    0,    0,    0,  304, 
            0,    0,    0,    0,    0,  519,  204,  205,    0,  206, 
          207,  208,  209,  210,  211,  212,  213,  214,  215,  216, 
            0,    0,  217,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  517,  517,  517,    0,    0,   96, 
          517,  517,    0,  517,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  519,    0,  115,    0,  519,    0, 
            0,    0,  517,  517,    0,  517,  517,  517,  517,  517, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  519,    0,    0,  517, 
          517,  517,  517,  517,  517,  517,  517,  517,  517,  517, 
          517,  517,  517,    0,    0,  517,  517,  517,  517,    0, 
          529,    0,    0,    0,  517,    0,    0,    0,    0,    0, 
            0,  517,    0,  517,    0,  517,  517,  517,  517,  517, 
          517,  517,  246,  517,    0,  517,    0,    0,    0,    0, 
            0,    0,  246,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  516,  516,  516,    0,    0,    0, 
          516,  516,    0,  516,    0,    0,    0,  517,    0,    0, 
            0,    0,    0,    0,    0,    0,  246,    0,    0,    0, 
            0,    0,  516,  516,    0,  516,  516,  516,  516,  516, 
          246,  246,    0,    0,    0,  246,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  516, 
          516,  516,  516,  516,  516,  516,  516,  516,  516,  516, 
          516,  516,  516,    0,    0,  516,  516,  516,  516,    0, 
          528,    0,    0,    0,  516,    0,    0,    0,    0,    0, 
            0,  516,    0,  516,    0,  516,  516,  516,  516,  516, 
          516,  516,    0,  516,    0,  516,    0,  519,  519,  519, 
            0,    0,    0,  519,  519,    0,  519,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  519,  519,  516,  519,  519, 
          519,  519,  519,    0,    0,    0,  310,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  310,    0,    0,    0, 
            0,    0,  519,  519,  519,  519,  519,  519,  519,  519, 
          519,  519,  519,  519,  519,  519,    0,    0,  519,  519, 
          519,  519,    0,  531,    0,    0,    0,  519,    0,    0, 
          110,    0,    0,    0,  519,    0,  519,    0,  519,  519, 
          519,  519,  519,  519,  519,  310,  519,  129,  519,  310, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  299,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  299,    0,    0,    0,  310,    0,    0, 
          519,    0,    0,    0,  246,  246,  246,    0,    0,  246, 
          246,  246,    0,  246,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  246,  246,    0,    0,  295,    0,    0, 
            0,    0,  246,  246,    0,  246,  246,  246,  246,  246, 
            0,  321,  299,    0,    0,    0,    0,  246,    0,    0, 
            0,  321,    0,    0,    0,    0,    0,    0,    0,  246, 
          246,  246,  246,  246,    0,    0,  246,  246,  246,  246, 
            0,  246,  246,    0,    0,  246,  246,    0,    0,    0, 
            0,  246,    0,    0,  246,  109,    0,  246,    0,  246, 
            0,  246,    0,  246,    0,  246,  246,  246,  246,    0, 
          321,  246,  128,  246,  321,  246,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  246,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  321,    0,    0,    0,    0,  246,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  233,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  233,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  310,  310, 
          310,    0,    0,    0,  310,  310,    0,  310,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  233,    0,    0,    0,    0,  310,  310,    0,  310, 
          310,  310,  310,  310,    0,  233,  233,    0,    0,    0, 
          233,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  310,  310,  310,  310,  310,  310,  310, 
          310,  310,  310,  310,  310,  310,  310,    0,    0,  310, 
          310,  310,  310,    0,    0,  299,  299,  299,  310,    0, 
          299,  299,  299,    0,  299,  310,  234,  310,    0,  310, 
          310,  310,  310,  310,  310,  310,  234,  310,    0,  310, 
            0,    0,    0,    0,    0,    0,  299,  299,  299,  299, 
          299,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          234,  310,    0,  321,  321,  321,    0,    0,    0,  321, 
          321,    0,  321,    0,  234,  234,    0,    0,    0,  234, 
            0,    0,  299,    0,    0,  299,    0,    0,    0,    0, 
            0,  321,  321,    0,  321,  321,  321,  321,  321,    0, 
            0,    0,    0,    0,    0,    0,  299,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  321,  321, 
          321,  321,  321,  321,  321,  321,  321,  321,  321,  321, 
          321,  321,    0,    0,  321,  321,  321,  321,  299,    0, 
            0,    0,    0,  321,    0,    0,    0,    0,    0,    0, 
          321,    0,  321,    0,  321,  321,  321,  321,  321,  321, 
          321,    0,  321,  297,  321,    0,    0,    0,    0,  233, 
          233,  233,    0,  297,  233,  233,  233,    0,  233,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  233,  233, 
            0,    0,    0,    0,    0,    0,  321,  233,  233,    0, 
          233,  233,  233,  233,  233,    0,    0,  292,    0,    0, 
            0,    0,  233,  310,    0,    0,    0,    0,    0,    0, 
            0,    0,  297,  310,    0,  233,  233,  233,  233,  233, 
          233,  233,  233,  233,  233,    0,  233,  233,    0,    0, 
          233,  233,    0,    0,    0,    0,  233,    0,    0,  233, 
            0,    0,  233,    0,  233,    0,    0,  310,  233,    0, 
            0,    0,  233,  233,  233,  233,  233,    0,  233,    0, 
          233,    0,  310,    0,  129,    0,  310,    0,  234,  234, 
          234,    0,  233,  234,  234,  234,    0,  234,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  234,  234,    0, 
            0,    0,  233,    0,  310,    0,  234,  234,    0,  234, 
          234,  234,  234,  234,    0,    0,    0,    0,    0,    0, 
            0,  234,  254,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  254,    0,  234,  234,  234,  234,  234,  234, 
          234,  234,  234,  234,    0,  234,  234,    0,    0,  234, 
          234,    0,    0,    0,    0,  234,    0,    0,  234,    0, 
            0,  234,    0,  234,    0,    0,  254,  234,    0,    0, 
            0,  234,  234,  234,  234,  234,    0,  234,    0,  234, 
          254,  254,    0,    0,    0,  254,    0,    0,    0,    0, 
            0,  234,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  234,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  297,  297,  297,    0,    0, 
            0,  297,  297,    0,  297,    0,    0,    0,    0,  255, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  255, 
            0,    0,    0,    0,    0,    0,  297,  297,  297,  297, 
          297,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  310,  310,  310,    0,    0, 
            0,  310,  310,  255,  310,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  255,  255,    0, 
            0,    0,  255,  310,  310,  297,  310,  310,  310,  310, 
          310,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  297,    0,    0,    0, 
          310,  310,  310,  310,  310,  310,  310,  310,  310,  310, 
          310,  310,  310,  310,    0,    0,  310,  310,  310,  310, 
            0,    0,    0,    0,    0,  310,    0,    0,  297,    0, 
            0,    0,  310,    0,  310,    0,  310,  310,  310,  310, 
          310,  310,  310,    0,  310,    0,  310,    0,    0,    0, 
            0,    0,    0,    0,  254,  254,  254,    0,    0,  254, 
          254,  254,    0,  254,    0,  244,    0,    0,    0,    0, 
            0,    0,    0,  254,  254,  244,    0,    0,  310,    0, 
            0,    0,  254,  254,    0,  254,  254,  254,  254,  254, 
            0,    0,    0,    0,    0,    0,    0,  254,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  244, 
          254,  254,  254,  254,  254,  254,  254,  254,  254,  254, 
            0,  254,  254,  244,  244,  254,  254,    0,  244,    0, 
            0,  254,    0,    0,  254,    0,    0,  254,    0,  254, 
            0,    0,    0,  254,    0,    0,    0,    0,    0,  254, 
          254,  254,    0,  254,    0,  254,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  254,    0,    0, 
            0,  255,  255,  255,    0,    0,  255,  255,  255,    0, 
          255,    0,  242,    0,    0,    0,    0,  254,    0,    0, 
          255,  255,  242,    0,    0,    0,    0,    0,    0,  255, 
          255,    0,  255,  255,  255,  255,  255,    0,    0,    0, 
            0,    0,    0,    0,  255,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  242,  255,  255,  255, 
          255,  255,  255,  255,  255,  255,  255,    0,  255,  255, 
          242,  242,  255,  255,    0,  242,    0,    0,  255,    0, 
            0,  255,    0,    0,  255,    0,  255,    0,    0,    0, 
          255,    0,    0,    0,    0,    0,  255,  255,  255,    0, 
          255,    0,  255,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  255,    0,    0,    0,    0,    0, 
            0,  243,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  243,    0,    0,  255,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  244,  244,  244, 
            0,    0,  244,  244,  244,  243,  244,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  244,  244,    0,  243, 
          243,    0,    0,    0,  243,  244,  244,    0,  244,  244, 
          244,  244,  244,    0,    0,    0,    0,    0,    0,    0, 
          244,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  244,  244,  244,  244,  244,  244,  244, 
          244,  244,  244,    0,  244,  244,    0,    0,    0,    0, 
            0,    0,    0,    0,  244,    0,    0,  244,  265,    0, 
          244,    0,  244,    0,    0,    0,  244,    0,  265,    0, 
            0,    0,  244,  244,  244,    0,  244,    0,  244,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          244,    0,    0,    0,  242,  242,  242,    0,    0,  242, 
          242,  242,  265,  242,    0,    0,    0,    0,    0,    0, 
          244,    0,    0,  242,  242,    0,  265,  265,    0,    0, 
            0,  265,  242,  242,    0,  242,  242,  242,  242,  242, 
            0,    0,    0,    0,    0,    0,    0,  242,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          242,  242,  242,  242,  242,  242,  242,  242,  242,  242, 
            0,  242,  242,    0,    0,    0,    0,    0,    0,    0, 
            0,  242,    0,    0,  242,    0,    0,  242,    0,  242, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  242, 
          242,  242,    0,  242,    0,  242,    0,    0,    0,    0, 
            0,    0,    0,  243,  243,  243,    0,  242,  243,  243, 
          243,    0,  243,    0,  266,    0,    0,    0,    0,    0, 
            0,    0,  243,  243,  266,    0,    0,  242,    0,    0, 
            0,  243,  243,    0,  243,  243,  243,  243,  243,    0, 
            0,    0,    0,    0,    0,    0,  243,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  266,  243, 
          243,  243,  243,  243,  243,  243,  243,  243,  243,    0, 
          243,  243,  266,  266,    0,    0,    0,  266,    0,    0, 
          243,    0,    0,  243,    0,    0,  243,    0,  243,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  243,  243, 
          243,    0,  243,    0,  243,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  243,    0,    0,    0, 
          265,  265,  265,    0,    0,  265,  265,  265,    0,  265, 
            0,  245,    0,    0,    0,    0,  243,    0,    0,  265, 
          265,  245,    0,    0,    0,    0,    0,    0,  265,  265, 
            0,  265,  265,  265,  265,  265,    0,    0,    0,    0, 
            0,    0,    0,  265,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  245,  265,  265,  265,  265, 
          265,  265,  265,  265,  265,  265,    0,  265,  265,  245, 
          245,    0,    0,    0,  245,    0,    0,  265,    0,    0, 
          265,    0,    0,  265,    0,  265,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  265,  265,    0,    0,    0, 
            0,  265,    0,    0,    0,    0,    0,    0,  247,    0, 
            0,    0,    0,  265,    0,    0,    0,    0,  247,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  265,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  247,    0,    0,    0,  266,  266,  266,    0, 
            0,  266,  266,  266,    0,  266,  247,  247,    0,    0, 
            0,  247,    0,    0,    0,  266,  266,    0,    0,    0, 
            0,    0,    0,    0,  266,  266,    0,  266,  266,  266, 
          266,  266,    0,    0,    0,    0,  248,    0,    0,  266, 
            0,    0,    0,    0,    0,    0,  248,    0,    0,    0, 
            0,    0,  266,  266,  266,  266,  266,  266,  266,  266, 
          266,  266,    0,  266,  266,    0,    0,    0,    0,    0, 
            0,    0,    0,  266,    0,    0,  266,    0,    0,  266, 
          248,  266,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  266,  266,    0,  248,  248,    0,  266,    0,  248, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  266, 
            0,    0,    0,  245,  245,  245,    0,    0,  245,  245, 
          245,    0,  245,    0,  249,    0,    0,    0,    0,  266, 
            0,    0,  245,  245,  249,    0,    0,    0,    0,    0, 
            0,  245,  245,    0,  245,  245,  245,  245,  245,    0, 
            0,    0,    0,    0,    0,    0,  245,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  249,    0, 
            0,    0,    0,    0,    0,  245,  245,    0,    0,    0, 
          245,  245,  249,  249,    0,    0,    0,  249,    0,    0, 
          245,    0,    0,  245,    0,    0,  245,    0,  245,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          247,  247,  247,    0,  245,  247,  247,  247,    0,  247, 
            0,  256,    0,    0,    0,    0,  245,    0,    0,  247, 
          247,  256,    0,    0,    0,    0,    0,    0,  247,  247, 
            0,  247,  247,  247,  247,  247,  245,    0,    0,    0, 
            0,    0,    0,  247,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  256,    0,    0,    0,    0, 
            0,    0,  247,  247,    0,    0,    0,  247,  247,  256, 
          256,    0,    0,    0,  256,    0,    0,  247,    0,    0, 
          247,    0,    0,  247,    0,  247,    0,    0,  248,  248, 
          248,    0,    0,  248,  248,  248,    0,  248,    0,  250, 
            0,  247,    0,    0,    0,    0,    0,  248,  248,  250, 
            0,    0,    0,  247,    0,    0,  248,  248,    0,  248, 
          248,  248,  248,  248,    0,    0,    0,    0,    0,    0, 
            0,  248,    0,  247,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  250,    0,    0,    0,    0,    0,    0, 
          248,  248,    0,    0,    0,  248,  248,  250,  250,    0, 
            0,    0,  250,    0,    0,  248,    0,    0,  248,    0, 
            0,  248,    0,  248,    0,    0,  249,  249,  249,    0, 
            0,  249,  249,  249,    0,  249,    0,  251,    0,  248, 
            0,    0,    0,    0,    0,  249,  249,  251,    0,    0, 
            0,  248,    0,    0,  249,  249,    0,  249,  249,  249, 
          249,  249,    0,    0,    0,    0,    0,    0,    0,  249, 
            0,  248,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,    0,    0,    0,    0,  249,  249, 
            0,    0,    0,  249,  249,  251,  251,    0,    0,    0, 
          251,    0,    0,  249,    0,    0,  249,    0,    0,  249, 
            0,  249,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  256,  256,  256,    0,  249,  256,  256, 
          256,    0,  256,    0,  257,    0,    0,    0,    0,  249, 
            0,    0,  256,  256,  257,    0,    0,    0,    0,    0, 
            0,  256,  256,    0,  256,  256,  256,  256,  256,  249, 
            0,    0,    0,    0,    0,    0,  256,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  257,    0, 
            0,    0,    0,    0,    0,  256,  256,    0,    0,    0, 
          256,  256,  257,  257,    0,    0,    0,  257,    0,    0, 
          256,    0,    0,  256,    0,    0,  256,    0,  256,    0, 
            0,  250,  250,  250,    0,    0,  250,  250,  250,    0, 
          250,    0,  231,    0,  256,    0,    0,    0,    0,    0, 
          250,  250,  231,    0,    0,    0,  256,    0,    0,  250, 
          250,    0,  250,  250,  250,  250,  250,    0,    0,    0, 
            0,    0,    0,    0,  250,    0,  256,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  231,    0,    0,    0, 
            0,    0,    0,  250,  250,    0,    0,    0,  250,  250, 
          231,  231,    0,    0,    0,  231,    0,    0,  250,    0, 
            0,  250,    0,    0,  250,    0,  250,    0,    0,  251, 
          251,  251,    0,    0,  251,  251,  251,    0,  251,    0, 
          232,    0,  250,    0,    0,    0,    0,    0,  251,  251, 
          232,    0,    0,    0,  250,    0,    0,  251,  251,    0, 
          251,  251,  251,  251,  251,    0,    0,    0,    0,    0, 
            0,    0,  251,    0,  250,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  232,    0,    0,    0,    0,    0, 
            0,  251,  251,    0,    0,    0,  251,  251,  232,  232, 
            0,    0,    0,  232,    0,    0,  251,    0,    0,  251, 
            0,    0,  251,    0,  251,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  257,  257,  257,  273, 
          251,  257,  257,  257,    0,  257,    0,    0,    0,  273, 
            0,    0,  251,    0,    0,  257,  257,    0,    0,    0, 
            0,    0,    0,    0,  257,  257,    0,  257,  257,  257, 
          257,  257,  251,    0,    0,    0,    0,    0,    0,  257, 
            0,    0,    0,  273,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  273,  273,  257, 
            0,    0,    0,  257,  257,    0,    0,    0,    0,    0, 
            0,    0,    0,  257,    0,    0,  257,    0,    0,  257, 
            0,  257,    0,    0,  231,  231,  231,    0,  259,  231, 
          231,  231,    0,  231,    0,    0,    0,  257,  259,    0, 
            0,    0,    0,  231,  231,    0,    0,    0,    0,  257, 
            0,    0,  231,  231,    0,  231,  231,  231,  231,  231, 
            0,    0,    0,    0,    0,    0,    0,  231,    0,  257, 
            0,    0,  259,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  259,  259,    0,    0, 
          224,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          224,  231,    0,    0,  231,    0,    0,  231,    0,  231, 
            0,    0,  232,  232,  232,    0,    0,  232,  232,  232, 
            0,  232,    0,    0,    0,  231,    0,    0,    0,    0, 
            0,  232,  232,    0,  224,    0,    0,  231,    0,    0, 
          232,  232,    0,  232,  232,  232,  232,  232,  224,  224, 
            0,    0,    0,    0,    0,  232,    0,  231,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  267,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  267,  232, 
            0,    0,  232,    0,    0,  232,    0,  232,    0,    0, 
            0,  273,  273,  273,    0,    0,  273,  273,  273,    0, 
          273,    0,    0,  232,    0,    0,    0,    0,    0,    0, 
          273,  273,  267,    0,    0,  232,    0,    0,    0,  273, 
          273,    0,  273,  273,  273,  273,  273,  267,    0,    0, 
            0,    0,    0,    0,  273,  232,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  272, 
            0,    0,    0,    0,    0,    0,    0,    0,  273,  272, 
            0,  273,    0,    0,  273,    0,  273,    0,    0,    0, 
          259,  259,  259,    0,    0,  259,  259,  259,    0,  259, 
            0,    0,  273,    0,    0,    0,    0,    0,    0,  259, 
          259,    0,    0,  272,  273,    0,    0,    0,  259,  259, 
            0,  259,  259,  259,  259,  259,    0,  272,  272,    0, 
            0,    0,    0,  259,  273,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  224,  224,  224,  294,    0,  224,  224,  224, 
            0,  224,    0,    0,    0,  294,    0,  259,    0,    0, 
          259,  224,  224,  259,    0,  259,    0,    0,    0,    0, 
          224,  224,    0,  224,  224,  224,  224,  224,    0,    0, 
            0,  259,  648,    0,    0,  224,    0,    0,    0,  294, 
            0,    0,  648,  259,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  294,    0,    0,    0,    0,    0, 
            0,    0,    0,  259,    0,    0,    0,    0,    0,  224, 
            0,    0,  224,    0,    0,  224,    0,  224,    0,    0, 
          267,  267,  267,    0,    0,  267,  267,  267,    0,  267, 
            0,  648,    0,  224,    0,    0,    0,    0,    0,  267, 
          267,    0,    0,    0,    0,  224,    0,    0,  267,  267, 
            0,  267,  267,  267,  267,  267,    0,    0,    0,    0, 
           61,    0,    0,  267,    0,  224,    0,    0,    0,    0, 
           61,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  267,    0,    0, 
          267,    0,    0,  267,    0,  267,    0,    0,    0,    0, 
            0,  272,  272,  272,    0,    0,  272,  272,  272,   61, 
          272,  267,   65,    0,    0,    0,    0,    0,    0,    0, 
          272,  272,   65,  267,    0,    0,    0,    0,    0,  272, 
          272,    0,  272,  272,  272,  272,    0,    0,    0,    0, 
            0,    0,    0,  267,  272,    0,    0,    0,    0,    0, 
            0,    0,    0,   63,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,   63,    0,    0,    0,    0,    0,   25, 
            0,   65,    0,    0,    0,    0,    0,    0,  272,   25, 
            0,  272,    0,    0,  272,    0,  272,  294,  294,  294, 
            0,    0,  294,  294,  294,    0,  294,    0,    0,    0, 
            0,    0,  272,    0,    0,    0,  294,  294,    0,    0, 
            0,    0,   63,    0,  272,  294,  294,    0,  294,  294, 
          294,  294,  294,    0,  648,  648,  648,    0,   25,  648, 
          648,  648,    0,  648,  272,    0,    0,    0,    0,    0, 
            0,    0,    0,  648,  648,    0,    0,    0,    0,    0, 
            0,    0,  648,  648,    0,  648,  648,  648,  648,  648, 
            0,    0,    0,    0,    0,    0,    0,  294,    0,    0, 
          294,    0,  294,    0,    0,    0,    0,    0,  308,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  294,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  648,    0,    0,  648,    0,  648, 
            0,    0,   61,   61,   61,    0,    0,   61,   61,   61, 
          294,   61,    0,    0,    0,  648,    0,    0,    0,    0, 
            0,   61,   61,    0,    0,    0,    0,    0,    0,    0, 
           61,   61,    0,   61,   61,   61,   61,   61,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  648,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   65,   65,   65,    0,    0,   65, 
           65,   65,    0,   65,    0,    0,    0,    0,    0,    0, 
            0,    0,   61,   65,   65,   61,    0,    0,    0,    0, 
            0,    0,   65,   65,    0,   65,   65,   65,   65,   65, 
            0,    0,    0,   61,    0,   63,   63,   63,  130,    0, 
           63,   63,   63,    0,   63,    0,    0,    0,    0,    0, 
            0,   25,   25,   25,   63,   63,    0,   25,   25,    0, 
           25,    0,    0,   63,   63,   61,   63,   63,   63,   63, 
           63,    0,    0,    0,   65,    0,    0,   65,    0,    0, 
            0,    0,   25,   25,   25,   25,   25,  130,    0,    0, 
            0,    0,    0,    0,    0,   65,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   63,    0,    0,   63,  515, 
            0,    0,    0,    0,    0,    0,    0,   65,    0,    0, 
            0,   25,    0,    0,    0,    0,   63,    0,    0,    0, 
            0,    0,    0,    0,    0,    4,    5,    6,    0,    8, 
            0,    0,   25,    9,   10,    0,    0,    0,   11,    0, 
           12,   13,   14,   15,   16,   17,   18,    0,   63,    0, 
            0,   19,   20,   21,  239,  240,  241,  242,  131,    0, 
          118,    0,    0,    0,   25,    0,    0,   28,    0,    0, 
          243,  244,  245,    0,  246,   35,  247,  248,  249,  250, 
          285,   40,   41,   42,   43,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  251,    0,  131,  131,    0, 
            0,   46,   47,    0,   48,    0,  286,    0,  287,    0, 
           50,    0,    0,    0,    0,    0,    0,    0,  288,    0, 
            0,    0,    0,   52,  289,   54,   55,   56,   57,  518, 
            0,    0,   58,    0,   59,   60,    0,   61,   62,   63, 
           64,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
           65,   66,  290,    0,  130,  130,  130,  130,  130,  130, 
          130,  130,  130,  130,  130,    0,    0,  130,  130,    0, 
          130,  130,  130,  130,  130,  130,  130,    0,  618,    0, 
            0,  130,  130,  130,  130,  130,  130,  130,  130,    0, 
          130,    0,    0,    0,    0,    0,  130,  130,    0,  130, 
          130,  130,  130,    0,  130,  130,  130,  130,  130,  130, 
            0,  130,  130,  130,  130,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  515,    0,    0,    0, 
            0,    0,    0,  515,  515,  130,    0,  130,  130,  618, 
            0,  130,  130,    0,  130,    0,  130,    0,    0,    0, 
          130,    0,    0,    0,    0,  130,    0,    0,  130,    0, 
          618,    0,    0,  130,  130,  130,  130,  130,  130,  625, 
            0,    0,  130,    0,  130,  130,    0,  130,  130,  130, 
          130,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          130,  130,    0,    0,  131,  131,  131,  131,  131,  131, 
          131,  131,  131,  131,  131,    0,    0,  131,  131,    0, 
          131,  131,  131,  131,  131,  131,  131,    0,  619,    0, 
            0,  131,  131,  131,  131,  131,  131,  131,  131,    0, 
          131,    0,    0,    0,    0,    0,  131,  131,    0,  131, 
          131,  131,  131,    0,  131,  131,  131,  131,  131,  131, 
            0,  131,  131,  131,  131,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  518,    0,    0,    0, 
            0,    0,    0,  518,  518,  131,    0,  131,  131,  619, 
            0,  131,  131,    0,  131,    0,  131,    0,    0,    0, 
          131,    0,    0,    0,    0,  131,    0,    0,  131,    0, 
          619,    0,    0,  131,  131,  131,  131,  131,  131,  320, 
            0,    0,  131,    0,  131,  131,    0,  131,  131,  131, 
          131,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          131,  131,    0,    0,  130,  130,  130,  130,  130,  130, 
          130,  130,  130,  130,  130,    0,    0,  130,  130,    0, 
          130,  130,  130,  130,  130,  130,  130,    0,  625,    0, 
            0,  130,  130,  130,  130,  130,  130,  130,  131,    0, 
          130,    0,    0,    0,    0,    0,  130,  130,    0,  130, 
          130,  130,  130,    0,  130,  130,  130,  130,  130,  130, 
            0,  130,  130,  130,  130,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  625,    0,    0,    0, 
            0,    0,    0,  625,  625,  130,    0,  131,  130,  621, 
            0,  130,  130,    0,  130,    0,  130,    0,    0,    0, 
          130,    0,    0,    0,    0,  130,    0,    0,  130,    0, 
          625,    0,    0,  130,  130,  130,  130,  130,  130,  321, 
            0,    0,  130,    0,  130,  130,    0,  130,  130,  130, 
          130,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          130,  130,    0,    0,  131,  131,  131,  131,  131,  131, 
          131,  131,  131,  131,  131,    0,    0,  131,  131,    0, 
          131,  131,  131,  131,  131,  131,  131,    0,    0,    0, 
            0,  131,  131,  131,  131,  131,  131,  131,  364,    0, 
          131,    0,    0,    0,    0,    0,  131,  131,    0,  131, 
          131,  131,  131,    0,  131,  131,  131,  131,  131,  131, 
            0,  131,  131,  131,  131,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  320,    0,    0,    0, 
            0,    0,    0,  320,  320,  131,    0,  364,  131,  622, 
            0,  131,  131,    0,  131,    0,  131,    0,    0,    0, 
          131,    0,    0,    0,    0,  131,    0,    0,  131,    0, 
            0,    0,    0,  131,  131,  131,  131,  131,  131,    0, 
            0,    0,  131,    0,  131,  131,    0,  131,  131,  131, 
          131,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          131,  131,    0,    0,  131,  131,  131,  131,  131,  131, 
          131,  131,  131,  131,  131,    0,    0,  131,  131,    0, 
          131,  131,  131,  131,  131,  131,  131,    0,    0,    0, 
            0,  131,  131,  131,  131,  131,  131,  131,  312,    0, 
          131,    0,    0,    0,    0,    0,  131,  131,    0,  131, 
          131,  131,  131,    0,  131,  131,  131,  131,  131,  131, 
            0,  131,  131,  131,  131,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  321,    0,    0,    0, 
            0,    0,    0,  321,  321,  131,    0,  312,  131,    0, 
            0,  131,  131,    0,  131,    0,  131,    0,    0,    0, 
          131,    0,    0,    0,    0,  131,    0,    0,  131,    0, 
            0,    0,    0,  131,  131,  131,  131,  131,  131,    0, 
            0,    0,  131,    0,  131,  131,    0,  131,  131,  131, 
          131,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          131,  131,    0,    0,  364,  364,  364,  364,  364,  364, 
          364,  364,  364,  364,  364,    0,  364,  364,  364,  364, 
          364,  364,  364,  364,  364,  364,  364,    0,    0,    0, 
            0,  364,  364,  364,  364,  364,  364,  364,  543,    0, 
          364,    0,    0,    0,    0,    0,  364,  364,    0,  364, 
          364,  364,  364,    0,  364,  364,  364,  364,  364,  364, 
            0,  364,  364,  364,  364,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  364,    0,  543,  364,    0, 
            0,  364,  364,    0,  364,    0,  364,    0,    0,    0, 
          364,    0,    0,    0,    0,    0,    0,    0,  364,    0, 
            0,    0,    0,  364,  364,  364,  364,  364,  364,    0, 
            0,    0,  364,    0,  364,  364,    0,  364,  364,  364, 
          364,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          364,  364,    0,    0,  312,  312,  312,  312,  312,  312, 
          312,  312,  312,  312,  312,    0,    0,  312,  312,    0, 
          312,  312,  312,  312,  312,  312,  312,    0,    0,    0, 
            0,  312,  312,  312,  312,  312,  312,  312,  647,    0, 
          312,    0,    0,    0,    0,    0,  312,  312,    0,  312, 
          312,  312,  312,    0,  312,  312,  312,  312,  312,  312, 
            0,  312,  312,  312,  312,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  312,    0,  647,  312,    0, 
            0,  312,  312,    0,  312,    0,  312,    0,    0,    0, 
          312,    0,    0,    0,    0,    0,    0,    0,  312,    0, 
           18,    0,    0,  312,  312,  312,  312,  312,  312,    0, 
            0,    0,  312,    0,  312,  312,    0,  312,  312,  312, 
          312,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          312,  312,    0,    0,  543,  543,  543,  543,  543,  543, 
          543,  543,  543,  543,  543,    0,    0,  543,  543,    0, 
          543,  543,  543,  543,  543,  543,  543,    0,  647,    0, 
            0,  543,  543,  543,  543,  543,  543,  543,    0,    0, 
          543,    0,    0,    0,    0,    0,  543,  543,    0,  543, 
          543,  543,  543,    0,  543,  543,  543,  543,  543,  543, 
            0,  543,  543,  543,  543,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  647,    0,    0, 
            0,    0,    0,    0,    0,  543,    0,    0,  543,    0, 
            0,  543,  543,    0,  543,    0,  543,    0,    0,    0, 
          543,    0,    0,    0,    0,    0,    0,    0,  543,    0, 
            0,    0,    0,  543,  543,  543,  543,  543,  543,    0, 
            0,    0,  543,    0,  543,  543,    0,  543,  543,  543, 
          543,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          543,  543,    0,    0,  647,  647,  647,  647,  647,  647, 
          647,  647,  647,  647,  647,    0,    0,  647,  647,    0, 
          647,  647,  647, 
      };
   }

   private static final short[] yyTable4() {
      return new short[] {

          647,  647,  647,  647,  647,    0,    0,    0,  647,  647, 
          647,  647,  647,  647,  647,    0,    0,  647,    0,    0, 
            0,    0,    0,  647,  647,    0,  647,  647,  647,  647, 
            0,  647,  647,  647,  647,  647,  647,    0,  647,  647, 
          647,  647,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  647,    0,    0,    0,    0,    0,    0, 
            0,    0,  647,    0,    0,  647,    0,    0,  647,  647, 
            0,  647,    0,  647,    0,    0,    0,  647,    0,    0, 
            0,    0,    0,    0,    0,  647,    0,    0,    0,    0, 
          647,  647,  647,  647,  647,  647,    0,    0,    0,  647, 
            0,  647,  647,    0,  647,  647,  647,  647,    0,    0, 
            0,  647,  647,  647,  647,  647,  647,  647,  647,  647, 
          647,  647,    0,    0,    0,  647,    0,  647,  647,  647, 
          647,  647,  647,  647,  314,    0,    0,    0,  647,  647, 
          647,  647,  647,  647,  647,    0,    0,  647,    0,    0, 
            0,    0,    0,  647,  647,    0,  647,  647,  647,  647, 
            0,  647,  647,  647,  647,  647,  647,    0,  647,  647, 
          647,  647,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  647,    0,    0,  647,    0,    0,  647,  647, 
            0,  647,    0,  647,    0,    0,    0,  647,    0,    0, 
            0,    0,    0,    0,    0,  647,    0,    0,    0,    0, 
          647,  647,  647,  647,  647,  647,    0,    0,    0,  647, 
            0,  647,  647,    0,  647,  647,  647,  647,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  647,  647,    0, 
          647,  647,  647,  647,  647,  647,    0,    0,    0,  647, 
          647,    0,    0,  260,  647,    0,  647,  647,  647,  647, 
          647,  647,  647,    0,    0,    0,    0,  647,  647,  647, 
          647,  647,  647,  647,    0,    0,  647,    0,    0,    0, 
            0,    0,  647,  647,    0,  647,  647,  647,  647,    0, 
          647,  647,  647,  647,  647,  647,    0,  647,  647,  647, 
          647,    0,  259,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  647,    0,    0,  647,    0,    0,  647,  647,    0, 
          647,    0,  647,    0,    0,    0,  647,    0,    0,    0, 
            0,    0,    0,    0,  647,    0,    0,  647,    0,  647, 
          647,  647,  647,  647,  647,    0,    0,    0,  647,    0, 
          647,  647,    0,  647,  647,  647,  647,    0,    0,    0, 
            0,  316,  316,  316,  316,  316,  647,  647,    0,  316, 
          316,    0,  308,    0,  316,    0,  316,  316,  316,  316, 
          316,  316,  316,    0,    0,    0,    0,  316,  316,  316, 
          316,  316,  316,  316,    0,    0,  316,    0,    0,    0, 
            0,    0,  316,  316,    0,  316,  316,  316,  316,    0, 
          316,  316,  316,  316,  316,  316,    0,  316,  316,  316, 
          316,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  316,    0,    0,  316,    0,  314,  316,  316,    0, 
          316,    0,  316,    0,    0,    0,  316,    0,    0,    0, 
            0,    0,    0,    0,  316,    0,    0,    0,    0,  316, 
          316,  316,  316,  316,  316,    0,    0,    0,  316,    0, 
          316,  316,    0,  316,  316,  316,  316,    0,    0,    0, 
            4,    5,    6,    0,    8,    0,  316,  316,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,   15,   16, 
           17,   18,    0,    0,    0,    0,   19,   20,   21,  239, 
          240,  241,  242,    0,    0,   26,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  353,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,    0,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,   51,    0,    0,    0,    0,   52,   53, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    4, 
            5,    6,    0,    8,    0,   65,   66,    9,   10,    0, 
            0,    0,   11,    0,   12,   13,   14,   15,   16,   17, 
           18,    0,    0,    0,    0,   19,   20,   21,  239,  240, 
          241,  242,    0,    0,   26,    0,    0,    0,    0,    0, 
            0,   28,    0,    0,  243,  244,  245,    0,  246,   35, 
          247,  248,  249,  250,  654,   40,   41,   42,   43,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  251, 
            0,    0,  131,    0,    0,   46,   47,    0,   48,    0, 
            0,    0,    0,    0,   50,    0,    0,    0,    0,    0, 
            0,    0,   51,    0,    0,    0,    0,   52,   53,   54, 
           55,   56,   57,    0,    0,    0,   58,    0,   59,   60, 
            0,   61,   62,   63,   64,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   65,   66,    0,    0,    4,    5, 
            6,    0,    8,    0,    0,    0,    9,   10,    0,    0, 
            0,   11,    0,   12,   13,   14,  109,  110,   17,   18, 
            0,    0,    0,    0,  111,  112,  113,  340,  341,  342, 
          343,    0,    0,  118,    0,    0,    0,    0,    0,    0, 
          119,    0,    0,  344,  345,  346,    0,  347,   35,  348, 
          349,  350,  351,    0,   40,    0,    0,  128,  471,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  352,    0, 
            0,  131,    0,    0,   46,   47,    0,   48,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,   52,   53,   54,   55, 
           56,   57,    0,    0,    0,   58,    0,   59,   60,    0, 
           61,   62,   63,   64,    0,    0,    0,    4,    5,    6, 
            0,    8,    0,   65,   66,    9,   10,    0,    0,    0, 
           11,    0,   12,   13,   14,  109,  110,   17,   18,    0, 
            0,    0,    0,  111,  112,  113,  340,  341,  342,  343, 
            0,    0,  118,    0,    0,    0,    0,    0,    0,  119, 
            0,    0,  344,  345,  346,    0,  347,   35,  348,  349, 
          350,  351,    0,   40,    0,    0,  128,  471,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  352,    0,    0, 
          131,    0,    0,   46,   47,    0,   48,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   52,   53,   54,   55,   56, 
           57,    0,    0,    0,   58,    0,   59,   60,    0,   61, 
           62,   63,   64,    0,    0,    0,    4,    5,    6,    7, 
            8,    0,   65,   66,    9,   10,    0,    0,    0,   11, 
            0,   12,   13,   14,   15,   16,   17,   18,    0,    0, 
            0,    0,   19,   20,   21,   22,   23,   24,   25,    0, 
            0,   26,    0,    0,    0,    0,    0,   27,   28,   29, 
           30,   31,   32,   33,    0,   34,   35,   36,   37,   38, 
           39,    0,   40,   41,   42,   43,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,   44,    0,    0,   45, 
            0,    0,   46,   47,    0,   48,    0,   49,    0,    0, 
            0,   50,    0,    0,    0,    0,    0,    0,    0,   51, 
            0,    0,    0,    0,   52,   53,   54,   55,   56,   57, 
            0,    0,    0,   58,    0,   59,   60,    0,   61,   62, 
           63,   64,    0,    0,    0,    4,    5,    6,    7,  321, 
            0,   65,   66,    9,   10,    0,    0,    0,   11,    0, 
           12,   13,   14,   15,   16,   17,   18,    0,    0,    0, 
            0,   19,   20,   21,   22,   23,   24,   25,    0,    0, 
           26,    0,    0,    0,    0,    0,   27,   28,    0,   30, 
           31,   32,   33,    0,   34,   35,   36,   37,   38,   39, 
            0,   40,   41,   42,   43,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   44,    0,    0,   45,    0, 
            0,   46,   47,    0,   48,    0,   49,    0,    0,    0, 
           50,    0,    0,    0,    0,    0,    0,    0,   51,    0, 
            0,    0,    0,   52,   53,   54,   55,   56,   57,    0, 
            0,    0,   58,    0,   59,   60,    0,   61,   62,   63, 
           64,    0,    0,    3,    4,    5,    6,    7,    8,    0, 
           65,   66,    9,   10,    0,    0,    0,   11,    0,   12, 
           13,   14,   15,   16,   17,   18,    0,    0,    0,    0, 
           19,   20,   21,   22,   23,   24,   25,    0,    0,   26, 
            0,    0,    0,    0,    0,   27,   28,   29,   30,   31, 
           32,   33,    0,   34,   35,   36,   37,   38,   39,    0, 
           40,   41,   42,   43,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   44,    0,    0,   45,    0,    0, 
           46,   47,    0,   48,    0,   49,    0,    0,    0,   50, 
            0,    0,    0,    0,    0,    0,    0,   51,    0,    0, 
            0,    0,   52,   53,   54,   55,   56,   57,    0,    0, 
            0,   58,    0,   59,   60,    0,   61,   62,   63,   64, 
            0,    0,  320,    4,    5,    6,    7,  321,    0,   65, 
           66,    9,   10,    0,    0,    0,   11,    0,   12,   13, 
           14,   15,   16,   17,   18,    0,    0,    0,    0,   19, 
           20,   21,   22,   23,   24,   25,    0,    0,   26,    0, 
            0,    0,    0,    0,   27,   28,    0,   30,   31,   32, 
           33,    0,   34,   35,   36,   37,   38,   39,    0,   40, 
           41,   42,   43,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,   44,    0,    0,  322,    0,    0,   46, 
           47,    0,   48,    0,   49,    0,    0,    0,   50,    0, 
            0,    0,    0,    0,    0,    0,   51,    0,    0,    0, 
            0,   52,   53,   54,   55,   56,   57,    0,    0,    0, 
           58,    0,   59,   60,    0,   61,   62,   63,   64,    0, 
            0,  320,    4,    5,    6,    7,  321,    0,   65,   66, 
            9,   10,    0,    0,    0,   11,    0,   12,   13,   14, 
           15,   16,   17,   18,    0,    0,    0,    0,   19,   20, 
           21,   22,   23,   24,   25,    0,    0,   26,    0,    0, 
            0,    0,    0,   27,   28,    0,   30,   31,   32,   33, 
            0,   34,   35,   36,   37,   38,   39,    0,   40,   41, 
           42,   43,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,   44,    0,    0,   45,    0,    0,   46,   47, 
            0,   48,    0,   49,    0,    0,    0,   50,    0,    0, 
            0,    0,    0,    0,    0,   51,    0,    0,    0,    0, 
           52,   53,   54,   55,   56,   57,    0,    0,    0,   58, 
            0,   59,   60,    0,   61,   62,   63,   64,    0,    0, 
            0,    4,    5,    6,    0,    8,    0,   65,   66,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,   15, 
           16,   17,   18,    0,    0,    0,    0,   19,   20,   21, 
          239,  240,  241,  242,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,  285,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,  286,    0,  287,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,  288,    0,    0,    0,    0,   52, 
          289,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            0,    4,    5,    6,    0,    8,   65,   66,  290,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,  109, 
          110,   17,   18,    0,    0,    0,    0,  111,  112,  113, 
          239,  240,  241,  242,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,  285,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,  708,    0,  287,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,  288,    0,    0,    0,    0,   52, 
          289,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            0,  287,  287,  287,    0,  287,   65,   66,  290,  287, 
          287,    0,    0,    0,  287,    0,  287,  287,  287,  287, 
          287,  287,  287,    0,    0,    0,    0,  287,  287,  287, 
          287,  287,  287,  287,    0,    0,  287,    0,    0,    0, 
            0,    0,    0,  287,    0,    0,  287,  287,  287,    0, 
          287,  287,  287,  287,  287,  287,  287,  287,  287,  287, 
          287,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  287,    0,    0,  287,    0,    0,  287,  287,    0, 
          287,    0,  287,    0,  287,    0,  287,    0,    0,    0, 
            0,    0,    0,    0,  287,    0,    0,    0,    0,  287, 
          287,  287,  287,  287,  287,    0,    0,    0,  287,    0, 
          287,  287,    0,  287,  287,  287,  287,    0,    0,    0, 
            0,    4,    5,    6,    0,    8,  287,  287,  287,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,  109, 
          110,   17,   18,    0,    0,    0,    0,  111,  112,  113, 
          239,  240,  241,  242,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,  285,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,  286,    0,    0,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,  288,    0,    0,    0,    0,   52, 
          289,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            0,    4,    5,    6,    0,    8,   65,   66,  290,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,  109, 
          110,   17,   18,    0,    0,    0,    0,  111,  112,  113, 
          239,  240,  241,  242,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,  285,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,    0,    0,  287,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,  288,    0,    0,    0,    0,   52, 
          289,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            0,    4,    5,    6,    0,    8,   65,   66,  290,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,  109, 
          110,   17,   18,    0,    0,    0,    0,  111,  112,  113, 
          239,  240,  241,  242,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,  285,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,  708,    0,    0,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,  288,    0,    0,    0,    0,   52, 
          289,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            0,    4,    5,    6,    0,    8,   65,   66,  290,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,  109, 
          110,   17,   18,    0,    0,    0,    0,  111,  112,  113, 
          239,  240,  241,  242,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,  285,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,    0,    0,    0,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,  288,    0,    0,    0,    0,   52, 
          289,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            0,    4,    5,    6,    7,    8,   65,   66,  290,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,   15, 
           16,   17,   18,    0,    0,    0,    0,   19,   20,   21, 
           22,   23,   24,   25,    0,    0,   26,    0,    0,    0, 
            0,    0,   27,   28,   29,   30,   31,   32,   33,    0, 
           34,   35,   36,   37,   38,   39,    0,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,   44,    0,    0,   45,    0,    0,   46,   47,    0, 
           48,    0,   49,    0,    0,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,   51,    0,    0,    0,    0,   52, 
           53,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            4,    5,    6,    7,    8,    0,   65,   66,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,   15,   16, 
           17,   18,    0,    0,    0,    0,   19,   20,   21,   22, 
           23,   24,   25,    0,    0,   26,    0,    0,    0,    0, 
            0,   27,   28,    0,   30,   31,   32,   33,    0,   34, 
           35,   36,   37,   38,   39,    0,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
           44,    0,    0,   45,    0,    0,   46,   47,    0,   48, 
            0,   49,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,   51,    0,    0,    0,    0,   52,   53, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    4, 
            5,    6,    0,    8,    0,   65,   66,    9,   10,    0, 
            0,    0,   11,    0,   12,   13,   14,  109,  110,   17, 
           18,    0,    0,    0,    0,  111,  112,  113,  239,  240, 
          241,  242,    0,    0,  118,    0,    0,    0,    0,    0, 
            0,   28,    0,    0,  243,  244,  245,    0,  246,   35, 
          247,  248,  249,  250,    0,   40,   41,   42,   43,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  251, 
            0,    0,  131,  506,    0,   46,   47,    0,   48,    0, 
            0,    0,    0,    0,   50,    0,    0,    0,    0,    0, 
            0,    0,  288,    0,    0,    0,    0,   52,   53,   54, 
           55,   56,   57,    0,    0,    0,   58,    0,   59,   60, 
            0,   61,   62,   63,   64,    0,    0,    0,    4,    5, 
            6,    0,    8,    0,   65,   66,    9,   10,    0,    0, 
            0,   11,    0,   12,   13,   14,   15,   16,   17,   18, 
            0,    0,    0,    0,   19,   20,   21,  239,  240,  241, 
          242,    0,    0,  118,    0,    0,    0,    0,    0,    0, 
           28,    0,    0,  243,  244,  245,    0,  246,   35,  247, 
          248,  249,  250,    0,   40,   41,   42,   43,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  251,    0, 
            0,  131,    0,    0,   46,   47,    0,   48,    0,  646, 
            0,    0,    0,   50,    0,    0,    0,    0,    0,    0, 
            0,  288,    0,    0,    0,    0,   52,   53,   54,   55, 
           56,   57,    0,    0,    0,   58,    0,   59,   60,    0, 
           61,   62,   63,   64,    0,    0,    0,    4,    5,    6, 
            0,    8,    0,   65,   66,    9,   10,    0,    0,    0, 
           11,    0,   12,   13,   14,  109,  110,   17,   18,    0, 
            0,    0,    0,  111,  112,  113,  239,  240,  241,  242, 
            0,    0,  118,    0,    0,    0,    0,    0,    0,   28, 
            0,    0,  243,  244,  245,    0,  246,   35,  247,  248, 
          249,  250,    0,   40,   41,   42,   43,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  251,    0,    0, 
          131,    0,    0,   46,   47,    0,   48,    0,  286,    0, 
            0,    0,   50,    0,    0,    0,    0,    0,    0,    0, 
          288,    0,    0,    0,    0,   52,   53,   54,   55,   56, 
           57,    0,    0,    0,   58,    0,   59,   60,    0,   61, 
           62,   63,   64,    0,    0,    0,    4,    5,    6,    0, 
            8,    0,   65,   66,    9,   10,    0,    0,    0,   11, 
            0,   12,   13,   14,  109,  110,   17,   18,    0,    0, 
            0,    0,  111,  112,  113,  239,  240,  241,  242,    0, 
            0,  118,    0,    0,    0,    0,    0,    0,   28,    0, 
            0,  243,  244,  245,    0,  246,   35,  247,  248,  249, 
          250,    0,   40,   41,   42,   43,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  251,    0,    0,  131, 
            0,    0,   46,   47,    0,   48,    0,  646,    0,    0, 
            0,   50,    0,    0,    0,    0,    0,    0,    0,  288, 
            0,    0,    0,    0,   52,   53,   54,   55,   56,   57, 
            0,    0,    0,   58,    0,   59,   60,    0,   61,   62, 
           63,   64,    0,    0,    0,    4,    5,    6,    0,    8, 
            0,   65,   66,    9,   10,    0,    0,    0,   11,    0, 
           12,   13,   14,  109,  110,   17,   18,    0,    0,    0, 
            0,  111,  112,  113,  239,  240,  241,  242,    0,    0, 
          118,    0,    0,    0,    0,    0,    0,   28,    0,    0, 
          243,  244,  245,    0,  246,   35,  247,  248,  249,  250, 
            0,   40,   41,   42,   43,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  251,    0,    0,  131,    0, 
            0,   46,   47,    0,   48,    0,  924,    0,    0,    0, 
           50,    0,    0,    0,    0,    0,    0,    0,  288,    0, 
            0,    0,    0,   52,   53,   54,   55,   56,   57,    0, 
            0,    0,   58,    0,   59,   60,    0,   61,   62,   63, 
           64,    0,    0,    0,    4,    5,    6,    0,    8,    0, 
           65,   66,    9,   10,    0,    0,    0,   11,    0,   12, 
           13,   14,  109,  110,   17,   18,    0,    0,    0,    0, 
          111,  112,  113,  239,  240,  241,  242,    0,    0,  118, 
            0,    0,    0,    0,    0,    0,   28,    0,    0,  243, 
          244,  245,    0,  246,   35,  247,  248,  249,  250,    0, 
           40,   41,   42,   43,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  251,    0,    0,  131,    0,    0, 
           46,   47,    0,   48,    0,  708,    0,    0,    0,   50, 
            0,    0,    0,    0,    0,    0,    0,  288,    0,    0, 
            0,    0,   52,   53,   54,   55,   56,   57,    0,    0, 
            0,   58,    0,   59,   60,    0,   61,   62,   63,   64, 
            0,    0,    0,  636,  636,  636,    0,  636,    0,   65, 
           66,  636,  636,    0,    0,    0,  636,    0,  636,  636, 
          636,  636,  636,  636,  636,    0,    0,    0,    0,  636, 
          636,  636,  636,  636,  636,  636,    0,    0,  636,    0, 
            0,    0,    0,    0,    0,  636,    0,    0,  636,  636, 
          636,    0,  636,  636,  636,  636,  636,  636,    0,  636, 
          636,  636,  636,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  636,    0,    0,  636,  636,    0,  636, 
          636,    0,  636,    0,    0,    0,    0,    0,  636,    0, 
            0,    0,    0,    0,    0,    0,  636,    0,    0,    0, 
            0,  636,  636,  636,  636,  636,  636,    0,    0,    0, 
          636,    0,  636,  636,    0,  636,  636,  636,  636,    0, 
            0,    0,    4,    5,    6,    0,    8,    0,  636,  636, 
            9,   10,    0,    0,    0,   11,    0,   12,   13,   14, 
           15,   16,   17,   18,    0,    0,    0,    0,   19,   20, 
           21,  239,  240,  241,  242,    0,    0,   26,    0,    0, 
            0,    0,    0,    0,   28,    0,    0,  243,  244,  245, 
            0,  246,   35,  247,  248,  249,  250,    0,   40,   41, 
           42,   43,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  251,    0,    0,  131,    0,    0,   46,   47, 
            0,   48,    0,    0,    0,    0,    0,   50,    0,    0, 
            0,    0,    0,    0,    0,   51,    0,    0,    0,    0, 
           52,   53,   54,   55,   56,   57,    0,    0,    0,   58, 
            0,   59,   60,    0,   61,   62,   63,   64,    0,    0, 
            0,    4,    5,    6,    0,    8,    0,   65,   66,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,  109, 
          110,   17,   18,    0,    0,    0,    0,  111,  112,  113, 
          239,  240,  241,  242,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,    0,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,    0,    0,    0,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,  288,    0,    0,    0,    0,   52, 
           53,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            4,    5,    6,    0,    8,    0,   65,   66,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,   15,   16, 
           17,   18,    0,    0,    0,    0,   19,   20,   21,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,    0,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,    0,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,   53, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,  636, 
          636,  636,    0,  636,    0,   65,   66,  636,  636,    0, 
            0,    0,  636,    0,  636,  636,  636,  636,  636,  636, 
          636,    0,    0,    0,    0,  636,  636,  636,  636,  636, 
          636,  636,    0,    0,  636,    0,    0,    0,    0,    0, 
            0,  636,    0,    0,  636,  636,  636,    0,  636,  636, 
          636,  636,  636,  636,    0,  636,  636,  636,  636,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  636, 
            0,    0,  636,    0,    0,  636,  636,    0,  636,    0, 
            0,    0,    0,    0,  636,    0,    0,    0,    0,    0, 
            0,    0,  636,    0,    0,    0,    0,  636,  636,  636, 
          636,  636,  636,    0,    0,    0,  636,    0,  636,  636, 
            0,  636,  636,  636,  636,    0,    0,    0,    4,    5, 
            6,    0,    8,    0,  636,  636,    9,   10,    0,    0, 
            0,   11,    0,   12,   13,   14,  109,  110,   17,   18, 
            0,    0,    0,    0,  111,  112,  113,  114,  115,  116, 
          117,    0,    0,  118,    0,    0,    0,    0,    0,    0, 
          119,    0,    0,  120,  121,  122,    0,  123,   35,  124, 
          125,  126,  127,    0,   40,    0,    0,  128,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  129,    0,    0,    0,  130,    0, 
            0,  131,    0,    0,   46,   47,    0,   48,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,   52,   53,   54,   55, 
           56,   57,    0,    0,    0,   58,    0,   59,   60,    0, 
           61,   62,   63,   64,    0,    0,    0,    4,    5,    6, 
            0,    8,    0,   65,   66,    9,   10,    0,    0,    0, 
           11,    0,   12,   13,   14,  109,  110,   17,   18,    0, 
            0,    0,    0,  111,  112,  113,  267,  268,  269,  270, 
            0,    0,  118,    0,    0,    0,    0,    0,    0,  119, 
            0,    0,  271,  272,  273,    0,  274,   35,  275,  276, 
          277,  278,    0,   40,    0,    0,  128,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  279,    0,    0, 
           45,    0,    0,   46,   47,    0,   48,    0,   49,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   52,   53,   54,   55,   56, 
           57,    0,    0,    0,   58,    0,   59,   60,    0,   61, 
           62,   63,   64,    0,    0,    0,    4,    5,    6,    0, 
            8,    0,   65,   66,    9,   10,    0,    0,    0,   11, 
            0,   12,   13,   14,  109,  110,   17,   18,    0,    0, 
            0,    0,  111,  112,  113,  340,  341,  342,  343,    0, 
            0,  118,    0,    0,    0,    0,    0,    0,  119,    0, 
            0,  344,  345,  346,    0,  347,   35,  348,  349,  350, 
          351,    0,   40,    0,    0,  128,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  352,    0,    0,  427, 
            0,    0,   46,   47,    0,   48,    0,  428,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   52,   53,   54,   55,   56,   57, 
            0,    0,    0,   58,    0,   59,   60,    0,   61,   62, 
           63,   64,    0,    0,    0,    4,    5,    6,    0,    8, 
            0,   65,   66,    9,   10,    0,    0,    0,   11,    0, 
           12,   13,   14,  109,  110,   17,   18,    0,    0,    0, 
            0,  111,  112,  113,  114,  115,  116,  117,    0,    0, 
          118,    0,    0,    0,    0,    0,    0,  119,    0,    0, 
          120,  121,  122,    0,  123,   35,  124,  125,  126,  127, 
            0,   40,    0,    0,  128,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  130,    0,    0,  131,    0, 
            0,   46,   47,    0,   48,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,   52,   53,   54,   55,   56,   57,    0, 
            0,    0,   58,    0,   59,   60,    0,   61,   62,   63, 
           64,    0,    0,    0,    4,    5,    6,    0,    8,    0, 
           65,   66,    9,   10,    0,    0,    0,   11,    0,   12, 
           13,   14,  109,  110,   17,   18,    0,    0,    0,    0, 
          111,  112,  113,  340,  341,  342,  343,    0,    0,  118, 
            0,    0,    0,    0,    0,    0,  119,    0,    0,  344, 
          345,  346,    0,  347,   35,  348,  349,  350,  351,    0, 
           40,    0,    0,  128,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  352,    0,    0,  427,    0,    0, 
           46,   47,    0,   48,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,   52,   53,   54,   55,   56,   57,    0,    0, 
            0,   58,    0,   59,   60,    0,   61,   62,   63,   64, 
            0,    0,    0,    4,    5,    6,    0,    8,    0,   65, 
           66,    9,   10,    0,    0,    0,   11,    0,   12,   13, 
           14,  109,  110,   17,   18,    0,    0,    0,    0,  111, 
          112,  113,  114,  115,  116,  117,    0,    0,  118,    0, 
            0,    0,    0,    0,    0,  119,    0,    0,  120,  121, 
          122,    0,  984,   35,  124,  125,  985,  127,    0,   40, 
            0,    0,  128,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  986,    0,    0,  131,    0,    0,   46, 
           47,    0,   48,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,   52,   53,   54,   55,   56,   57,    0,    0,    0, 
           58,    0,   59,   60,    0,   61,   62,   63,   64,    0, 
            0,    0,    4,    5,    6,    0,    8,    0,   65,   66, 
            9,   10,    0,    0,    0,   11,    0,   12,   13,   14, 
          109,  110,   17,   18,    0,    0,    0,    0,  111,  112, 
          113,  997,  998,  999, 1000,    0,    0,  118,    0,    0, 
            0,    0,    0,    0,  119,    0,    0, 1001, 1002, 1003, 
            0, 1004,   35, 1005, 1006, 1007, 1008,    0,   40,    0, 
            0,  128,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0, 1009,    0,    0,  131,    0,    0,   46,   47, 
            0,   48,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
           52,   53,   54,   55,   56,   57,    0,    0,    0,   58, 
            0,   59,   60,    0,   61,   62,   63,   64,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,   65,   66,  139, 
          140,  141,  142,  143,  144,  145,  146,  147,  148,  149, 
          150,  151,  152,  153,  154,  155,  156,  157,  158,  159, 
          160,  161,  162,    0,    0,  163,  164,  165,  223,  224, 
          225,  226,  170,  171,  172,  173,  174,  175,  176,  177, 
          178,  179,  180,  181,  227,  228,  229,    0,  230,  186, 
          310,    0,  231,    0,    0,    0,  188,  189,    0,  190, 
          191,  192,  193,  194,  195,  196,    0,    0,  197,  198, 
            0,    0,    0,  199,  200,  201,  202,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          204,  205,    0,  206,  207,  208,  209,  210,  211,  212, 
          213,  214,  215,  216,    0,    0,  217,   52,  139,  140, 
          141,  142,  143,  144,  145,  146,  147,  148,  149,  150, 
          151,  152,  153,  154,  155,  156,  157,  158,  159,  160, 
          161,  162,    0,    0,  163,  164,  165,  223,  224,  225, 
          226,  170,  171,  172,  173,  174,  175,  176,  177,  178, 
          179,  180,  181,  227,  228,  229,    0,  230,  186,    0, 
            0,  231,    0,    0,    0,  188,  189,    0,  190,  191, 
          192,  193,  194,  195,  196,    0,    0,  197,  198,    0, 
            0,    0,  199,  200,  201,  202,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  204, 
          205,    0,  206,  207,  208,  209,  210,  211,  212,  213, 
          214,  215,  216,    0,    0,  217,   52,  139,  140,  141, 
          142,  143,  144,  145,  146,  147,  148,  149,  150,  151, 
          152,  153,  154,  155,  156,  157,  158,  159,  160,  161, 
          162,    0,    0,  163,  164,  165,  166,  167,  168,  169, 
          170,  171,  172,  173,  174,  175,  176,  177,  178,  179, 
          180,  181,  182,  183,  184,    0,  185,  186,  124,  125, 
          187,  127,    0,    0,  188,  189,    0,  190,  191,  192, 
          193,  194,  195,  196,    0,    0,  197,  198,    0,    0, 
            0,  199,  200,  201,  202,    0,    0,    0,    0,    0, 
            0,  203,    0,    0,    0,    0,    0,    0,  204,  205, 
            0,  206,  207,  208,  209,  210,  211,  212,  213,  214, 
          215,  216,    0,    0,  217,  139,  140,  141,  142,  143, 
          144,  145,  146,  147,  148,  149,  150,  151,  152,  153, 
          154,  155,  156,  157,  158,  159,  160,  161,  162,    0, 
            0,  163,  164,  165,  223,  224,  225,  226,  170,  171, 
          172,  173,  174,  175,  176,  177,  178,  179,  180,  181, 
          227,  228,  229,    0,  230,  186,  360,  361,  231,  362, 
            0,    0,  188,  189,    0,  190,  191,  192,  193,  194, 
          195,  196,    0,    0,  197,  198,    0,    0,    0,  199, 
          200,  201,  202,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  204,  205,    0,  206, 
          207,  208,  209,  210,  211,  212,  213,  214,  215,  216, 
            0,    0,  217,  139,  140,  141,  142,  143,  144,  145, 
          146,  147,  148,  149,  150,  151,  152,  153,  154,  155, 
          156,  157,  158,  159,  160,  161,  162,    0,    0,  163, 
          164,  165,  223,  224,  225,  226,  170,  171,  172,  173, 
          174,  175,  176,  177,  178,  179,  180,  181,  227,  228, 
          229,    0,  230,  186,    0,    0,  231,    0,    0,    0, 
          188,  189,    0,  190,  191,  192,  193,  194,  195,  196, 
            0,    0,  197,  198,    0,    0,    0,  199,  200,  201, 
          202,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  204,  205,    0,  206,  207,  208, 
          209,  210,  211,  212,  213,  214,  215,  216,  731,  621, 
          217,    0,  732,    0,    0,    0,  188,  189,    0,  190, 
          191,  192,  193,  194,  195,  196,    0,    0,  197,  198, 
            0,    0,    0,  199,  200,  201,  202,    0,    0,    0, 
            0,    0,    0,  304,    0,    0,    0,    0,    0,    0, 
          204,  205,    0,  206,  207,  208,  209,  210,  211,  212, 
          213,  214,  215,  216,  734,  629,  217,    0,  735,    0, 
            0,    0,  188,  189,    0,  190,  191,  192,  193,  194, 
          195,  196,    0,    0,  197,  198,    0,    0,    0,  199, 
          200,  201,  202,    0,    0,    0,    0,    0,    0,  304, 
            0,    0,    0,    0,    0,    0,  204,  205,    0,  206, 
          207,  208,  209,  210,  211,  212,  213,  214,  215,  216, 
          919,  621,  217,    0,  920,    0,    0,    0,  188,  189, 
            0,  190,  191,  192,  193,  194,  195,  196,    0,    0, 
          197,  198,    0,    0,    0,  199,  200,  201,  202,    0, 
            0,    0,    0,    0,    0,  304,    0,    0,    0,    0, 
            0,    0,  204,  205,    0,  206,  207,  208,  209,  210, 
          211,  212,  213,  214,  215,  216,  922,  629,  217,    0, 
          923,    0,    0,    0,  188,  189,    0,  190,  191,  192, 
          193,  194,  195,  196,    0,    0,  197,  198,    0,    0, 
            0,  199,  200,  201,  202,    0,    0,    0,    0,    0, 
            0,  304,    0,    0,    0,    0,    0,    0,  204,  205, 
            0,  206,  207,  208,  209,  210,  211,  212,  213,  214, 
          215,  216, 1078,  621,  217,    0, 1079,    0,    0,    0, 
          188,  189,    0,  190,  191,  192,  193,  194,  195,  196, 
            0,    0,  197,  198,    0,    0,    0,  199,  200,  201, 
          202,    0,    0,    0,    0,    0,    0,  304,    0,    0, 
            0,    0,    0,    0,  204,  205,    0,  206,  207,  208, 
          209,  210,  211,  212,  213,  214,  215,  216, 1081,  621, 
          217,    0, 1082,    0,    0,    0,  188,  189,    0,  190, 
          191,  192,  193,  194,  195,  196,    0,    0,  197,  198, 
            0,    0,    0,  199,  200,  201,  202,    0,    0,    0, 
            0,    0,    0,  304,    0,    0,    0,    0,    0,    0, 
          204,  205,    0,  206,  207,  208,  209,  210,  211,  212, 
          213,  214,  215,  216, 1084,  629,  217,    0, 1085,    0, 
            0,    0,  188,  189,    0,  190,  191,  192,  193,  194, 
          195,  196,    0,    0,  197,  198,    0,    0,    0,  199, 
          200,  201,  202,    0,    0,    0,    0,    0,    0,  304, 
            0,    0,    0,    0,    0,    0,  204,  205,    0,  206, 
          207,  208,  209,  210,  211,  212,  213,  214,  215,  216, 
            0,    0,  217, 
      };
   }

   private static final short[] yyCheck1() {
      return new short[] {

           26,   74,   28,    2,    3,    4,    5,  100,    6,  468, 
            9,   10,   11,    3,    7,   14,   15,   16,   51,  296, 
           19,   27,   88,    6,    7,   91,   49,  508,   15,   16, 
           11,  258,   19,   10,   27,  448,   10,   10,   11,   15, 
           16,  488,    7,   19,   27,  580,   45,  473,  485,    0, 
           49,  488,   51,   10,   52,   21,   10,   10,  424,   10, 
           47,   48,   27,  335,  366,  431,  368,  339,   10,   52, 
            2,    3,   15,   16,   45,  418,   19,   10,  430,   10, 
          527,   10,   59,  673,   32,  372,    4,    5,   10,    0, 
          105,    0,  417,  119,  419,   59,   14,   10,  423,  689, 
           10,  100,   59,  446,   47,  130,   59,   10,   59,   10, 
           10,   44,  702,   10,  864,   10,   10,   21,   59,  462, 
           10,  105,  447,   10,  105,  772,   59,  758,   59,   10, 
          129,   49,  131,  476,   44,  980,  461,   59,  463,  280, 
          960,  107,   10,  733,   44,   10,   59,  113,   10,   59, 
          307,    0,  477,  683,  308,  558,   59,  322,   59,   59, 
          341,   10,  692,  341,   59,   59,   43,   10,  342,   59, 
           10,  137,  329,  330,   10,   10,   10,   10,   59,  342, 
          482,   10,  100,   10,  305,  359,  511,  362,  254,   10, 
          310,   59,  535,  312,   59,   10,  359,   59,  317,  282, 
           10,   44,  280,  107,   44,    2,    3,    4,    5,   44, 
           59,  536,   10,  305,  280,   44,   59,   14,   32,   59, 
          497,  362,  376,  377,   59,   61,   59, 1072,   10,  587, 
           59,  350,  351,    9,   10,   11,   10,   10,   59,  238, 
          887, 1061,   10,   32,  725,   10,  548,   10,   45,   59, 
          316,  128,   49,   10,  375,  668,   10,  424,  366,  426, 
          353,  619,  427,  341,  431,  678,  265,  266,   10,  716, 
           10,  558,   10, 1023,  677,  394,  679,  261,  715,  716, 
          261,  583,  345,  375,  362,   10,   10,   61,  354,  373, 
           44,  557,  265,  266,  378,  308,  309,   10,  311,   32, 
          363,   10,  301,  100,  303,  304,   44,   61,  307,  324, 
          309,  784,  328,  317,  301,  314,  303,  790,   10,  335, 
          336,  320,   10,  322,  373,  301,   32,  303,  304,  378, 
          682,  921,  331,  490,  131,   44,  967,   61,  290,  291, 
          324,  498,  499,  324,   10,  349,   59,  882,  306,  353, 
          354,  322,  310,  129,  353,   10,  891,  267,  301,  516, 
          303,  896,   32,  376,  377,   44,  267,  780,  342,   61, 
          264,  345,  267,   61,  264,  269,   45,  271,  281,  269, 
          379,  380,  381,  382,  383,  384,  385,  341,  320,   44, 
          677,  424,  679,  426,  807,  359,  264,  430,  431,  331, 
          342,  342,  359,  345,   59,  428,  379,  380,  381,  382, 
          363,  342,  363,  342,  840,  781,  782,  947,  395,  418, 
          445,  787,  788,  371,  372,  424,  359,  426,  427,  428, 
          378,  430,  431,  710,  433,  353,  419,  359,  339,  741, 
          423,  238,  418,  433,  537,  342,  359,  446,  345,  448, 
          449,  383,  363,  419,  363,  798,  427,  729,  345,  359, 
          459,  342,  131,  462,  447,  305,  768,  694,  467,  756, 
          446,  758,  797,  269,  799,  271,  305,  476,  744,  745, 
          463,  447,  537,  426,  305,  772,  462,  430,  947,  265, 
          266,  472,  473, 1083,  477,  305,  653,  463,  485,  505, 
          476,  433,  342,   10,  951,  279,  342,  506, 1043,  508, 
          428,  477,  505,  342,  951,  342,  359,  314,  345,  676, 
          519,  342,  505,  320,  359,  322,  469,  342,  511,    2, 
            3,  307,  342,  309,  331,  375,  535,  524,  537,  338, 
          505,   14,  908,   61,  342,  511,  375,   91,   10,    2, 
          263,  264,   59,  536,  375,  279,  353,  371,  372,  535, 
          342,  654,   10,  720,  378,  375,  756,  844,  342,  342, 
          536,  728,   45,   10,  342,  641,  508,  342,   61,  342, 
          571,  572,   44,  996,   91,  342,  383,  279,  342,   10, 
          616,  279,  749,   10,   47,   48,  862,   59,  864,  654, 
          342,  627,  342,  379,  380,  381,  382,   44,  384,  385, 
          338,  372,  267,   61,  781,  782,   10,  342,  342,  537, 
          787,  788,   59,  341,  290,  291,  325,  424,  357,  426, 
          427,  428,   10,  800,  431,   10,  433,  730,   59,  371, 
          372,  373,  322,  323,   61,  314,  378,  673,  268,  269, 
          342,  448,   44,  322,  342,  654,  813,  623,  131,  310, 
          359,  338,  459,  689,  396,  631,   10,   61,  667,  668, 
          467,  637,   61,  449,   10,  681,  702,  357,  358,  678, 
          967,  338,  680,   61,  339,  666,   61,  271,  681,  976, 
          956,  957,   59,  980,  667,    2,    3,  680,  681,  310, 
           44,  350,  351,  279,  861,  671,  705,  733,  674,  936, 
         1053,  508,  805,   15,   16,  942,  681,   19,  339,  623, 
           44,  794,  519,   44,  750,   61,  725,  631,  715, 1054, 
          506,  730,  705,  637,  328,  756,  654,  758,   45,  682, 
          537,  908,  336,  263,  264,  362,  772,  317,  781,  782, 
          805,  772,   44,  317,  787,  788,  338, 1023,  427, 1025, 
          267,   10,  263,  264,   44,  238,  345,  800,  269,   61, 
          342,  961,  363,  280,  310,  774,  862,   44,  777,   44, 
          347,  780,  781,  782,  328, 1072,  976,  341,  787,  788, 
          459,  335,  336,  725, 1060,   44,  795,  558,  467,  798, 
          341,  800,   44,  262,  264,  342,  805,  264,  807,  267, 
           59,   44,  730,  371,  372,   44,  799,  363,   44,  362, 
          378,  328,  798,  822,  131,  383,  825,  808,  335,  336, 
          306,  279,  306,  799,  815,   44,   91,  836,  837,  838, 
           44,  314,   91,  305,   44,   10,  267,  320,   44,  322, 
          306,  304,  825,   91,   44,  362,   61,  654,  331,  840, 
          363,  887,  279,  795,   44,  264,  338,  338,  305,  338, 
          956,  668,  871,  872,  338,  908,   61,  868, 1068,   44, 
          342,  678,  297,  874,  875,  279,  877,  805,  879,   44, 
           44,  667,  424,  264,  342,  921,   61,  359,  897,  431, 
          269,  279,   44,  338,  279,  342,  338,  338,  338,  908, 
          383,   91,  264,  375,  363,   32,  677,   44,  679,   32, 
           58,  345,  359,   44,  338,  342,   91,  363,  725,  705, 
          264,  238,  464,  730,  466,   10,  935,  264,  375, 1025, 
          939,  738,  339,  279, 1010,   91,  967,  263,  342,   91, 
          264,  424,  338,  426,  427,  976,  264,   44,  431,  980, 
          433,  371,  372,  373,  342,  418,   44,  342,  378,   10, 
           91,  371,  372,  373, 1060,  448,  363,  774,  378,  897, 
          777,  513,  342,  780,  781,  782,  459,  279,   59,  988, 
          787,  788,   44,  446,  467,   44,   44,  996,  795,  301, 
           44,  303, 1028,  800,   44,   44,  342,  314,  805,  462, 
          807,   44,   10,  320, 1013,  322, 1015,  935,   59,   44, 
           44,  359, 1021,  476,  331,  822,  306,   61,  308,  309, 
          310,  311,  264,  264,  264,  508,  264,   44,  363,  836, 
          837,  838, 1033, 1034, 1035, 1036,  519,  264,  359,  825, 
          310, 1072,  310,   44, 1053,  508,   44, 1083,   91, 1058, 
           44,   59,   10,   44,  306,   44,  308,  309,  310,  311, 
          988, 1054,   44,  328,  871,  872,  383, 1053,  264,  328, 
          335,  336,  535,  264,   44,  395,  335,  336, 1054,  271, 
          328,  345,  341,   91,  363,  341,   44,  335,  336,  279, 
          897, 1092,   10,  271,  345,  774,   10,  363,  777,  345, 
          359,  908,   58,   61,  279,    0,   10,  424,  342,  426, 
          427,  264,  345,   70,  431,   10,  433,  289,   10,   54, 
            5,  583,  585, 1056,  587,  519,    6,  279,  935,  637, 
           44,  448,  939,   91,  379,  380,  381,  382,  328,  934, 
          756,   59,  459,  822,  772,  335,  336,   61,  279,  280, 
          467,  341,   44,  328,  760,   59,  619,  836,  837,  838, 
          335,  336,   44,  978,   59,  976,  341,  342,  631,   61, 
          793, 1055,  328,   91,  637,   91,  328,   91,   14,  335, 
          336,  988,  744,  335,  336,  668,  664,   91, 1061,  996, 
           89,  508,  871,  872,   -1,  678,   -1,  328,   -1,   91, 
           -1,   -1,  519,   -1,  335,  336, 1013,   44, 1015,   91, 
          341,  674,   -1,  264, 1021,   -1,   -1,  268,  269,  682, 
          683,  306,  685,  308,  309,  310,  311,  312,   10,  692, 
           -1,  362,  317,   -1,   -1,   -1,  279,  280,   -1,  781, 
          782,   -1,  725,   -1,   -1,  787,  788,   -1,   91,   -1, 
           -1, 1058,   -1,   -1,   91,  340,   -1,   10,   91,  267, 
          939,   -1,  725,  348,  349,  350,  351,   -1,   -1,   10, 
           -1,   -1,    0,   -1,   -1,   15,   16,   59,   -1,   19, 
           -1,   10,   10,   -1,   -1,  328,  828,  829,   -1,  831, 
          832,  774,  335,  336,  777,   -1,   -1,  780,  781,  782, 
           -1,   41,   42,   -1,  787,  788,   59,   47,   48,  394, 
           50,   51,  795,   10,   -1,   44,   44,  800,   59,  362, 
          328,  279,  280,   -1,  807,   -1,   -1,  335,  336,   -1, 
           59,   59,   -1,  341, 1013,  798, 1015,   -1,   -1,  822, 
           91,   -1, 1021,   -1,   10,   44,   -1,   44,   -1,  267, 
           91,  668,   -1,  836,  837,  838,   10,  820,   -1,   -1, 
          823,  678,   91,  267,   61,  279,  280,  262,  263,  264, 
          328,   91,  267,  268,  269,   -1,  271,  335,  336, 1058, 
           -1,   -1,   -1,  341,  342,   -1,  281,  279,  871,  872, 
           44,   -1,   91,   59,   91,  290,  291,  279,  293,  294, 
          295,  296,  297,   -1,  362,   -1,   -1,   61,  725,   -1, 
          328,   -1,  328,   -1,  328,   -1,   -1,  335,  336,  335, 
          336,  335,  336,   -1,  328,  908,   -1,  341,  342,    4, 
            5,  335,  336,  328,   -1,   -1,  328,   91,   -1,   14, 
          335,  336,  279,  335,  336,   -1,  328,  342,  362,   -1, 
          342,   -1,  935,  335,  336,   -1,  939,  774,   -1,   10, 
          777,   -1,   -1,  780,  781,  782,   41,   42,  363,   -1, 
          787,  788,   47,   48,   49,   50,   -1,   -1,  795,   48, 
          262,  263,  264,  800,  947,  328,  268,  269,   -1,  271, 
          807,  328,  335,  336,   -1,  328,   -1,   -1,  335,  336, 
          395,   -1,  335,  336,   -1,  822,   -1,   -1,   59,  262, 
          263,  264,    0,  996,   -1,  268,  269,   -1,  271,  836, 
          837,  838,   10,   -1,   -1,  100,  267,   -1,   -1,   -1, 
         1013,   -1, 1015,   -1,  262,  263,  264,   -1, 1021,  267, 
          268,  269,   -1,  271,   -1,  285,  286,  287,  288,   -1, 
          290,  280,   -1,   -1,  871,  872,   -1,   -1,   -1,   -1, 
          342,  301,   -1,  303,  304,  293,  294,  295,  296,  297, 
           10,   59,  312,   -1,   -1, 1058,   -1,  328,   -1,   -1, 
          279,  363,  279,   -1,  335,  336,   -1,  328,   -1,  342, 
         1053,  908,   -1, 1056,  335,  336,  262,  263,  264,  328, 
           -1,   -1,   -1,  269,   44,   -1,  335,  336,  328,   -1, 
          363,  339,  341,  395,  342,  335,  336,  371,  372,  373, 
           -1,   61,  939,   -1,  378,  279,  280,   -1,   -1,  328, 
          359,  328,   -1,  362,   -1,  363,  335,  336,  335,  336, 
           -1,   -1,  395,   -1,   -1,  342,  386,  387,  388,  389, 
          390,   91,   -1,  393,  394,  395,  396,  397,  398,  399, 
          400,  401,  402,  403,  404,  405,  406,  395,   -1,  409, 
          410,  411,  412,   44,  328,   -1,   -1,   -1,  418,  996, 
          420,  335,  336,   -1,  424,   -1,  426,  341,  342,   -1, 
          430,  431,   -1,   -1,   -1,   -1, 1013,  306, 1015,  308, 
          309,  310,  311,   -1, 1021,   -1,  446,   -1,  362,   -1, 
          285,  286,  287,  288,   -1,  290,  285,  286,  287,   -1, 
           91,  290,  462,    0,  464,  465,  466,  753,   -1,  469, 
           -1,  340,  758,   10,   -1,   -1,  476,  312,   -1,   44, 
           -1, 1058,   -1,   -1,  484,  485,  772,   -1,  488,  371, 
          372,  373,   -1,   -1,   -1,  306,  378,  308,  309,  310, 
          311,  312,   -1,  371,  372,  373,  317,   44,   -1,   -1, 
          378,   -1,  512,  513,  262,  263,  264,  328,  353,  267, 
          268,  269,   59,  271,  524,  336,   91,  527,   -1,  340, 
          341,   -1,   -1,  281,   -1,  535,   -1,  348,  349,  350, 
          351,   -1,  290,  291,   10,  293,  294,  295,  296,  297, 
           -1,  386,  387,  388,  389,  390,   -1,   -1,  393,  394, 
          395,  396,  397,  398,  399,  400,  401,  402,  403,  404, 
          405,  406,   44,   10,  409,  410,  411,  412,   44,  279, 
          280,  581,   -1,  394,   10,  420,   10,   -1,   10,  371, 
          372,  373,   -1,  428,  342,   61,  378,  426,   44,   -1, 
           -1,  430,   -1,   10,   -1,   -1,   -1,   44,   -1,  306, 
           -1,  308,  309,  310,  311,  363,   -1,   -1,   44,   91, 
           44,   -1,   44,   -1,   61,   91,   -1,   -1,  328,  464, 
          465,  466,   -1,   59,  469,  335,  336,   61,   -1,   61, 
          469,  341,  342,  340,   -1,   91,  646,  395,  279,  484, 
          485,  348,   59,  488,   91,  484,  485,   -1,   -1,  488, 
           -1,   -1,  362,   -1,   -1,   91,   10,   91,  306,   91, 
          308,  309,  310,  311,   -1,   -1,   -1,  512,  513,  965, 
           -1,  967,  682,  969,   91,   -1,   -1,  973,   -1,  524, 
           -1,   -1,  527,   -1,  980,  524,   -1,  328,  527,   -1, 
           44,   -1,  537,   -1,  335,  336,  706,   -1,  708,  293, 
          294,  295,  296,  297,  279,  715,  716,   61,    0,   -1, 
           -1,   -1,   -1,   -1,   -1,  262,  263,  264,   10,  760, 
          267,  268,  269,  306,  271,  308,  309,  310,  311,   -1, 
           -1,  772,   -1,   44,  281,  282,  581,   91,   -1,   -1, 
           -1,   -1,  581,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,    0,  328,   -1,   -1,  766,  340,   -1,   -1, 
          335,  336,   10,   -1,   -1,  348,   -1,   59, 1064,   -1, 
           -1,  781,  782,   44, 1070,   10, 1072,  787,  788, 1075, 
           91,   -1,   -1,   44,   -1,   -1,  796,   -1,  798,   -1, 
          800,   -1,  802,   -1,   -1,  342,   -1,  279,  345,   -1, 
          347,  646,   10,  279,  280, 1101,   -1,  646,   -1,  654, 
           -1,   59,   -1,   -1,   -1,   -1,  363,   -1,  828,  829, 
           91,  831,  832,  279,   59,  835,   -1,   -1,   -1,   -1, 
           91,   -1,  279,  280,   -1,   -1,   -1,  682,   -1,   -1, 
           -1,   -1,   -1,  682,  280,  279,  328,  279,  395,   -1, 
          267,   59,  328,  335,  336,   -1,   91,   -1,   -1,  335, 
          336,  706,   -1,  708,   -1,  341,  342,  706,   -1,  708, 
          715,  716,  328,   -1,   -1,   -1,  715,  716,   -1,  335, 
          336,  328,   -1,   91,   -1,  730,  362,    0,  335,  336, 
           -1,  901,  328,   -1,  328,  342,  328,   10,  908,  335, 
          336,  335,  336,  335,  336,  341,   -1,   -1,  342,   -1, 
          342,  328,   -1,   -1,  924,  362,   -1,  317,  335,  336, 
           -1,  766,   -1,  359,  888,  279,  362,  766,  892,   -1, 
           -1,   44,  973,   -1,  944,  976,   -1,  978,   -1,  980, 
           -1,  951,   -1,   -1,   -1,   58,   59,   -1,   61,  349, 
           63,  796,   -1,  353,  354,  355,  356,  802,   -1,   -1, 
          805,   -1,   -1,  802,   -1,   -1,   -1,   -1,  279,  280, 
          262,  263,  264,   -1,  328,  267,  268,  269,   91,  271, 
           -1,  335,  336,  828,  829,   -1,  831,  832,  342,  281, 
          835,   -1,   -1,   -1,   -1,   -1,  835,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,  279,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,  328,  279,  267, 
          268,  269,   -1,  271,  335,  336,   -1, 1068,    0, 1070, 
          341, 1072,  267,  281, 1075,   -1,   -1,   -1,   10,   -1, 
           -1,   -1,   -1, 1053,   -1,  293,  294,  295,  296,  297, 
          342,  362,  897,   -1,   -1,   -1,  901,  328,   -1,  267, 
         1101,   -1,   -1,   -1,  335,  336,   -1,  328,   -1,   -1, 
           -1,  363,   44, 1037,  335,  336, 1040, 1041,   -1,  924, 
         1044, 1045,   -1,   -1,   -1,  924,   58,   59,   -1,   61, 
          935,   63,   -1,  328,  342,   -1,   -1,   -1,   -1,  944, 
          335,  336,   -1,  395,   -1,   -1,  951,   -1,   -1,   -1, 
           -1,   -1,  951,   -1,   -1,  363,   -1,   10,   -1,   91, 
          328,   -1,   -1,   -1,   -1,   -1,   -1,  335,  336, 1093, 
         1094, 1095, 1096,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1, 1105,   -1,  988,  257,  258,  259,  395,  261,  262, 
          263,  264,  265,  266,  267,  268,  269,  270,  271,  272, 
          273,  274,  275,  276,  277,  278,   59,  280,  281,  282, 
          283,  284,  285,  286,  287,  288,  289,  290,  291,  292, 
          293,  294,  295,  296,  297,   -1,  299,   -1,   -1,  302, 
          303,  304,  305,  306,  307,  308,  309,  310,  311,  312, 
          313,  314,  315,  316,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   10,   -1, 
          333,  334,  335,  336,  337,  338,  339,  340,  341,  342, 
          343,  344,  345,  346,  347,  348,  349,  350,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,  362, 
          363,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           44,  374,  375,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   63,   -1,   -1,   -1,  772,   -1,   -1,   -1,  392, 
          393,  394,  395,   -1,   -1,  257,  258,  259,   -1,  261, 
          262,  263,  264,  265,  266,  267,  268,  269,  270,  271, 
          272,  273,  274,  275,  276,  277,  278,   91,  280,  281, 
          282,  283,  284,  285,  286,  287,  288,  289,  290,  291, 
          292,  293,  294,  295,  296,  297,   -1,  299,   -1,   -1, 
          302,  303,  304,  305,  306,  307,  308,  309,  310,  311, 
          312,  313,  314,  315,  316,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   44, 
           -1,  333,  334,  335,  336,  337,  338,  339,  340,  341, 
          342,  343,  344,  345,  346,  347,  348,  349,  350,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
          362,  363,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,    0,  374,  375,  376,  377,   91,  379,  380,  381, 
          382,   10,   -1,  306,  317,  308,  309,  310,  311,  312, 
          392,  393,  394,  395,  317,   -1,   -1,   -1,   -1,   -1, 
          333,  334,   -1,   -1,   -1,  328,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  336,   -1,   44,  349,  340,  341,   -1, 
          353,  354,  355,  356,   -1,  348,  349,  350,  351,   58, 
           59,   -1,   61,   -1,   63,   -1,  306,   -1,  308,  309, 
          310,  311,  312,   -1,   -1,   -1,  973,  317,   -1,  976, 
           -1,  978,   -1,  980,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   91,   -1,   -1,  279,  262,  263,  264,   -1, 
          340,  394,  268,  269,   -1,  271,  262,  263,  264,   -1, 
          350,  351,  268,  269,   -1,  271,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,   -1,  329,  330,   -1, 
            0,  333,  334,   -1,  328,   -1,   -1,   -1,   -1,   -1, 
           10,  335,  336,   -1,  394,   -1,   -1,  349,   -1,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
           -1, 1068,   -1, 1070,   -1, 1072,  342,   -1, 1075,   -1, 
           -1,   -1,   -1,   -1,   44,   -1,  342,   -1,   -1,  317, 
           -1,   -1,   -1,   -1,  279,   -1,   -1,  363,   58,   59, 
           -1,   61,   -1,   63, 1101,  333,  334,  363,   -1,   -1, 
           -1,  306,   -1,  308,  309,  310,  311,  312,   -1,   -1, 
           -1,  349,  317,  351,   -1,  353,  354,  355,  356,  395, 
           -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395, 
           -1,   -1,   -1,  328,   -1,  340,  341,   -1,   -1,   -1, 
          335,  336,   10,  348,  349,  350,  351,   -1,  257,  258, 
          259,   -1,  261,  262,  263,  264,  265,  266,  267,  268, 
          269,  270,  271,  272,  273,  274,  275,  276,  277,  278, 
           -1,  280,  281,  282,  283,  284,  285,  286,  287,  288, 
          289,  290,  291,  292,  293,  294,  295,  296,  297,  394, 
          299,   59,   -1,  302,  303,  304,  305,  306,  307,  308, 
          309,  310,  311,  312,  313,  314,  315,  316,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   44,   -1,  333,  334,  335,  336,  337,   10, 
          339,  340,  341,  342,  343,  344,  345,  346,  347,  348, 
          349,  350,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,  362,  363,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,  375,  376,  377,   91, 
          379,  380,  381,  382,   -1,   -1,   -1,   -1,   59,   -1, 
           -1,   10,   -1,  392,  393,  394,  395,  257,  258,  259, 
           -1,  261,  262,  263,  264,  265,  266,  267,  268,  269, 
          270,  271,  272,  273,  274,  275,  276,  277,  278,   -1, 
          280,  281,  282,  283,  284,  285,  286,  287,  288,  289, 
          290,  291,  292,  293,  294,  295,  296,  297,   -1,  299, 
           59,   -1,  302,  303,  304,  305,  306,  307,  308,  309, 
          310,  311,  312,  313,  314,  315,  316,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   91,  333,  334,  335,  336,  337,   -1,  339, 
          340,  341,  342,  343,  344,  345,  346,  347,  348,  349, 
          350,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,  362,  363,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,    0,  374,  375,  376,  377,   -1,  379, 
          380,  381,  382,   10,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  392,  393,  394,  395,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  306,   -1, 
          308,  309,  310,  311,  312,   -1,   -1,   44,   -1,  317, 
           -1,   -1,   -1,  317,   -1,   -1,   -1,  279,  280,   -1, 
          328,   58,   59,   -1,   61,   -1,   63,   -1,  336,  333, 
          334,   -1,  340,  341,   -1,   -1,   -1,   -1,   -1,   -1, 
          348,  349,  350,  351,   -1,  349,   -1,  351,   -1,  353, 
          354,  355,  356,   -1,   91,  359,   -1,  361,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           -1,   -1,   -1,  335,  336,   -1,   -1,   -1,   -1,  341, 
           -1,   -1,   -1,   -1,   -1,  306,  394,  308,  309,  310, 
          311,  312,   -1,   -1,   -1,   -1,  317,   -1,  267,   -1, 
          362,   -1,    0,   -1,   -1,   -1,   -1,  328,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,  336,   -1,   -1,   -1,  340, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350, 
          351,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,   -1,  328, 
           58,   59,   -1,  333,  334,   63,  335,  336,   -1,   -1, 
           -1,   -1,   -1,  394,   -1,   -1,   -1,   -1,   -1,  349, 
           -1,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  262,  263,  264,  265,  266, 
          267,  268,  269,  270,  271,  272,  273,  274,  275,  276, 
          277,  278,   -1,  280,  281,  282,  283,  284,  285,  286, 
          287,  288,  289,  290,  291,  292,  293,  294,  295,  296, 
          297,   -1,  299,   59,   -1,  302,  303,  304,  305,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   44,   -1,  333,  334,  335,  336, 
          337,   10,  339,  340,  341,  342,  343,  344,  345,  346, 
          347,  348,  349,  350,  351,  352,  353,  354,  355,  356, 
          357,  358,  359,  360,  361,  362,  363,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,  375,  376, 
          377,   91,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           59,   -1,   -1,   10,   -1,  392,  393,  394,  395,  257, 
          258,  259,   -1,  261,  262,  263,  264,  265,  266,  267, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,   -1,  280,  281,  282,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
           -1,  299,   59,   -1,  302,  303,  304,  305,  306,  307, 
          308,  309,  310,  311,  312,  313,  314,  315,  316,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   91,  333,  334,  335,  336,  337, 
           -1,  339,  340,  341,  342,  343,  344,  345,  346,  347, 
          348,  349,  350,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  362,  363,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,    0,  374,  375,  376,  377, 
           -1,  379,  380,  381,  382,   10,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  392,  393,  394,  395,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          306,   -1,  308,  309,  310,  311,  312,   -1,   -1,   44, 
           -1,  317,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  279, 
          280,  317,  328,   58,   59,   -1,  322,  323,   63,   -1, 
          336,   -1,   -1,   -1,  340,  341,   -1,  333,  334,   -1, 
           -1,   -1,  348,  349,  350,  351,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  349,   -1,  351,   91,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   -1,   -1,  328,   -1, 
           -1,   -1,   -1,   -1,   -1,  335,  336,   -1,   -1,   -1, 
           -1,  341,   -1,   -1,   -1,   -1,   -1,  306,  394,  308, 
          309,  310,  311,  312,   -1,   -1,   -1,   -1,  317,   -1, 
          267,   -1,  362,   -1,    0,   -1,   -1,   -1,   -1,  328, 
           -1,   -1,   -1,   -1,   10,   -1,   -1,  336,   -1,   -1, 
           -1,  340,  341,   -1,   -1,   -1,   -1,   -1,   -1,  348, 
          349,  350,  351,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          317,  318,  319,  320,  321,  322,  323,  324,   44,  326, 
          327,   -1,   -1,   -1,   -1,   -1,  333,  334,   -1,   -1, 
           -1,  328,   58,   59,   -1,   -1,   -1,   63,  335,  336, 
           -1,   -1,  349,   -1,  351,  394,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1, 
           -1,   -1,  257,  258,  259,   -1,  261,  262,  263,  264, 
          265,  266,  267,  268,  269,  270,  271,  272,  273,  274, 
          275,  276,  277,  278,   -1,  280,  281,  282,  283,  284, 
          285,  286,  287,  288,  289,  290,  291,  292,  293,  294, 
          295,  296,  297,   -1,  299,   59,   -1,  302,  303,  304, 
          305,  306,  307,  308,  309,  310,  311,  312,  313,  314, 
          315,  316,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   44,   -1,  333,  334, 
          335,  336,  337,   10,  339,  340,  341,  342,  343,  344, 
          345,  346,  347,  348,  349,  350,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,  362,  363,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
          375,  376,  377,   91,  379,  380,  381,  382,   -1,   -1, 
           -1,   -1,   59,   -1,   -1,   10,   -1,  392,  393,  394, 
          395,  257,  258,  259,   -1,  261,  262,  263,  264,  265, 
          266,  267,  268,  269,  270,  271,  272,  273,  274,  275, 
          276,  277,  278,   -1,  280,  281,  282,  283,  284,  285, 
          286,  287,  288,  289,  290,  291,  292,  293,  294,  295, 
          296,  297,   -1,  299,   59,   -1,  302,  303,  304,  305, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   91,  333,  334,  335, 
          336,  337,   -1,  339,  340,  341,  342,  343,  344,  345, 
          346,  347,  348,  349,  350,  351,  352,  353,  354,  355, 
          356,  357,  358,  359,  360,  361,  362,  363,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,    0,  374,  375, 
          376,  377,   -1,  379,  380,  381,  382,   10,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  394,  395, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  306,   -1,  308,  309,  310,  311,  312,   -1, 
           -1,   44,   -1,  317,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  279,  280,   -1,  328,   58,   59,   -1,   61,   -1, 
           63,   -1,  336,   -1,   -1,   -1,  340,  341,   -1,   -1, 
           -1,   -1,   -1,   -1,  348,  349,  350,  351,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          328,   -1,   -1,   -1,   -1,   -1,   -1,  335,  336,   -1, 
           -1,   -1,   -1,  341,   -1,   -1,   -1,   -1,   -1,  306, 
          394,  308,  309,  310,  311,  312,   -1,   -1,   -1,   -1, 
          317,   -1,  267,   -1,  362,   -1,    0,   -1,   -1,   -1, 
           -1,  328,   -1,   -1,   -1,   -1,   10,   -1,   -1,  336, 
           -1,   -1,   -1,  340,  341,   -1,   -1,   -1,   -1,   -1, 
           -1,  348,  349,  350,  351,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  317,  318,  319,  320,  321,  322,  323,   -1, 
           44,  326,  327,   -1,   -1,   -1,   -1,   -1,  333,  334, 
           -1,   -1,   -1,  328,   -1,   59,   -1,   -1,   -1,   63, 
          335,  336,   -1,   -1,  349,   -1,  351,  394,  353,  354, 
          355,  356,  357,  358,  359,   -1,  361,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1, 
           -1,   -1,   -1,   -1,  257,  258,  259,   -1,  261,  262, 
          263,  264,  265,  266,  267,  268,  269,  270,  271,  272, 
          273,  274,  275,  276,  277,  278,   -1,   -1,  281,  282, 
          283,  284,  285,  286,  287,  288,  289,  290,  291,  292, 
          293,  294,  295,  296,  297,   -1,  299,   59,   -1,  302, 
          303,  304,  305,  306,  307,  308,  309,  310,  311,  312, 
          313,  314,  315,  316,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,   -1, 
          333,  334,  335,  336,  337,   10,  339,  340,  341,  342, 
          343,  344,  345,  346,  347,  348,  349,  350,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,   63, 
          363,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,  375,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,   -1,   -1,   59,   -1,   -1,   10,   -1,  392, 
          393,  394,  395,  257,  258,  259,   -1,  261,  262,  263, 
          264,  265,  266,  267,  268,  269,  270,  271,  272,  273, 
          274,  275,  276,  277,  278,   -1,  280,  281,  282,  283, 
          284,  285,  286,  287,  288,  289,  290,  291,  292,  293, 
          294,  295,  296,  297,   -1,  299,   59,   -1,  302,  303, 
          304,   -1,  306,  307,  308,  309,  310,  311,  312,  313, 
          314,  315,  316,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,   63,   91,  333, 
          334,  335,  336,  337,   -1,  339,  340,   -1,  342,  343, 
          344,  345,  346,  347,  348,  349,  350,  351,  352,  353, 
          354,  355,  356,  357,  358,  359,  360,  361,  362,  363, 
           -1,  365,  366,  367,  368,  369,  370,   -1,   -1,    0, 
          374,   -1,  376,  377,   -1,  379,  380,  381,  382,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  392,  393, 
          394,  395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  306,   -1,  308,  309,  310,  311, 
          312,   -1,   -1,   44,   -1,  317,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   59,   -1, 
           -1,   -1,   63,   -1,  336,   -1,   -1,   -1,  340,  341, 
           -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351, 
           -1,   -1,   -1,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,   -1,  329,  330,   -1,   -1,  333, 
          334,  306,  394,  308,  309,  310,  311,  312,   -1,   -1, 
           -1,   -1,  317,   -1,  267,  349,   -1,  351,    0,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,   10,   -1, 
           -1,   -1,   -1,   -1,   -1,  340,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  348,  349,  350,  351,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  328,   -1,   59,   -1,   -1, 
           -1,   63,  335,  336,   -1,   -1,   -1,   -1,   -1,  394, 
           -1,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,   -1,  329,  330,   -1,   -1,  333,  334,   91, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  349,   -1,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,  257,  258,  259,   -1, 
          261,  262,  263,  264,  265,  266,  267,  268,  269,  270, 
          271,  272,  273,  274,  275,  276,  277,  278,   -1,  280, 
          281,  282,  283,  284,  285,  286,  287,  288,  289,  290, 
          291,  292,  293,  294,  295,  296,  297,   -1,  299,   -1, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,  312,  313,  314,  315,  316,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
            0,   -1,  333,  334,  335,  336,  337,   10,  339,  340, 
           10,  342,  343,  344,  345,  346,  347,  348,  349,  350, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,  362,  363,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   44,   -1,  374,   44,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,   -1,   59,   -1,   58,   59, 
           -1,  392,  393,  394,  395,  257,  258,  259,   -1,  261, 
          262,  263,  264,  265,  266,  267,  268,  269,  270,  271, 
          272,  273,  274,  275,  276,  277,  278,   -1,   91,  281, 
          282,  283,  284,  285,  286,  287,  288,  289,  290,  291, 
          292,  293,  294,  295,  296,  297,   -1,  299,   -1,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
          312,  313,  314,  315,  316,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   -1, 
           -1,  333,  334,  335,  336,  337,   -1,  339,  340,   -1, 
          342,  343,  344,  345,  346,  347,  348,  349,  350,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
            0,  363,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           10,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,  394,  395,  306,   -1,  308,  309,  310,  311, 
          312,   -1,   -1,   -1,   44,  317,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59, 
           -1,   61,   -1,   63,   -1,   -1,   -1,   -1,  340,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   91,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,   -1,   -1,  280,   -1,   -1, 
           -1,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          290,  291,  394,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  305,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1,   -1, 
           -1,   -1,  335,  336,   -1,   -1,   -1,   -1,  341,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1, 
           -1,   -1,   -1,   44,   -1,   -1,  359,   -1,   -1,  362, 
           -1,   -1,   -1,  363,   -1,  306,  307,   -1,   59,  310, 
           61,   -1,   63,  314,  315,  375,  317,  318,  319,  320, 
          321,  322,  323,   -1,   -1,  326,  327,   -1,   -1,   -1, 
          331,  332,  333,  334,   -1,  395,   -1,   -1,   -1,   -1, 
           91,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,   -1,   -1,  364,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,  262,  263,  264,  265,  266,   -1,  268,  269, 
          270,  271,  272,  273,  274,  275,  276,  277,  278,   -1, 
          280,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
          290,  291,  292,  293,  294,  295,  296,  297,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,  312,  313,  314,  315,  316,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,  337,  338,   10, 
          340,  341,  342,  343,  344,   -1,  346,   -1,  348,  349, 
          350,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,  362,  363,   -1,  365,  366,  367,  368,  369, 
          370,   -1,  760,   -1,  374,  763,  376,  377,   -1,  379, 
          380,  381,  382,   -1,  772,   -1,   -1,   -1,   59,   -1, 
           10,   -1,  392,  393,  394,  395,  257,  258,  259,   -1, 
          261,  262,  263,  264,  265,  266,   -1,  268,  269,  270, 
          271,  272,  273,  274,  275,  276,  277,  278,   -1,  280, 
           91,   -1,  283,  284,  285,  286,  287,  288,  289,  290, 
          291,  292,  293,  294,  295,  296,  297,   -1,  299,   59, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,  312,  313,  314,  315,  316,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           -1,   91,  333,  334,  335,  336,  337,  338,   -1,  340, 
          341,  342,  343,  344,   -1,  346,   -1,  348,  349,  350, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,  362,  363,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,    0,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  392,  393,  394,  395,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  973,   -1,   -1,  976,   -1, 
          978,   -1,  980,   -1,   -1,   -1,  267,   -1,   -1,   -1, 
           -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,  280, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  267,   -1,   -1, 
           -1,    0,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1, 
           -1,   10,   -1,   -1,  335,  336,   -1,   -1,   -1,   -1, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
         1068,   -1, 1070,   -1, 1072,   -1,   -1, 1075,   -1,   -1, 
           -1,  362,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  328,   -1, 
           59,   -1,   61, 1101,   63,  335,  336,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257, 
          258,  259,   -1,  261,  262,  263,  264,  265,  266,   -1, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,   -1,  280,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,  312,  313,  314,  315,  316,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,  337, 
           -1,   10,  340,  341,  342,  343,  344,   -1,  346,   -1, 
          348,  349,  350,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  362,  363,   -1,  365,  366,  367, 
          368,  369,  370, 
      };
   }

   private static final short[] yyCheck2() {
      return new short[] {

           -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,   -1,   -1,   59,   -1,   10, 
           -1,  392,  393,  394,  395,  257,  258,  259,   -1,  261, 
          262,  263,  264,  265,  266,   -1,  268,  269,  270,  271, 
          272,  273,  274,  275,  276,  277,  278,   -1,  280,   91, 
           -1,  283,  284,  285,  286,  287,  288,  289,  290,  291, 
          292,  293,  294,  295,  296,  297,   -1,  299,   59,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
          312,  313,  314,  315,  316,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   -1, 
           91,  333,  334,  335,  336,  337,   -1,   -1,  340,  341, 
          342,  343,  344,   -1,  346,   -1,  348,  349,  350,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
          362,  363,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,    0,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,  394,  395,   -1,   -1,  306,  307,   -1,   -1, 
          310,   -1,   -1,   -1,  314,  315,   -1,  317,  318,  319, 
          320,  321,  322,  323,   -1,   44,  326,  327,   -1,   -1, 
           -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1, 
           59,  341,   61,   -1,   63,   -1,   -1,   -1,  348,  349, 
           -1,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,   -1,   -1,  364,  267,   -1,   -1,   -1,   -1, 
           -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,  280,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  267,   -1,   -1,   -1, 
            0,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           10,   -1,   -1,  335,  336,   -1,   -1,   -1,   -1,  341, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          362,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   59, 
           -1,   61,   -1,   63,  335,  336,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,  258, 
          259,   -1,  261,  262,  263,  264,  265,  266,   -1,  268, 
          269,  270,  271,  272,  273,  274,  275,  276,  277,  278, 
           -1,  280,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,  290,  291,  292,  293,  294,  295,  296,  297,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,  312,  313,  314,  315,  316,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,  337,   -1, 
           10,  340,  341,  342,  343,  344,   -1,  346,   -1,  348, 
          349,  350,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,  362,  363,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1,   59, 
           -1,   -1,   -1,  392,  393,  394,  395,  257,  258,  259, 
           -1,  261,  262,  263,  264,  265,  266,   -1,  268,  269, 
          270,  271,  272,  273,  274,  275,  276,  277,  278,   -1, 
          280,   91,   -1,  283,  284,  285,  286,  287,  288,  289, 
          290,  291,  292,  293,  294,  295,  296,  297,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,  312,  313,  314,  315,  316,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,  337,   -1,   -1, 
          340,  341,  342,  343,  344,   -1,  346,   -1,  348,  349, 
          350,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,  362,  363,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,    0,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   10,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  392,  393,  394,  395,   -1,   -1,  306,  307, 
           -1,   -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317, 
          318,  319,  320,  321,  322,  323,   -1,   44,  326,  327, 
           -1,   -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1, 
           -1,   -1,   59,  341,   61,   -1,   63,   -1,   -1,   -1, 
          348,  349,   -1,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,   -1,   -1,  364,  267,   -1,   -1, 
           -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1, 
          280,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,  328,   -1, 
           -1,   -1,   10,   -1,   -1,  335,  336,   -1,   -1,   -1, 
           -1,  341,   -1,   -1,  306,  307,   -1,   -1,  310,   -1, 
           -1,   -1,  314,  315,   -1,  317,  318,  319,  320,  321, 
          322,  323,  362,   -1,  326,  327,   44,   -1,   -1,  331, 
          332,  333,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341, 
           -1,   59,   -1,   61,   -1,   63,  348,  349,   -1,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
           -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  262,  263,  264,  265,  266, 
           -1,  268,  269,  270,  271,  272,  273,  274,  275,  276, 
          277,  278,   -1,  280,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,  290,  291,  292,  293,  294,  295,  296, 
          297,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
          337,   -1,   10,  340,  341,  342,  343,  344,   -1,  346, 
           -1,  348,  349,  350,  351,  352,  353,  354,  355,  356, 
          357,  358,  359,  360,  361,  362,  363,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           -1,   59,   -1,   -1,   -1,  392,  393,  394,  395,  257, 
          258,  259,   -1,  261,  262,  263,  264,  265,  266,   -1, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,   -1,  280,   91,   -1,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,  312,  313,  314,  315,  316,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,  337, 
           -1,   -1,  340,  341,  342,  343,  344,   -1,  346,   -1, 
          348,  349,  350,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  362,  363,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,    0,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   10,   -1,   -1,   10,   -1, 
           -1,   -1,   -1,   -1,  392,  393,  394,  395,   -1,   -1, 
          306,  307,   -1,   -1,  310,   -1,   -1,   -1,  314,  315, 
           -1,  317,  318,  319,  320,  321,  322,  323,   -1,   44, 
          326,  327,   44,   -1,   -1,  331,  332,  333,  334,   -1, 
           -1,   -1,   -1,   -1,   59,  341,   61,   -1,   63,   61, 
           -1,   -1,  348,  349,   -1,  351,  352,  353,  354,  355, 
          356,  357,  358,  359,  360,  361,   -1,   -1,  364,  267, 
           -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,   91, 
           -1,   -1,  280,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1, 
          328,   -1,   -1,   -1,   10,   -1,   -1,  335,  336,   -1, 
           -1,   -1,   -1,  341,   -1,   -1,  306,  307,   -1,   -1, 
          310,   -1,   -1,   -1,  314,  315,   -1,  317,  318,  319, 
          320,  321,  322,  323,  362,   -1,  326,  327,   44,   -1, 
           -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1, 
           -1,  341,   -1,   59,   -1,   61,   -1,   63,  348,  349, 
           -1,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,   -1,   -1,  364,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  257,  258,  259,   -1,  261,  262,  263,  264, 
          265,  266,   -1,  268,  269,  270,  271,  272,  273,  274, 
          275,  276,  277,  278,   -1,   -1,   -1,  279,  283,  284, 
          285,  286,  287,  288,  289,  290,  291,  292,  293,  294, 
          295,  296,  297,   -1,  299,   -1,   -1,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,  312,  313,  314, 
          315,  316,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,  328,   -1,  333,  334, 
          335,  336,  337,  335,  336,  340,  341,  342,  343,  344, 
          342,  346,   -1,  348,  349,  350,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,    0,  363,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   10,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  394, 
          395,  257,  258,  259,   -1,  261,  262,  263,  264,  265, 
          266,   -1,  268,  269,  270,  271,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   59,  283,  284,  285, 
          286,  287,  288,  289,  290,  291,  292,  293,  294,  295, 
          296,  297,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   -1,  333,  334,  335, 
          336,  337,   -1,   -1,  340,  341,  342,  343,  344,   -1, 
          346,   -1,  348,  349,  350,  351,  352,  353,  354,  355, 
          356,  357,  358,  359,  360,  361,    0,  363,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   10,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  394,  395, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           44,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   58,   59,   -1,   61,   -1,   63, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  257,  258,  259,   -1,  261,  262, 
          263,  264,  265,  266,  267,  268,  269,  270,  271,  272, 
          273,  274,  275,  276,  277,  278,   -1,  280,  281,   -1, 
          283,  284,  285,  286,  287,  288,  289,  290,  291,  292, 
          293,  294,  295,  296,  297,   59,  299,   -1,   -1,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,  312, 
          313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  328,   -1,   91,   -1,   -1, 
           -1,   -1,  335,  336,  337,   -1,   -1,  340,   -1,  342, 
          343,  344,   -1,  346,   -1,  348,   -1,  350,   -1,  352, 
           -1,    0,   -1,   -1,   -1,   -1,   -1,  360,   -1,  362, 
          363,   10,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  392, 
          393,  394,  395,   -1,   -1,   44,   -1,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,   58, 
           59,   -1,   -1,   -1,   63,   -1,  280,  281,  282,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,   -1,   -1,  333, 
          334,  335,  336,   -1,    0,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   10,  349,   -1,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,  362,  363, 
           -1,   -1,   -1,  267,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   -1,   -1,  280,   -1,   44,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  395,   58,   59,   -1,   61,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  328,   91,   -1,   -1,   -1,   -1, 
           -1,  335,  336,   -1,   -1,   -1,   -1,  341,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1, 
           -1,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  280,  281,  282,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   58, 
           59,   -1,   61,   -1,   63,   -1,  305,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   91,   -1,  333,  334,  335,  336,   -1,   -1, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
          349,   10,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   -1,  361,  362,  363,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   44,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,  395,   -1,   -1,   -1, 
           -1,   -1,   61,   -1,   -1,  281,  282,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305, 
            0,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   -1,  333,  334,  335, 
          336,   -1,  338,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   -1,  349,   44,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   -1,  363,   58,   59, 
           -1,   61,   -1,   63,   -1,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,  395, 
           -1,   91,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,    0,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,   -1,  338, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
          349,   44,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   -1,  361,   -1,  363,   58,   59,   -1,   61,   -1, 
           63,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,   -1, 
          279,   -1,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   91,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  328, 
            0,   -1,   -1,   61,   -1,   -1,  335,  336,   -1,   -1, 
           10,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  281,  282,   -1,   44,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   58,   59, 
           -1,   61,   -1,   63,   -1,  305,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   91,   -1,  333,  334,  335,  336,   -1,  338,  339, 
           -1,   -1,  342,   10,   -1,  345,   -1,  347,   -1,  349, 
           -1,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
            0,  361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,   -1,   -1,   -1,   -1,  375,   -1,   44,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,   -1, 
           -1,   -1,   -1,   -1,   61,  395,   -1,   -1,  281,  282, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   -1,   59, 
           -1,   -1,  305,    0,   91,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,   -1, 
          333,  334,  335,  336,   -1,  338,  339,   -1,   -1,  342, 
           -1,  279,  345,   -1,  347,   -1,  349,   44,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,   -1,  361,   -1, 
          363,   58,   59,   -1,   61,   -1,   63,   -1,   -1,   -1, 
           -1,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
          328,  271,  395,   -1,   91,   -1,   -1,  335,  336,   -1, 
           -1,  281,  282,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  305,    0,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,   -1,  338,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349, 
           44,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   -1,  363,   58,   59,   -1,   61,   -1,   63, 
           -1,   -1,  279,   -1,   -1,  375,   -1,   -1,   -1,   -1, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,  395,   -1,   91,   -1,   -1, 
           -1,  281,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,  328,   -1,    0,   -1,   -1,   -1,   -1,  335,  336, 
           -1,   -1,   -1,   10,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  342,   -1,  281,  282,   -1,   44,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   58,   59,  363,   61,   -1,   63,   -1,  305,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   91,  395,  333,  334,  335,  336, 
           -1,  338,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,  349,   10,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  375,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,  395,   -1, 
           -1,   -1,   -1,   -1,   61,   -1,   -1,  281,  282,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,    0,   -1,   91,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,   -1,   -1,  333, 
          334,  335,  336,   -1,  338,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,  349,   44,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,   -1,  363, 
           58,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   -1,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,  395,   -1,   91,  281,  282,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,  338,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,  349,   44,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   58,   59,   -1, 
           61,   -1,   63,   -1,   -1,   -1,   -1,   -1,  375,   -1, 
           -1,   -1,  279,   -1,   10,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1, 
           91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  328,    0,   -1,   -1,   61,   -1,   -1,  335,  336, 
           -1,   -1,   10,   -1,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,   91,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  281,  282,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           58,   59,   -1,   -1,   -1,   63,   -1,  305,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   91,   -1,  333,  334,  335,  336,   -1, 
          338,  339,   -1,   -1,  342,   10,   -1,  345,   -1,  347, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,    0,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,   -1,   -1,  375,   -1,   44, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   61,  395,   -1,   -1, 
          281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   59,   -1,   -1,  305,    0,   91,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   10,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           -1,   -1,  333,  334,  335,  336,   -1,  338,  339,   -1, 
           -1,  342,   -1,  279,  345,   -1,  347,   -1,  349,   44, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   -1, 
          361,   -1,  363,   58,   59,   -1,   61,   -1,   63,   -1, 
           -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,  328,  271,  395,   -1,   91,   -1,   -1,  335, 
          336,   -1,  280,  281,  282,   -1,  342,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,    0,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,   -1, 
           -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,  349,   44,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   -1,  361,  362,  363,   58,   59,   -1,   61, 
           -1,   63,   -1,   -1,  279,   -1,   -1,  375,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,   -1,   -1,  395,   -1,   91, 
           -1,   -1,   -1,  281,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,  328,   -1,    0,   -1,   -1,   -1,   -1, 
          335,  336,   -1,   -1,   -1,   10,   -1,  342,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264, 
           -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  342,   -1,  281,  282,   -1,   44, 
           -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294, 
          295,  296,  297,   58,   59,  363,   -1,   -1,   63,   -1, 
          305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   91,  395,  333,  334, 
          335,  336,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1, 
          345,   -1,  347,   -1,  349,   -1,  351,   -1,  353,  354, 
          355,  356,  357,  358,  359,   -1,  361,   -1,  363,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          375,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
          395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  281, 
          282,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  305,    0,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   10,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   -1, 
           -1,  333,  334,  335,  336,   -1,   -1,  339,   -1,   -1, 
          342,   -1,   -1,  345,   -1,  347,   -1,  349,   44,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
           -1,  363,   58,   59,   -1,   -1,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,  375,   -1,   -1,   -1,  262,  263,  264, 
           -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1, 
           -1,   -1,   -1,  395,   -1,   91,  281,  282,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294, 
          295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          305,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   10,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   -1,   -1,  333,  334, 
          335,  336,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1, 
          345,   -1,  347,   -1,  349,   44,  351,   -1,  353,  354, 
          355,  356,  357,  358,  359,   -1,  361,   -1,  363,   58, 
           59,   -1,   -1,   -1,   63,   -1,   -1,   -1,  306,  307, 
          375,   -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317, 
          318,  319,  320,  321,  322,  323,   -1,   -1,  326,  327, 
          395,   -1,   91,  331,  332,  333,  334,   -1,   -1,   -1, 
           -1,   -1,   -1,  341,   -1,   -1,   -1,   -1,   -1,   -1, 
          348,  349,   -1,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,    0,   -1,  364,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  281,  282,   -1,   44,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   58,   59,   -1,   -1,   -1,   63,   -1,  305, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   91,   -1,  333,  334,  335, 
          336,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   -1,  349,   -1,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,    0,  361,   -1,  363,   -1,   -1, 
           -1,   -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,  395, 
           -1,   -1,  281,  282,   -1,   -1,   -1,   -1,   44,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   59,   -1,   -1,  305,    0,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,   -1,   -1, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
          349,   44,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   -1,  361,   -1,  363,   58,   59,   -1,   -1,   -1, 
           63,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,  395,   -1,   91,   -1, 
           -1,   -1,   -1,   -1,   -1,  281,  282,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305, 
            0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   -1,  333,  334,  335, 
          336,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   -1,  349,   44,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   -1,  363,   -1,   59, 
           -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1,  395, 
           -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,    0,   -1,   -1,   -1,  293,  294,  295, 
          296,  297,   -1,   10,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   44,  271,   -1, 
           -1,   -1,   -1,  339,   -1,   -1,  342,   -1,  281,  282, 
           -1,   -1,   59,   -1,   -1,   -1,   63,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,  363,   -1,   -1, 
           -1,   -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   91,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,  395, 
          333,  334,  335,  336,   -1,    0,  339,   -1,   -1,  342, 
           -1,   -1,  345,   -1,  347,   10,  349,   -1,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,   -1,  361,   -1, 
          363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,   44, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,  395,   -1,   59,   -1,   -1,   -1,   63,   -1, 
           -1,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,   -1,   -1,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349, 
           -1,  351,    0,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   10,  363,   -1,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  281,  282,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,  395,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0, 
           -1,   59,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,  349,   44,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   58,   59,   -1, 
           -1,   -1,   63,   -1,   -1,   -1,   -1,  262,  263,  264, 
           -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  281,  282,  395,   -1, 
           -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294, 
          295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,    0,   -1,  333,  334, 
          335,  336,   -1,   -1,  339,   -1,   10,  342,   -1,   -1, 
          345,   -1,  347,   -1,  349,   -1,  351,   -1,  353,  354, 
          355,  356,  357,  358,  359,   -1,  361,   -1,  363,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           44,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   58,   59,   -1,   -1,   -1,   63, 
          395,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  293,  294,  295,  296,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          281,  282,   -1,   44,  342,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   58,   59,   -1, 
           -1,   -1,   63,   -1,  305,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,   -1,  329,  330, 
           -1,   -1,  333,  334,   -1,   -1,   -1,  395,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   -1, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  375,   -1,    0,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  395,   -1,   -1,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1, 
           44,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   58,   59,   -1,   -1,   -1,   63, 
           -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,   -1,  329,  330,   -1,   -1,  333, 
          334,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,  349,   -1,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,   -1,  363, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,  375,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          281,  282,   10,   -1,   -1,   -1,   -1,   -1,   -1,  290, 
          291,  395,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,   -1,  329,  330, 
           58,   59,  333,  334,   -1,   63,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   -1, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1, 
           -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   10,   -1,   -1,  395,   -1,   -1,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1, 
           -1,   -1,   -1,   -1,   -1,   44,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   58, 
           59,  305,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,   -1,  329,  330,   -1,   -1,  333, 
          334,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,  349,   -1,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,    0,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1, 
           -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  395,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   44,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  281,  282,   -1,   -1,   59,   -1,   61, 
           -1,   63,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,    0,   -1,   -1,   -1,  305,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,   91, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
           -1,  329,  330,   -1,   -1,  333,  334,   -1,   -1,   -1, 
           -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   59,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,  375,  267,  268, 
          269,   -1,  271,   -1,    0,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  281,  282,   10,   -1,   -1,  395,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,    0, 
          329,  330,   -1,   59,  333,  334,   -1,   -1,   -1,   10, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
          349,   -1,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   -1,  361,   -1,  363,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   44,   -1,   -1,  375,   -1,   -1,   -1, 
          262,  263,  264,   -1,   -1,   -1,  268,  269,   59,  271, 
           61,   -1,   63,   -1,   -1,   -1,  395,   -1,  280,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,    0, 
           -1,  333,  334,  335,  336,  262,  263,  264,   -1,   10, 
          342,  268,  269,   -1,  271,   -1,   -1,  349,   -1,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
          362,  363,   -1,   -1,   -1,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   -1, 
           61,   -1,   63,  395,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1, 
      };
   }

   private static final short[] yyCheck3() {
      return new short[] {

           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  342,  262,  263,  264,   -1,   91,   -1,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,  363,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,  290,  291,   -1,  293,  294,  295,  296,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  395,   -1,   -1,  262,  263, 
          264,   -1,   -1,   44,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  280,   -1,   59,   -1, 
           61,   -1,   63,  342,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  363,   -1,   -1,   -1,   -1,   -1, 
           91,   -1,   -1,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,    0,   -1,  333, 
          334,  335,  336,   -1,   -1,   -1,  395,   10,  342,   -1, 
           -1,   -1,   -1,   -1,   -1,  349,   -1,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,  362,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   44,   -1,   -1,   -1,   -1,   -1,   -1,  262,  263, 
          264,   -1,   -1,   -1,  268,  269,   59,  271,   61,   -1, 
           63,  395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   91,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,   -1,   -1,  333, 
          334,  335,  336,   -1,  338,   -1,   -1,   -1,  342,   -1, 
           -1,   -1,   -1,   -1,   -1,  349,   -1,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,   -1,  363, 
           -1,  262,  263,  264,   -1,   -1,   -1,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290, 
          291,  395,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
            0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,   -1,   -1,   -1,   -1,   -1,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           -1,   -1,  333,  334,  335,  336,   -1,  338,   -1,   -1, 
           -1,  342,   -1,   -1,   44,   -1,   -1,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   59, 
          361,   61,  363,   63,   -1,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   -1,   -1,  268,  269,   -1,  271,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   91,   -1,   -1,  395,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,    0,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1, 
           -1,   -1,   -1,   -1,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,   -1, 
          333,  334,  335,  336,   -1,  338,   -1,   -1,   -1,  342, 
           -1,   -1,   44,   -1,   -1,   -1,  349,   -1,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,   59,  361,   61, 
          363,   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          306,  307,   -1,   -1,  310,   -1,   -1,   -1,  314,  315, 
           -1,  317,  318,  319,  320,  321,  322,  323,   -1,   91, 
          326,  327,  395,   -1,   -1,  331,  332,  333,  334,   -1, 
           -1,    0,   -1,   -1,   -1,  341,   -1,   -1,   -1,   -1, 
           -1,   10,  348,  349,   -1,  351,  352,  353,  354,  355, 
          356,  357,  358,  359,  360,  361,   -1,   -1,  364,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,   -1,  262,  263,  264,   -1,   -1,   -1,  268,  269, 
           59,  271,   61,   -1,   63,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,    0,   -1,  333,  334,  335,  336,   -1,  338,   -1, 
           -1,   10,  342,   -1,   -1,   -1,   -1,   -1,   -1,  349, 
           -1,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
          262,  263,  264,   -1,   -1,   44,  268,  269,   -1,  271, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           59,   -1,   61,   -1,   63,  395,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   91,   -1,   -1,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   -1, 
           -1,  333,  334,  335,  336,   -1,  338,   -1,   -1,   -1, 
          342,   -1,   -1,   -1,   -1,   -1,   -1,  349,   -1,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
           -1,  363,   -1,  262,  263,  264,   -1,   -1,   -1,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,  395,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,   -1,  338, 
           -1,   -1,   -1,  342,   -1,   -1,   44,   -1,   -1,   -1, 
          349,   -1,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   59,  361,   61,  363,   63,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1,  268, 
          269,   -1,  271,   91,   -1,   -1,  395,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,   -1,  338, 
           -1,   -1,   -1,  342,   -1,   -1,   44,   -1,   -1,   -1, 
          349,   -1,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   59,  361,   61,  363,   63,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  306,  307,   -1,   -1,  310,   -1, 
           -1,   -1,  314,  315,   -1,  317,  318,  319,  320,  321, 
          322,  323,   -1,   91,  326,  327,  395,   -1,   -1,  331, 
          332,  333,  334,   -1,   -1,    0,   -1,   -1,   -1,  341, 
           -1,   -1,   -1,   -1,   -1,   10,  348,  349,   -1,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
           -1,   -1,  364,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,   44, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   59,   -1,   61,   -1,   63,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,    0,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,  395,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           58,   59,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   -1,  361,   -1,  363,   -1,  262,  263,  264, 
           -1,   -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  290,  291,  395,  293,  294, 
          295,  296,  297,   -1,   -1,   -1,    0,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1, 
           -1,   -1,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   -1,   -1,  333,  334, 
          335,  336,   -1,  338,   -1,   -1,   -1,  342,   -1,   -1, 
           44,   -1,   -1,   -1,  349,   -1,  351,   -1,  353,  354, 
          355,  356,  357,  358,  359,   59,  361,   61,  363,   63, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   -1,   -1,   91,   -1,   -1, 
          395,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  281,  282,   -1,   -1,   44,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,    0,   59,   -1,   -1,   -1,   -1,  305,   -1,   -1, 
           -1,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,   -1,   -1,  324,  325,  326,  327, 
           -1,  329,  330,   -1,   -1,  333,  334,   -1,   -1,   -1, 
           -1,  339,   -1,   -1,  342,   44,   -1,  345,   -1,  347, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,   -1, 
           59,  359,   61,  361,   63,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   91,   -1,   -1,   -1,   -1,  395,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,    0,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,  263, 
          264,   -1,   -1,   -1,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   44,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   58,   59,   -1,   -1,   -1, 
           63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,   -1,   -1,  333, 
          334,  335,  336,   -1,   -1,  262,  263,  264,  342,   -1, 
          267,  268,  269,   -1,  271,  349,    0,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   10,  361,   -1,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           44,  395,   -1,  262,  263,  264,   -1,   -1,   -1,  268, 
          269,   -1,  271,   -1,   58,   59,   -1,   -1,   -1,   63, 
           -1,   -1,  339,   -1,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,  395,   -1, 
           -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1,   -1, 
          349,   -1,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   -1,  361,    0,  363,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   10,  267,  268,  269,   -1,  271,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282, 
           -1,   -1,   -1,   -1,   -1,   -1,  395,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   44,   -1,   -1, 
           -1,   -1,  305,    0,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   59,   10,   -1,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,   -1,  329,  330,   -1,   -1, 
          333,  334,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342, 
           -1,   -1,  345,   -1,  347,   -1,   -1,   44,  351,   -1, 
           -1,   -1,  355,  356,  357,  358,  359,   -1,  361,   -1, 
          363,   -1,   59,   -1,   61,   -1,   63,   -1,  262,  263, 
          264,   -1,  375,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1, 
           -1,   -1,  395,   -1,   91,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,   -1,  329,  330,   -1,   -1,  333, 
          334,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,   -1,   44,  351,   -1,   -1, 
           -1,  355,  356,  357,  358,  359,   -1,  361,   -1,  363, 
           58,   59,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1, 
           -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1, 
           -1,  268,  269,   44,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   -1, 
           -1,   -1,   63,  290,  291,  342,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  363,   -1,   -1,   -1, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,   -1,   -1,   -1,   -1,  342,   -1,   -1,  395,   -1, 
           -1,   -1,  349,   -1,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,    0,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  281,  282,   10,   -1,   -1,  395,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   44, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
           -1,  329,  330,   58,   59,  333,  334,   -1,   63,   -1, 
           -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,   -1,   -1,  351,   -1,   -1,   -1,   -1,   -1,  357, 
          358,  359,   -1,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,    0,   -1,   -1,   -1,   -1,  395,   -1,   -1, 
          281,  282,   10,   -1,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,   -1,  329,  330, 
           58,   59,  333,  334,   -1,   63,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1,   -1, 
          351,   -1,   -1,   -1,   -1,   -1,  357,  358,  359,   -1, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1, 
           -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   10,   -1,   -1,  395,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264, 
           -1,   -1,  267,  268,  269,   44,  271,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1,   58, 
           59,   -1,   -1,   -1,   63,  290,  291,   -1,  293,  294, 
          295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,   -1,  329,  330,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  339,   -1,   -1,  342,    0,   -1, 
          345,   -1,  347,   -1,   -1,   -1,  351,   -1,   10,   -1, 
           -1,   -1,  357,  358,  359,   -1,  361,   -1,  363,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          375,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   44,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
          395,   -1,   -1,  281,  282,   -1,   58,   59,   -1,   -1, 
           -1,   63,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
           -1,  329,  330,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  357, 
          358,  359,   -1,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,  375,  267,  268, 
          269,   -1,  271,   -1,    0,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  281,  282,   10,   -1,   -1,  395,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,   -1, 
          329,  330,   58,   59,   -1,   -1,   -1,   63,   -1,   -1, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  357,  358, 
          359,   -1,  361,   -1,  363,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,   -1, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
           -1,    0,   -1,   -1,   -1,   -1,  395,   -1,   -1,  281, 
          282,   10,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  305,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   44,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,   -1,  329,  330,   58, 
           59,   -1,   -1,   -1,   63,   -1,   -1,  339,   -1,   -1, 
          342,   -1,   -1,  345,   -1,  347,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  357,  358,   -1,   -1,   -1, 
           -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,    0,   -1, 
           -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   10,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  395,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   44,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,   58,   59,   -1,   -1, 
           -1,   63,   -1,   -1,   -1,  281,  282,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,    0,   -1,   -1,  305, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1, 
           -1,   -1,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,   -1,  329,  330,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           44,  347,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  357,  358,   -1,   58,   59,   -1,  363,   -1,   63, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,    0,   -1,   -1,   -1,   -1,  395, 
           -1,   -1,  281,  282,   10,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,   -1, 
           -1,   -1,   -1,   -1,   -1,  324,  325,   -1,   -1,   -1, 
          329,  330,   58,   59,   -1,   -1,   -1,   63,   -1,   -1, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          262,  263,  264,   -1,  363,  267,  268,  269,   -1,  271, 
           -1,    0,   -1,   -1,   -1,   -1,  375,   -1,   -1,  281, 
          282,   10,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,  395,   -1,   -1,   -1, 
           -1,   -1,   -1,  305,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,   -1,  324,  325,   -1,   -1,   -1,  329,  330,   58, 
           59,   -1,   -1,   -1,   63,   -1,   -1,  339,   -1,   -1, 
          342,   -1,   -1,  345,   -1,  347,   -1,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,    0, 
           -1,  363,   -1,   -1,   -1,   -1,   -1,  281,  282,   10, 
           -1,   -1,   -1,  375,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,   -1,  395,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1, 
          324,  325,   -1,   -1,   -1,  329,  330,   58,   59,   -1, 
           -1,   -1,   63,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,   -1,    0,   -1,  363, 
           -1,   -1,   -1,   -1,   -1,  281,  282,   10,   -1,   -1, 
           -1,  375,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305, 
           -1,  395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   44,   -1,   -1,   -1,   -1,   -1,   -1,  324,  325, 
           -1,   -1,   -1,  329,  330,   58,   59,   -1,   -1,   -1, 
           63,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,  363,  267,  268, 
          269,   -1,  271,   -1,    0,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,  281,  282,   10,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,  395, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,   -1, 
           -1,   -1,   -1,   -1,   -1,  324,  325,   -1,   -1,   -1, 
          329,  330,   58,   59,   -1,   -1,   -1,   63,   -1,   -1, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,    0,   -1,  363,   -1,   -1,   -1,   -1,   -1, 
          281,  282,   10,   -1,   -1,   -1,  375,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  305,   -1,  395,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,   -1,  324,  325,   -1,   -1,   -1,  329,  330, 
           58,   59,   -1,   -1,   -1,   63,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,   -1, 
            0,   -1,  363,   -1,   -1,   -1,   -1,   -1,  281,  282, 
           10,   -1,   -1,   -1,  375,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  305,   -1,  395,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1, 
           -1,  324,  325,   -1,   -1,   -1,  329,  330,   58,   59, 
           -1,   -1,   -1,   63,   -1,   -1,  339,   -1,   -1,  342, 
           -1,   -1,  345,   -1,  347,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264,    0, 
          363,  267,  268,  269,   -1,  271,   -1,   -1,   -1,   10, 
           -1,   -1,  375,   -1,   -1,  281,  282,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,  395,   -1,   -1,   -1,   -1,   -1,   -1,  305, 
           -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59,  325, 
           -1,   -1,   -1,  329,  330,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   -1,   -1,  262,  263,  264,   -1,    0,  267, 
          268,  269,   -1,  271,   -1,   -1,   -1,  363,   10,   -1, 
           -1,   -1,   -1,  281,  282,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,  395, 
           -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   58,   59,   -1,   -1, 
            0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,  281,  282,   -1,   44,   -1,   -1,  375,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   58,   59, 
           -1,   -1,   -1,   -1,   -1,  305,   -1,  395,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
          281,  282,   44,   -1,   -1,  375,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   59,   -1,   -1, 
           -1,   -1,   -1,   -1,  305,  395,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,   10, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1,   -1, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
           -1,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,  281, 
          282,   -1,   -1,   44,  375,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   58,   59,   -1, 
           -1,   -1,   -1,  305,  395,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  262,  263,  264,    0,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,   10,   -1,  339,   -1,   -1, 
          342,  281,  282,  345,   -1,  347,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,  363,    0,   -1,   -1,  305,   -1,   -1,   -1,   44, 
           -1,   -1,   10,  375,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   59,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  395,   -1,   -1,   -1,   -1,   -1,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
           -1,   59,   -1,  363,   -1,   -1,   -1,   -1,   -1,  281, 
          282,   -1,   -1,   -1,   -1,  375,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
            0,   -1,   -1,  305,   -1,  395,   -1,   -1,   -1,   -1, 
           10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1, 
          342,   -1,   -1,  345,   -1,  347,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   59, 
          271,  363,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          281,  282,   10,  375,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  395,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,    0, 
           -1,   59,   -1,   -1,   -1,   -1,   -1,   -1,  339,   10, 
           -1,  342,   -1,   -1,  345,   -1,  347,  262,  263,  264, 
           -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1, 
           -1,   -1,  363,   -1,   -1,   -1,  281,  282,   -1,   -1, 
           -1,   -1,   59,   -1,  375,  290,  291,   -1,  293,  294, 
          295,  296,  297,   -1,  262,  263,  264,   -1,   59,  267, 
          268,  269,   -1,  271,  395,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  281,  282,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  342,   -1,   -1, 
          345,   -1,  347,   -1,   -1,   -1,   -1,   -1,   10,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  363,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
          395,  271,   -1,   -1,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  342,  281,  282,  345,   -1,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,  363,   -1,  262,  263,  264,   10,   -1, 
          267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,  281,  282,   -1,  268,  269,   -1, 
          271,   -1,   -1,  290,  291,  395,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,  342,   -1,   -1,  345,   -1,   -1, 
           -1,   -1,  293,  294,  295,  296,  297,   59,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  342,   -1,   -1,  345,   91, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   -1, 
           -1,  342,   -1,   -1,   -1,   -1,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  257,  258,  259,   -1,  261, 
           -1,   -1,  363,  265,  266,   -1,   -1,   -1,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,  395,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,  395,   -1,   -1,  299,   -1,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
          312,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   59,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,  350,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   91, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,  394,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,  280,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           -1,   -1,   -1,  335,  336,  337,   -1,   59,  340,  341, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,  357,   -1,   -1,  360,   -1, 
          362,   -1,   -1,  365,  366,  367,  368,  369,  370,   91, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,  280,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           -1,   -1,   -1,  335,  336,  337,   -1,   59,  340,  341, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,  357,   -1,   -1,  360,   -1, 
          362,   -1,   -1,  365,  366,  367,  368,  369,  370,   91, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,  280,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           -1,   -1,   -1,  335,  336,  337,   -1,   59,  340,  341, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,  357,   -1,   -1,  360,   -1, 
          362,   -1,   -1,  365,  366,  367,  368,  369,  370,   91, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           -1,   -1,   -1,  335,  336,  337,   -1,   59,  340,  341, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,  357,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           -1,   -1,   -1,  335,  336,  337,   -1,   59,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,  357,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,  268,  269,  270,  271, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   59,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   10,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   59,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
          362,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   10,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   -1,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   59,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          392,  393,   -1,   -1,  256,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,   -1,   -1,  269,  270,   -1, 
          272,  273,  274, 
      };
   }

   private static final short[] yyCheck4() {
      return new short[] {

          275,  276,  277,  278,   10,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,  298,  299,   -1,  301,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,  314, 
          315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   59,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344, 
           -1,  346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,  256,  257,  258,  259,  260,  261,  392,  393,  264, 
          265,  266,   -1,   -1,   -1,  270,   -1,  272,  273,  274, 
          275,  276,  277,  278,   10,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,  298,  299,   -1,  301,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,  314, 
          315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344, 
           -1,  346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1, 
          256,  257,  258,  259,  260,  261,   -1,   -1,   -1,  265, 
          266,   -1,   -1,   10,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   59,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,  363,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,  260,  261,  392,  393,   -1,  265, 
          266,   -1,   10,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,  342,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,   -1,  392,  393,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   44,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,   -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,  257, 
          258,  259,   -1,  261,   -1,  392,  393,  265,  266,   -1, 
           -1,   -1,  270,   -1,  272,  273,  274,  275,  276,  277, 
          278,   -1,   -1,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,   44,  313,  314,  315,  316,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337, 
           -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1, 
           -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,   -1,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  392,  393,   -1,   -1,  257,  258, 
          259,   -1,  261,   -1,   -1,   -1,  265,  266,   -1,   -1, 
           -1,  270,   -1,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,   -1,  313,   -1,   -1,  316,   59,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1, 
           -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,   -1,  392,  393,  265,  266,   -1,   -1,   -1, 
          270,   -1,  272,  273,  274,  275,  276,  277,  278,   -1, 
           -1,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
           -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,   -1,  313,   -1,   -1,  316,   59,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1, 
          340,   -1,   -1,  343,  344,   -1,  346,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,  257,  258,  259,  260, 
          261,   -1,  392,  393,  265,  266,   -1,   -1,   -1,  270, 
           -1,  272,  273,  274,  275,  276,  277,  278,   -1,   -1, 
           -1,   -1,  283,  284,  285,  286,  287,  288,  289,   -1, 
           -1,  292,   -1,   -1,   -1,   -1,   -1,  298,  299,  300, 
          301,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,   -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340, 
           -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1, 
           -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360, 
           -1,   -1,   -1,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,  257,  258,  259,  260,  261, 
           -1,  392,  393,  265,  266,   -1,   -1,   -1,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   -1,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,  256,  257,  258,  259,  260,  261,   -1, 
          392,  393,  265,  266,   -1,   -1,   -1,  270,   -1,  272, 
          273,  274,  275,  276,  277,  278,   -1,   -1,   -1,   -1, 
          283,  284,  285,  286,  287,  288,  289,   -1,   -1,  292, 
           -1,   -1,   -1,   -1,   -1,  298,  299,  300,  301,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,   -1, 
          313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1, 
          343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1,  352, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1, 
           -1,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,  256,  257,  258,  259,  260,  261,   -1,  392, 
          393,  265,  266,   -1,   -1,   -1,  270,   -1,  272,  273, 
          274,  275,  276,  277,  278,   -1,   -1,   -1,   -1,  283, 
          284,  285,  286,  287,  288,  289,   -1,   -1,  292,   -1, 
           -1,   -1,   -1,   -1,  298,  299,   -1,  301,  302,  303, 
          304,   -1,  306,  307,  308,  309,  310,  311,   -1,  313, 
          314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343, 
          344,   -1,  346,   -1,  348,   -1,   -1,   -1,  352,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1, 
           -1,  365,  366,  367,  368,  369,  370,   -1,   -1,   -1, 
          374,   -1,  376,  377,   -1,  379,  380,  381,  382,   -1, 
           -1,  256,  257,  258,  259,  260,  261,   -1,  392,  393, 
          265,  266,   -1,   -1,   -1,  270,   -1,  272,  273,  274, 
          275,  276,  277,  278,   -1,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,  298,  299,   -1,  301,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,  314, 
          315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344, 
           -1,  346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,   -1,  392,  393,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,  350,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,  392,  393,  394,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,  350,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,  392,  393,  394,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,  350,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,  392,  393,  394,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,  392,  393,  394,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,  392,  393,  394,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,  392,  393,  394,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,   -1,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,  257,  258,  259,  260,  261,  392,  393,  394,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,  300,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
          257,  258,  259,  260,  261,   -1,  392,  393,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,   -1,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,  257, 
          258,  259,   -1,  261,   -1,  392,  393,  265,  266,   -1, 
           -1,   -1,  270,   -1,  272,  273,  274,  275,  276,  277, 
          278,   -1,   -1,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,   -1,  313,  314,  315,  316,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337, 
           -1,   -1,  340,  341,   -1,  343,  344,   -1,  346,   -1, 
           -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,   -1,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,  257,  258, 
          259,   -1,  261,   -1,  392,  393,  265,  266,   -1,   -1, 
           -1,  270,   -1,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,   -1,  313,  314,  315,  316,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1, 
           -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1,  348, 
           -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,   -1,  392,  393,  265,  266,   -1,   -1,   -1, 
          270,   -1,  272,  273,  274,  275,  276,  277,  278,   -1, 
           -1,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
           -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,   -1,  313,  314,  315,  316,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1, 
          340,   -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1, 
           -1,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          360,   -1,   -1,   -1,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,  257,  258,  259,   -1, 
          261,   -1,  392,  393,  265,  266,   -1,   -1,   -1,  270, 
           -1,  272,  273,  274,  275,  276,  277,  278,   -1,   -1, 
           -1,   -1,  283,  284,  285,  286,  287,  288,  289,   -1, 
           -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,   -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340, 
           -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1, 
           -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360, 
           -1,   -1,   -1,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,  257,  258,  259,   -1,  261, 
           -1,  392,  393,  265,  266,   -1,   -1,   -1,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   -1,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1, 
          392,  393,  265,  266,   -1,   -1,   -1,  270,   -1,  272, 
          273,  274,  275,  276,  277,  278,   -1,   -1,   -1,   -1, 
          283,  284,  285,  286,  287,  288,  289,   -1,   -1,  292, 
           -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,   -1, 
          313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1, 
          343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1,  352, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1, 
           -1,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1,  392, 
          393,  265,  266,   -1,   -1,   -1,  270,   -1,  272,  273, 
          274,  275,  276,  277,  278,   -1,   -1,   -1,   -1,  283, 
          284,  285,  286,  287,  288,  289,   -1,   -1,  292,   -1, 
           -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303, 
          304,   -1,  306,  307,  308,  309,  310,  311,   -1,  313, 
          314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  337,   -1,   -1,  340,  341,   -1,  343, 
          344,   -1,  346,   -1,   -1,   -1,   -1,   -1,  352,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1, 
           -1,  365,  366,  367,  368,  369,  370,   -1,   -1,   -1, 
          374,   -1,  376,  377,   -1,  379,  380,  381,  382,   -1, 
           -1,   -1,  257,  258,  259,   -1,  261,   -1,  392,  393, 
          265,  266,   -1,   -1,   -1,  270,   -1,  272,  273,  274, 
          275,  276,  277,  278,   -1,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,  314, 
          315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344, 
           -1,  346,   -1,   -1,   -1,   -1,   -1,  352,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,   -1,  392,  393,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,   -1,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,   -1,  392,  393,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,   -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,  257, 
          258,  259,   -1,  261,   -1,  392,  393,  265,  266,   -1, 
           -1,   -1,  270,   -1,  272,  273,  274,  275,  276,  277, 
          278,   -1,   -1,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,   -1,  313,  314,  315,  316,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337, 
           -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1, 
           -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,   -1,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,  257,  258, 
          259,   -1,  261,   -1,  392,  393,  265,  266,   -1,   -1, 
           -1,  270,   -1,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,   -1,  313,   -1,   -1,  316,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  333,   -1,   -1,   -1,  337,   -1, 
           -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,   -1,  392,  393,  265,  266,   -1,   -1,   -1, 
          270,   -1,  272,  273,  274,  275,  276,  277,  278,   -1, 
           -1,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
           -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,   -1,  313,   -1,   -1,  316,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1, 
          340,   -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,  257,  258,  259,   -1, 
          261,   -1,  392,  393,  265,  266,   -1,   -1,   -1,  270, 
           -1,  272,  273,  274,  275,  276,  277,  278,   -1,   -1, 
           -1,   -1,  283,  284,  285,  286,  287,  288,  289,   -1, 
           -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,   -1,  313,   -1,   -1,  316,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340, 
           -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,  257,  258,  259,   -1,  261, 
           -1,  392,  393,  265,  266,   -1,   -1,   -1,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   -1,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1, 
          392,  393,  265,  266,   -1,   -1,   -1,  270,   -1,  272, 
          273,  274,  275,  276,  277,  278,   -1,   -1,   -1,   -1, 
          283,  284,  285,  286,  287,  288,  289,   -1,   -1,  292, 
           -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,   -1, 
          313,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1, 
          343,  344,   -1,  346,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1,  392, 
          393,  265,  266,   -1,   -1,   -1,  270,   -1,  272,  273, 
          274,  275,  276,  277,  278,   -1,   -1,   -1,   -1,  283, 
          284,  285,  286,  287,  288,  289,   -1,   -1,  292,   -1, 
           -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303, 
          304,   -1,  306,  307,  308,  309,  310,  311,   -1,  313, 
           -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343, 
          344,   -1,  346,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  365,  366,  367,  368,  369,  370,   -1,   -1,   -1, 
          374,   -1,  376,  377,   -1,  379,  380,  381,  382,   -1, 
           -1,   -1,  257,  258,  259,   -1,  261,   -1,  392,  393, 
          265,  266,   -1,   -1,   -1,  270,   -1,  272,  273,  274, 
          275,  276,  277,  278,   -1,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,   -1, 
           -1,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344, 
           -1,  346,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  257, 
          258,  259,  260,  261,  262,  263,  264,  265,  266,  267, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,  279,  280,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
          298,  299,  300,  301,  302,  303,  304,   -1,  306,  307, 
          308,   -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317, 
          318,  319,  320,  321,  322,  323,   -1,   -1,  326,  327, 
           -1,   -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          348,  349,   -1,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,   -1,   -1,  364,  365,  257,  258, 
          259,  260,  261,  262,  263,  264,  265,  266,  267,  268, 
          269,  270,  271,  272,  273,  274,  275,  276,  277,  278, 
          279,  280,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,  290,  291,  292,  293,  294,  295,  296,  297,  298, 
          299,  300,  301,  302,  303,  304,   -1,  306,  307,   -1, 
           -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317,  318, 
          319,  320,  321,  322,  323,   -1,   -1,  326,  327,   -1, 
           -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348, 
          349,   -1,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,   -1,   -1,  364,  365,  257,  258,  259, 
          260,  261,  262,  263,  264,  265,  266,  267,  268,  269, 
          270,  271,  272,  273,  274,  275,  276,  277,  278,  279, 
          280,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
          290,  291,  292,  293,  294,  295,  296,  297,  298,  299, 
          300,  301,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,   -1,   -1,  314,  315,   -1,  317,  318,  319, 
          320,  321,  322,  323,   -1,   -1,  326,  327,   -1,   -1, 
           -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1, 
           -1,  341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349, 
           -1,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,   -1,   -1,  364,  257,  258,  259,  260,  261, 
          262,  263,  264,  265,  266,  267,  268,  269,  270,  271, 
          272,  273,  274,  275,  276,  277,  278,  279,  280,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,  290,  291, 
          292,  293,  294,  295,  296,  297,  298,  299,  300,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,   -1,  314,  315,   -1,  317,  318,  319,  320,  321, 
          322,  323,   -1,   -1,  326,  327,   -1,   -1,   -1,  331, 
          332,  333,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
           -1,   -1,  364,  257,  258,  259,  260,  261,  262,  263, 
          264,  265,  266,  267,  268,  269,  270,  271,  272,  273, 
          274,  275,  276,  277,  278,  279,  280,   -1,   -1,  283, 
          284,  285,  286,  287,  288,  289,  290,  291,  292,  293, 
          294,  295,  296,  297,  298,  299,  300,  301,  302,  303, 
          304,   -1,  306,  307,   -1,   -1,  310,   -1,   -1,   -1, 
          314,  315,   -1,  317,  318,  319,  320,  321,  322,  323, 
           -1,   -1,  326,  327,   -1,   -1,   -1,  331,  332,  333, 
          334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  348,  349,   -1,  351,  352,  353, 
          354,  355,  356,  357,  358,  359,  360,  361,  306,  307, 
          364,   -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317, 
          318,  319,  320,  321,  322,  323,   -1,   -1,  326,  327, 
           -1,   -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1, 
           -1,   -1,   -1,  341,   -1,   -1,   -1,   -1,   -1,   -1, 
          348,  349,   -1,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  306,  307,  364,   -1,  310,   -1, 
           -1,   -1,  314,  315,   -1,  317,  318,  319,  320,  321, 
          322,  323,   -1,   -1,  326,  327,   -1,   -1,   -1,  331, 
          332,  333,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341, 
           -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
          306,  307,  364,   -1,  310,   -1,   -1,   -1,  314,  315, 
           -1,  317,  318,  319,  320,  321,  322,  323,   -1,   -1, 
          326,  327,   -1,   -1,   -1,  331,  332,  333,  334,   -1, 
           -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,   -1,   -1, 
           -1,   -1,  348,  349,   -1,  351,  352,  353,  354,  355, 
          356,  357,  358,  359,  360,  361,  306,  307,  364,   -1, 
          310,   -1,   -1,   -1,  314,  315,   -1,  317,  318,  319, 
          320,  321,  322,  323,   -1,   -1,  326,  327,   -1,   -1, 
           -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1, 
           -1,  341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349, 
           -1,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,  306,  307,  364,   -1,  310,   -1,   -1,   -1, 
          314,  315,   -1,  317,  318,  319,  320,  321,  322,  323, 
           -1,   -1,  326,  327,   -1,   -1,   -1,  331,  332,  333, 
          334,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1, 
           -1,   -1,   -1,   -1,  348,  349,   -1,  351,  352,  353, 
          354,  355,  356,  357,  358,  359,  360,  361,  306,  307, 
          364,   -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317, 
          318,  319,  320,  321,  322,  323,   -1,   -1,  326,  327, 
           -1,   -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1, 
           -1,   -1,   -1,  341,   -1,   -1,   -1,   -1,   -1,   -1, 
          348,  349,   -1,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  306,  307,  364,   -1,  310,   -1, 
           -1,   -1,  314,  315,   -1,  317,  318,  319,  320,  321, 
          322,  323,   -1,   -1,  326,  327,   -1,   -1,   -1,  331, 
          332,  333,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341, 
           -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
           -1,   -1,  364, 
      };
   }

}
